﻿using System;
using System.Collections.Generic;
using System.Text;
using Ice.Orm.Core.Entity; using Ice.Orm.Core.Criteria;
using Ice.Net.Ecommerce.BusinessObjects.Database.VISTACART;
using Ice.Net.Ecommerce.CacheProvider;

namespace Ice.Net.Ecommerce.BusinessObjects.Common
{
	public class PagedCollection<T>
	{
		public int totalRecords;
		public List<T> collection;
	}

	public class SmartObjectLoader
	{
		/// <summary>
		/// Load from ApplicationStore, if not found, load from database, if not found, return null, else SmartObject.
		/// </summary>
		/// <typeparam name="T"></typeparam>
		/// <param name="identityID"></param>
		/// <returns></returns>
		public static T LoadSmartObject<T>(int identityID) where T : EntityBase
		{
			string key = typeof(T).FullName + "_" + identityID;

			T smartObject =  ApplicationStore.RetrieveByName<T>(key);
			if (smartObject == null)
			{
				// not loaded from ApplicationStore, load from database
				smartObject = LoadFromDBByIdentity<T>(identityID, true);

				if (smartObject != null)
				{
					// save to application store
					ApplicationStore.Save(smartObject, key);
				}
			}
			return smartObject;
		}


		/// <summary>
		/// <para>Load SmartObject by identityID from database.</para> 
		/// <para>If not loaded, returns an clean instance or null if nullIfNotLoaded == true</para>
		/// </summary>
		/// <typeparam name="T"></typeparam>
		/// <param name="identityID"></param>
		/// <returns></returns>
		public static T LoadFromDBByIdentity<T>(int identityID, bool nullIfNotLoaded) where T : EntityBase
		{
			T smartObject = System.Activator.CreateInstance<T>();
			if (smartObject.LoadFromDB(identityID) == false && nullIfNotLoaded)
			{
				return null;
			}
			else
			{
				return smartObject;
			}
		}

		/// <summary>
		/// <para>Load SmartObject by identity from database</para>
		/// <para>If fails, load by defaultIdentityID. If fails too, return null.</para>
		/// </summary>
		/// <typeparam name="T"></typeparam>
		/// <param name="identityID"></param>
		/// <param name="defaultIdentityID"></param>
		/// <returns></returns>
		public static T LoadFromDBByIdentityOrDefault<T>(int identityID, int defaultIdentityID) where T : EntityBase
		{
			T smartObject = System.Activator.CreateInstance<T>();
			if (smartObject.LoadFromDB(identityID) || smartObject.LoadFromDB(defaultIdentityID))
			{
				return smartObject;
			}
			else
			{
				return null;
			}
		}

		public static List<T> LoadCollection<T>(int maxObjects, bool activeOnly)
			where T : EntityBase
		{
			return LoadCollection<T>(maxObjects, activeOnly, null, false);
		}

		public static List<T> LoadCollection<T>(int maxObjects, bool activeOnly, object orderByColumn, bool orderAsc)
			where T : EntityBase
		{
			return new Query()
			.Case(activeOnly, o => o.Where(new SqlName("IsActive"), true))
			.Case(orderByColumn != null, o => o.OrderBy(new SqlName(orderByColumn + ""), !orderAsc))
			.Take(maxObjects)
			.ToList<T>();
		}

		public static PagedCollection<T> LoadCollectionPaged<T>(int recordsPerPage, int currentPageIndex, object orderByColumn, bool orderAsc, bool activeOnly)
			where T : EntityBase
		{
			List<T> collection = new Query()
			.Case(activeOnly, o => o.Where(new SqlName("IsActive"), true))
			.Case(orderByColumn != null, o => o.OrderBy(new SqlName(orderByColumn + ""), !orderAsc))
			.Skip(currentPageIndex * recordsPerPage)
			.Take(recordsPerPage)
			.ToList<T>();

			EntitySetting es = EntitySetting.Get<T>();

			Query queryCount = new Query()
			.Count(new SqlName(es.SchemaTableHelper.TableName))
			.Case(activeOnly, o => o.Where(new SqlName("IsActive"), true));

			PagedCollection<T> pc = new PagedCollection<T>();
			pc.totalRecords = queryCount.ExecuteScalar<T, int>();
			pc.collection = collection;
			return pc;
		}

	}
}
