﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using System.Data.SqlClient;
using Ice.Orm.Core.Entity; using Ice.Orm.Core.Criteria;

namespace Ice.Net.Ecommerce.BusinessObjects.Database.VISTACART
{
	public partial class Order
	{
		// PREDEFINED ORDER STATUSES
		public enum OrderStatusEnum
		{
			Unknown = 0,
			ShoppingCart = 1,
			Ordered = 2,
			Shipped = 3
		}

		// PREDEFINED PAYMENT STATUSES
		public enum PaymentStatusEnum
		{
			/// <summary>
			/// (-2)  A customer initiated a chargeback with his credit card company to get her money back.
			/// </summary>
			Chargeback = -2,
			/// <summary>
			/// (-1) successfully refunded the customer's credit card
			/// </summary>
			Refunded = -1,
			/// <summary>
			/// (0)
			/// </summary>
			Unknown = 0,
			/// <summary>
			/// (1) order not processed
			/// </summary>
			NotProcessed = 1,
			/// <summary>
			/// (2) order not paid yet
			/// </summary>
			NotYetApproved = 2,
			/// <summary>
			/// (3) rejected
			/// </summary>
			Rejected = 3,
			/// <summary>
			/// (4) order is waiting for approval
			/// </summary>
			Waiting = 4,
			/// <summary>
			/// (5) order is peding approval
			/// </summary>
			Pending = 5,
			/// <summary>
			/// (6) CC approved
			/// </summary>
			Approved = 6,
			/// <summary>
			/// (7) CC charged, order paid
			/// </summary>
			Paid = 7
		}

		#region Database methods
		// Database methods

		public static Order LoadByUserIDAndOrderStatusEnum(int userID, OrderStatusEnum orderStatusEnum)
		{
			return new Query<Order>().Where(o => o.UserID == userID && o.OrderStatusEnumId == (int)orderStatusEnum).FirstOrDefault();
		}


		#endregion

		public PaymentStatusEnum PaymentStatus
		{
			get { return (PaymentStatusEnum)this.PaymentStatusEnumId.GetValueOrDefault(); }
			set { this.PaymentStatusEnumId = (int)value; }
		}

		public OrderStatusEnum OrderStatus
		{
			get { return (OrderStatusEnum)this.OrderStatusEnumId.GetValueOrDefault(); }
			set { this.OrderStatusEnumId = (int)value; }
		}

		private OrderDetailCollection mOrderDetailCollection = null;

		public OrderDetailCollection GetOrderDetailCollection()
		{
			if (mOrderDetailCollection == null)
			{
				mOrderDetailCollection = new OrderDetailCollection();
				mOrderDetailCollection.AddRange(OrderDetailCollection.LoadByOrderID(this.OrderID));
			}
			return mOrderDetailCollection;
		}

		private List<LicenseInfo> mLicenseInfoCollection = null;

		public List<LicenseInfo> GetLicenseInfoCollection()
		{
			if (mLicenseInfoCollection == null)
			{
				mLicenseInfoCollection = LicenseInfoCollection.LoadByOrderID(this.OrderID);
			}
			return mLicenseInfoCollection;
		}

		public void SetOrderDetailCollection(OrderDetailCollection orderDetails)
		{
			mOrderDetailCollection = orderDetails;
		}


		///////// <summary>
		///////// Save Order to database, OrderDetailCollection is not saved.
		///////// </summary>
		///////// <returns></returns>
		//////public override bool SaveToDBLegacy()
		//////{
		//////    return this.SaveToDB(false);
		//////}

		/// <summary>
		/// <para>true: save Order and OrderDetailCollection to database.</para>
		/// <para>false: save Order object only.</para>
		/// </summary>
		/// <returns></returns>
		public bool SaveToDBInternal(bool updateOrderDetails)
		{
			if (updateOrderDetails != true)  // save Order, don't save OrderDetailCollection
			{
				base.SaveToDB(WriteModeEnum.SaveAllWithIDCheck);
				return true;
			}
			else  // save Order and OrderDetailCollection
			{
				base.SaveToDB(WriteModeEnum.SaveAllWithIDCheck);

				List<OrderDetail> detailColection = GetOrderDetailCollection();

				new Query<OrderDetail>().DeleteDirect(o => o.OrderID == this.OrderID);

				foreach (OrderDetail detail in detailColection)
				{
					detail.OrderID = this.OrderID;
					detail.OrderDetailID = 0; // insert new
				}
				detailColection.ForEach(o => o.SaveToDB(WriteModeEnum.SaveAllWithIDCheck));

				return true;
			}
			return false;
		}

		/// <summary>
		/// Delete Order, OrderDetailCollection, BillingAddress, ShippingAddress and LicenseInfoCollection from database.
		/// </summary>
		/// <returns></returns>
		public override bool DeleteFromDB()
		{
			if (this.OrderID < 1)
			{
				return false;
			}
			bool success = true;
			foreach (OrderDetail detail in GetOrderDetailCollection())
			{
				success = success && detail.DeleteFromDB();
			}
			foreach (LicenseInfo licInfo in GetLicenseInfoCollection())
			{
				success = success && licInfo.DeleteFromDB();
			}
			success = success && base.DeleteFromDB();

			if (this.BillingAddressID > 0)
			{
				Address billingAddress = new Address();
				billingAddress.AddressID = this.BillingAddressID.GetValueOrDefault();
				success = success && billingAddress.DeleteFromDB();
			}
			if (this.ShippingAddressID > 0 && this.ShippingAddressID != this.BillingAddressID)
			{
				Address shippingAddress = new Address();
				shippingAddress.AddressID = this.ShippingAddressID.GetValueOrDefault();
				success = success && shippingAddress.DeleteFromDB();
			}
			return success;
		}

		public bool RecalculateRedundantValues()
		{
			if (Ecommerce.BusinessObjects.Common.VISTACARTHelper.AppDefaults.AppGlobalSettingsItem.SingleCartItem.GetValueOrDefault())
			{
				return RecalculateRedundantValues2();
			}
			else
			{
				return RecalculateRedundantValues3();
			}
		}

		/// <summary>
		/// Recalculate Redundant prices based on current Product price, tax, shipping, weight...
		/// </summary>
		/// <returns></returns>
		private bool RecalculateRedundantValues2()
		{
			this.RedundantProductsSupplyPrice = 0;
			this.RedundantProductsSalePrice = 0;
			this.RedundantTotalProductsCount = 0;
			this.RedundantProductsTaxPrice = 0;
			this.RedundantShippingPrice = 0;
			this.RedundantShippingTaxPrice = 0;
			this.RendundantTotalWeight = 0;
			this.RedundantTotalOrderPrice = 0;
			this.RedundantTotalTaxPrice = 0;
			this.RedundantTotalToPay = 0;

			OrderDetailCollection detailCollection = GetOrderDetailCollection();
			List<OrderDetail> toBeRemovedFromOrderDetailCollection = new List<OrderDetail>();

			// we have to remember campaign if ApplyToNextNProducts > 0, meaning don't apply to current product, but to next N products
			CampaignHelper campaignHelper = null;
			int creditLeft = 0;

			foreach (OrderDetail detail in detailCollection)
			{
				if(detail.Quantity > 0 && detail.GetProduct().IsActive && detail.GetProduct().StockQuantity > 0)
				{
					if (campaignHelper != null && creditLeft > 0)
					{
						// campaign helper wasn't used for product for which it was retrieved, use campaign helper with current product
						// use campaign helper for next N products
						if (campaignHelper.Product.SalePrice >= detail.GetProduct().SalePrice)
						{
							// use credit
							detail.RecalculateRedundantValues(campaignHelper);
							creditLeft = creditLeft - detail.Quantity;
						}
						else
						{
							// don't use credit
							detail.RecalculateRedundantValues();
						}
					}

					if (campaignHelper == null)
					{
						detail.RecalculateRedundantValues();

						campaignHelper = detail.GetProduct().GetCampaignHelper(); // get campaign helper
						if (campaignHelper.ApplyToCurrentProduct)
						{
							campaignHelper = null; // campaign helper was used for current product, don't use it again
						}
						else
						{
							creditLeft = detail.Quantity * campaignHelper.Campaign.ApplyToNextNProducts.GetValueOrDefault();
						}
					}

					if (creditLeft <= 0)
					{
						creditLeft = 0;
						campaignHelper = null;
					}
					
					this.RedundantProductsSupplyPrice += detail.Quantity * detail.RedundantSupplyPrice;
					this.RedundantProductsSalePrice += detail.Quantity * detail.RedundantSalePrice;
					this.RedundantTotalProductsCount += detail.Quantity;
					this.RedundantProductsTaxPrice += detail.RedundantTaxPrice * detail.Quantity;
					this.RendundantTotalWeight += detail.Quantity * detail.RedundantWeight;
				}
				else
				{
					// product doesn't exist or not active or quantity < 1, don't add it to newDetailCollection
					detail.Quantity = 0;
					toBeRemovedFromOrderDetailCollection.Add(detail);
					detailCollection.ToBeRemovedFromDB(detail);
				}
			}

			foreach (OrderDetail detail in toBeRemovedFromOrderDetailCollection)
			{
				detailCollection.Remove(detail);
			}

			if (!Ecommerce.BusinessObjects.Common.VISTACARTHelper.AppDefaults.AppGlobalSettingsItem.SingleCartItem.GetValueOrDefault())
			{
				bool recalculateRequired = false;

				// merge OrderDetail where values are the same
				// OrderDetailCollection mergedDetailCollection = new OrderDetailCollection();
				for (int nn = 0; nn < detailCollection.Count; nn++)
				{
					OrderDetail detail = detailCollection[nn];

					toBeRemovedFromOrderDetailCollection.Clear();
					// find next OrderDetail whith the same values
					for (int kk = nn + 1; kk < detailCollection.Count; kk++)
					{
						OrderDetail nextDetail = detailCollection[kk];
						if (nextDetail.OrderID == detail.OrderID
							&& nextDetail.ProductID == detail.ProductID
							&& nextDetail.RedundantSalePrice == detail.RedundantSalePrice
							&& nextDetail.Attributes == detail.Attributes)
						{
							detail.Quantity += nextDetail.Quantity;

							nextDetail.Quantity = 0;
							toBeRemovedFromOrderDetailCollection.Add(nextDetail);
							detailCollection.ToBeRemovedFromDB(nextDetail);

							recalculateRequired = true;
						}
					}

					foreach (OrderDetail odToRemove in toBeRemovedFromOrderDetailCollection)
					{
						detailCollection.Remove(odToRemove);
					}
				}

				if (recalculateRequired)
				{
					RecalculateRedundantValues2();
				}
			}

			// add shipping price and tax
			if (this.RedundantTotalProductsCount > 0)
			{
				ShippingType shipping = new ShippingType();
				if (shipping.LoadFromDB(this.ShippingTypeID))
				{
					this.RedundantShippingPrice = shipping.StartPrice + (shipping.PricePerWeight * this.RendundantTotalWeight);
					this.RedundantShippingTaxPrice = Math.Round(((this.RedundantShippingPrice * BusinessObjects.Common.VISTACARTHelper.AppDefaults.Tax.RatePercent / 100)), 2);
				}
			}

			// Totals
			// this.RendundantTotalWeight // we already have it
			this.RedundantTotalTaxPrice = this.RedundantProductsTaxPrice + this.RedundantShippingTaxPrice;
			this.RedundantTotalOrderPrice = this.RedundantProductsSalePrice + this.RedundantShippingPrice;
			this.RedundantTotalToPay = this.RedundantTotalTaxPrice + this.RedundantTotalOrderPrice;

			return true;
		}

		/// <summary>
		/// Recalculate Redundant prices based on current Product price, tax, shipping, weight...
		/// </summary>
		/// <returns></returns>
		public bool RecalculateRedundantValues3()
		{
			this.RedundantProductsSupplyPrice = 0;
			this.RedundantProductsSalePrice = 0;
			this.RedundantTotalProductsCount = 0;
			this.RedundantProductsTaxPrice = 0;
			this.RedundantShippingPrice = 0;
			this.RedundantShippingTaxPrice = 0;
			this.RendundantTotalWeight = 0;
			this.RedundantTotalOrderPrice = 0;
			this.RedundantTotalTaxPrice = 0;
			this.RedundantTotalToPay = 0;

			OrderDetailCollection detailCollection = GetOrderDetailCollection();
			OrderDetailCollection toBeRemovedFromOrderDetailCollection = new OrderDetailCollection();

			foreach (OrderDetail detail in detailCollection)
			{
				if (detail.Quantity > 0 && detail.GetProduct().IsActive && detail.GetProduct().StockQuantity > 0)
				{
					detail.RecalculateRedundantValues();

					this.RedundantProductsSupplyPrice += detail.Quantity * detail.RedundantSupplyPrice;
					this.RedundantProductsSalePrice += detail.Quantity * detail.RedundantSalePrice;
					this.RedundantTotalProductsCount += detail.Quantity;
					this.RedundantProductsTaxPrice += detail.RedundantTaxPrice * detail.Quantity;
					this.RendundantTotalWeight += detail.Quantity * detail.RedundantWeight;
				}
				else
				{
					// product doesn't exist or not active or quantity < 1, don't add it to newDetailCollection
					detail.Quantity = 0;
					toBeRemovedFromOrderDetailCollection.Add(detail);
					detailCollection.ToBeRemovedFromDB(detail);
				}
			}

			foreach (OrderDetail detail in toBeRemovedFromOrderDetailCollection)
			{
				detailCollection.Remove(detail);
			}

			// add shipping price and tax
			if (this.RedundantTotalProductsCount > 0)
			{
				ShippingType shipping = new ShippingType();
				if (shipping.LoadFromDB(this.ShippingTypeID))
				{
					this.RedundantShippingPrice = shipping.StartPrice + (shipping.PricePerWeight * this.RendundantTotalWeight);
					this.RedundantShippingTaxPrice = Math.Round(((this.RedundantShippingPrice * BusinessObjects.Common.VISTACARTHelper.AppDefaults.Tax.RatePercent / 100)), 2);
				}
			}

			// Totals
			// this.RendundantTotalWeight // we already have it
			this.RedundantTotalTaxPrice = this.RedundantProductsTaxPrice + this.RedundantShippingTaxPrice;
			this.RedundantTotalOrderPrice = this.RedundantProductsSalePrice + this.RedundantShippingPrice;
			this.RedundantTotalToPay = this.RedundantTotalTaxPrice + this.RedundantTotalOrderPrice;

			return true;
		}

		/// <summary>
		/// <para>true: if products were paid, they has to be removed from stock.</para>
		/// <para>false: restore products back to stock (egg. if payment was rejected, or payment was refunded)</para>
		/// </summary>
		/// <returns></returns>
		public bool AdjustStockQuantity(bool removeFromStock)
		{
			if (this.IsProductsRemovedFromStock == false && removeFromStock == true) // remove from stock
			{
				this.IsProductsRemovedFromStock = true;
				foreach (OrderDetail detail in this.GetOrderDetailCollection())
				{
					detail.GetProduct().AdjustStockQuantity(-detail.Quantity); // remove products from stock
				}
			}
			else if (this.IsProductsRemovedFromStock == true && removeFromStock == false) //restore to stock
			{
				this.IsProductsRemovedFromStock = false;
				foreach (OrderDetail detail in this.GetOrderDetailCollection())
				{
					detail.GetProduct().AdjustStockQuantity(detail.Quantity); // add products to stock, restore product back to stock if payment was refunded
				}
			}
			return this.SaveToDBInternal(false);
		}

		private ShippingType mShippingType = null;
		public ShippingType GetShippingType()
		{
			if (mShippingType == null)
			{
				mShippingType = new ShippingType();
				mShippingType.LoadFromDB(this.ShippingTypeID);
			}
			return mShippingType;
		}

		private PaymentType mPaymentType = null;
		public PaymentType GetPaymentType()
		{
			if (mPaymentType == null)
			{
				mPaymentType = new PaymentType();
				mPaymentType.LoadFromDB(this.PaymentTypeID);
			}
			return mPaymentType;
		}

		private User mUser = null;

		public User GetUser()
		{
			if (mUser == null || mUser.UserID != this.UserID)
			{
				mUser = new User();
				mUser.LoadFromDB(this.UserID);
			}
			return mUser;
		}

		public string Username
		{
			get { return GetUser().Username; }
		}

		public string Email
		{
			get { return GetUser().Email;}
		}

		

	}

}
