﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using System.Data.SqlClient;
using Ice.Orm.Core.Entity; using Ice.Orm.Core.Criteria;
using Ice.Net.Ecommerce.CacheProvider;

namespace Ice.Net.Ecommerce.BusinessObjects.Database.VISTACART
{
	public partial class ShoppingCart
	{
		#region CookieDetail class

		// due to cookie size limitation, we'll save only int productID, int quantity, string attributes values to cookie
		class CookieDetail
		{
			public int? ProductID = 0;
			public int? Quantity = 0;
			public string Attributes = "";

			public CookieDetail()
			{
			}

			public CookieDetail(int? productID, int? quantity, string attributes)
			{
				this.ProductID = productID;
				this.Quantity = quantity;
				this.Attributes = attributes;
			}
		}

		class CookieCartData
		{
			public List<CookieDetail> CookieDetailCollection = new List<CookieDetail>();
			public int? ShippingTypeID;
			public int? PaymentTypeID;
		}

		#endregion

		public Order Order = null;

		public ShoppingCart()
		{
			this.Order = new Order();
			this.Order.CreationDate = DateTime.Now;
		}

		public ShoppingCart(Order order)
		{
			this.Order = order;
		}

		public OrderDetailCollection OrderDetailCollection
		{
			get 
			{
				return this.Order.GetOrderDetailCollection();
			}
			set 
			{ 
				this.Order.SetOrderDetailCollection(value); 
			}
		}

		#region Database methods
		// Database methods

		#endregion

		#region Load & Save Cart methods

		/// <summary>
		/// Get ShoppingCart from database if user is authenticated or from cookie if not authenticated.
		/// </summary>
		/// <returns></returns>
		public static ShoppingCart GetCart()
		{
			if (BusinessObjects.Common.VISTACARTHelper.CurrentUser.UserTicket.IsAuthenticated)
			{
				// get shopping cart from database
				return GetCartFromDatabase(BusinessObjects.Common.VISTACARTHelper.CurrentUser.UserTicket.UserID, true);
			}
			else
			{
				// get shopping cart from cookie
				return GetCartFromCookie();
			}
		}

		/// <summary>
		/// Recalculate order prices and save ShoppingCart to database if user is authenticated or to cookie if not authenticated.
		/// </summary>
		/// <returns></returns>
		public bool SaveCartWithRecalculate()
		{
			this.Order.RecalculateRedundantValues();
			this.Order.IsActive = true;

			if (BusinessObjects.Common.VISTACARTHelper.CurrentUser.UserTicket.IsAuthenticated)
			{
				// save to database
				return SaveCartToDatabase(BusinessObjects.Common.VISTACARTHelper.CurrentUser.UserTicket.UserID);
			}
			else
			{
				// save to cookie
				return SaveCartToCookie();
			}
		}

		private static void SetDefaults(Order order)
		{
			if (order.PaymentTypeID < 1)
			{
				order.PaymentTypeID = BusinessObjects.Common.VISTACARTHelper.AppDefaults.AppGlobalSettingsItem.PaymentTypeID;
			}
			if (order.ShippingTypeID < 1)
			{
				order.ShippingTypeID = BusinessObjects.Common.VISTACARTHelper.AppDefaults.AppGlobalSettingsItem.ShippingTypeID;
			}
		}

		private static ShoppingCart GetCartFromDatabase(int userID, bool setDefaults)
		{
			Order dbCart = Order.LoadByUserIDAndOrderStatusEnum(userID, Order.OrderStatusEnum.ShoppingCart);
			
			if (dbCart == null)
			{
				// if object doesn't already exist in database, we have to set UserID and OrderStatus
				dbCart = new Order();
				dbCart.UserID = userID;
				dbCart.OrderStatus = Order.OrderStatusEnum.ShoppingCart;
			}
			if (setDefaults)
			{
				SetDefaults(dbCart);
			}
			return new ShoppingCart(dbCart);
		}

		private static ShoppingCart GetCartFromCookie()
		{
			ShoppingCart shoppCart = new ShoppingCart();

			// due to JavascriptSerializer deserialize limitations, we serialize List<CookieDetail> only!!!
			CookieCartData cookieCart = CookieStore.Retrieve<CookieCartData>();
			if (cookieCart == null)
			{
				return shoppCart;
			}
			OrderDetailCollection orderDetails = new OrderDetailCollection();
			shoppCart.Order.ShippingTypeID = cookieCart.ShippingTypeID;
			shoppCart.Order.PaymentTypeID = cookieCart.PaymentTypeID;

			if (cookieCart != null)
			{
				foreach (CookieDetail cookieDetail in cookieCart.CookieDetailCollection)
				{
					orderDetails.Add(new OrderDetail(cookieDetail.ProductID, cookieDetail.Quantity, cookieDetail.Attributes));
				}
			}

			SetDefaults(shoppCart.Order);

			shoppCart.Order.SetOrderDetailCollection(orderDetails);
			shoppCart.Order.RecalculateRedundantValues();
			return shoppCart;
		}

		private bool SaveCartToDatabase(int userID)
		{
			if (this.Order.CreationDate == null)
			{
				this.Order.CreationDate = DateTime.Now;
			}
			this.Order.IsActive = true;
			this.Order.UserID = userID;
			this.Order.OrderStatus = Order.OrderStatusEnum.ShoppingCart;
			return this.Order.SaveToDBInternal(true);
		}

		private bool SaveCartToCookie()
		{
			// due to JavascriptSerializer deserialize limitations, we serialize List<CookieDetail> only!!!
			CookieCartData cookieCart = new CookieCartData();
			cookieCart.ShippingTypeID = this.Order.ShippingTypeID;
			cookieCart.PaymentTypeID = this.Order.PaymentTypeID;

			foreach (OrderDetail detail in this.OrderDetailCollection)
			{
				cookieCart.CookieDetailCollection.Add(new CookieDetail(detail.ProductID, detail.Quantity, detail.Attributes));
			}
			return CookieStore.Save(cookieCart, DateTime.Now.AddYears(100));
		}


		private static bool DeleteCartFromCookie()
		{
			// due to JavascriptSerializer deserialize limitations, we serialize OrderDetailCollection only!!!
			return CookieStore.Delete(typeof(CookieCartData));
		}

		private int GetAddressCopyID(int? addressID)
		{
			Address address = new Address();
			address.LoadFromDB(addressID);

			// create address copy
			address.AddressID = 0;
			
			address.SaveToDB(WriteModeEnum.SaveAllWithIDCheck); // address copy!!!
			{
				// success
				return address.AddressID;
			}
		}

		/// <summary>
		/// <para>If not already created, copy User's Address 2 times for Billing and Shipping address and assing it's AddressID's to Order.BillingAddressID and Order.ShippingAddressID</para>
		/// <para>If changes were made, Order object is saved to database</para>
		/// </summary>
		public void CreateBillingAndShippingAddresses()
		{
			if (BusinessObjects.Common.VISTACARTHelper.CurrentUser.UserTicket.IsAuthenticated)
			{
				User user = new User();
				if (user.LoadFromDB(BusinessObjects.Common.VISTACARTHelper.CurrentUser.UserTicket.UserID))
				{
					ShoppingCart dbCart = this;
					bool saveOrder = false;

					// create 2 copies of current user address, 1 for billing address, 1 for shipping address
					// and assign them to current ShoppingCart
					if (dbCart.Order.BillingAddressID < 1)
					{
						saveOrder = true;
						dbCart.Order.BillingAddressID = GetAddressCopyID(user.AddressID);
					}
					if (dbCart.Order.ShippingAddressID < 1)
					{
						saveOrder = true;
						dbCart.Order.ShippingAddressID = GetAddressCopyID(user.AddressID);
					}

					if (saveOrder)
					{
						dbCart.Order.SaveToDB(WriteModeEnum.SaveAllWithIDCheck);
					}
				}
				else
				{
					BusinessObjects.Common.Logging.WriteError("Can't load User object, UserID=" + BusinessObjects.Common.VISTACARTHelper.CurrentUser.UserTicket.UserID);
				}
			}
		}

		public static bool MoveCookieCartToDatabase(int userID)
		{
			// when user log-in, we have to merge he's Cookie Shopping Cart with Database Shopping Cart and delte Cookie Shopping Cart
			// get cookie cart
			ShoppingCart cookieCart = GetCartFromCookie();
			// get cart from database
			ShoppingCart dbCart = GetCartFromDatabase(userID, false); // don't set default shippingID and paymentTypeID // we'll copy them from cookie cart if database cart is just created
			dbCart.Order.IsActive = true;

			foreach (OrderDetail detail in cookieCart.OrderDetailCollection)
			{
				dbCart.AddProduct(detail.ProductID, detail.Quantity, detail.Attributes);
			}
	
			if (dbCart.Order.ShippingAddressID < 1)
			{
				dbCart.Order.ShippingTypeID = cookieCart.Order.ShippingTypeID;
			}
			if (dbCart.Order.PaymentTypeID < 1)
			{
				dbCart.Order.PaymentTypeID = cookieCart.Order.PaymentTypeID;
			}

			dbCart.Order.RecalculateRedundantValues();
			// save database cart and delte cookie cart
			return (DeleteCartFromCookie() && dbCart.Order.SaveToDBInternal(true));
		}

		/// <summary>
		/// Delete shopping cart from cookie and from database by userID
		/// </summary>
		/// <param name="userID"></param>
		/// <returns></returns>
		public bool DeleteCart(int userID)
		{
			ShoppingCart cart = GetCartFromDatabase(userID, false);
			return DeleteCartFromCookie() && cart.Order.DeleteFromDB();
		}

		#endregion

		#region Add / Remove products methods

		public int AddProduct(int? productID, int? quantity, string attributes)
		{
			int count = OrderDetailCollection.AddProduct(productID, quantity, attributes);
			return count;
		}

		public int SetQuantity(int? detailIndex, int? quantity)
		{
			OrderDetailCollection.SetQuantity(detailIndex, quantity);
			return quantity.GetValueOrDefault();
		}

		public int RemoveAtDetailIndex(int detailIndex)
		{
			OrderDetailCollection.RemoveAtDetailIndex(detailIndex);
			return 0;
		}

		#endregion

		/// <summary>
		/// Lock shopping cart - disable cart items modifications any more. SaveToDB is invoked automatically.
		/// </summary>
		public bool CloseShoppingCart()
		{
			this.Order.OrderStatus = Order.OrderStatusEnum.Ordered;
			this.Order.CheckoutDate = DateTime.Now;
			this.Order.IPAddress = System.Web.HttpContext.Current.Request.UserHostAddress;
			return this.Order.SaveToDBInternal(false);
		}

	}

}
