﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using Ice.Xml.Parser;
using System.Web;
using Ice.Net.Ecommerce.WebService.Code;
using System.Linq;

namespace Ice.Net.Ecommerce.CurrencyUpdater
{
	#region CurrencyRate class

	public class CurrencyRate
	{
		/// <summary>
		/// Currency name in English.
		/// </summary>
		public string Name;
		
		/// <summary>
		/// Currency ISO code.
		/// </summary>
		public string ISOCode;

		/// <summary>
		/// Conversion rate. 1 unit of default currency expressed in this currency.
		/// </summary>
		public decimal ConversionRate;

		/// <summary>
		/// Reverse conversion rate. 1 unit of this currency expressed in default currency.
		/// </summary>
		public decimal ReverseRate;

		public CurrencyRate()
		{ }

		public CurrencyRate(string name, string isoCode, decimal conversionRate, decimal reverseRate)
		{
			this.Name = name;
			this.ISOCode = isoCode;
			this.ConversionRate = conversionRate;
			this.ReverseRate = reverseRate;
		}
	}

	#endregion

	#region CurrencyRatesInfo class

	public class CurrencyRatesInfo
	{

		/// <summary>
		/// Last update from currency server date and time.
		/// </summary>
		public DateTime Date = DateTime.UtcNow;

		/// <summary>
		/// True if Items data is valid, else false.
		/// </summary>
		public bool Success = false;

		/// <summary>
		/// Error message if there was any error.
		/// </summary>
		public string ErrorMessage = "";

		/// <summary>
		/// CurrencyRate collection.
		/// </summary>
		public List<CurrencyRate> Items = new List<CurrencyRate>();

	}

	#endregion

	#region CurrencyRatesUpdaterResult class

	public class CurrencyRatesUpdaterResult
	{
		public bool Success = false;
		public int Count = 0;
		public string Message = "";

		public CurrencyRatesUpdaterResult()
		{ }

		public CurrencyRatesUpdaterResult(bool success, int count, string message)
		{
			this.Success = success;
			this.Count = count;
			this.Message = message;
		}
	}

	#endregion

	#region CurrencyRatesUpdater class

	public class CurrencyRatesUpdater
	{
		public static CurrencyRatesInfo GetCurrencyRatesInfo(string defaultCurrencyISOCode, string remoteUrl, string userTicket)
		{
			CurrencyRatesUpdater updater = new CurrencyRatesUpdater();
			return updater.GetCurrencyRatesInfoWorker(defaultCurrencyISOCode, remoteUrl);
		}

		/// <summary>
		/// Returns number of objects updated
		/// </summary>
		/// <returns></returns>
		private CurrencyRatesInfo GetCurrencyRatesInfoWorker(string defaultCurrencyISOCode, string remoteUrl)
		{
			string logMsg = "CurrencyRatesInfo, request from IP: " + HttpContext.Current.Request.UserHostAddress + ", ISO code: " + defaultCurrencyISOCode + ", remote URL: " + remoteUrl;

			/*
		basecur	USD
		day	17
		hide_inverse	true
		historical	false
		image.x	34
		image.y	14
		month	8
		sort_by	name
		year	2008
		*/
			if (defaultCurrencyISOCode == null || defaultCurrencyISOCode.Length < 3)
			{
				defaultCurrencyISOCode = "USD";
			}

			CurrencyRatesInfo currencyRatesInfo = null;			
			// get from cache first
			Object oCached = System.Web.HttpContext.Current.Cache[defaultCurrencyISOCode];
			if (oCached != null && oCached.GetType() == typeof(CurrencyRatesInfo))
			{
				currencyRatesInfo = (CurrencyRatesInfo)oCached;
				logMsg = "Cached! " + logMsg;
			}
			else
			{
				currencyRatesInfo = new CurrencyRatesInfo();

				// get currency conversion html table at: http://www.xe.com/ict/
				string parameters = String.Format("day={0}&month={1}&year={2}&basecur={3}&hide_inverse=false&historical=false&sort_by=name&image.x=34&image.y=14",
					DateTime.UtcNow.Day,
					DateTime.UtcNow.Month,
					DateTime.UtcNow.Year,
					defaultCurrencyISOCode.ToUpper());

				HttpResponseStatus status = WebService.Code.HttpRequest.MakePostRequest("http://www.xe.com/ict/", parameters);
				if (status.StatusCode != System.Net.HttpStatusCode.OK)
				{
					currencyRatesInfo.ErrorMessage = "Currency server is temporary down. Please, try again in few minutes or send us email: support@iceasp.net.";
					//Logging.WriteError(logMsg + ". ERROR: " + currencyRatesInfo.ErrorMessage);
					return currencyRatesInfo;
				}

				// get table with class="tbl_main printFullWidth"
				// parse it using next code:
				XmlDocument doc = XmlDocument.Parse(status.ResponseText);

				List<XmlNode> nodes = doc.DescendantNodes().WhereAttribute("class", "tbl_sub").ParentNodes().ToList();
				List<XmlNode> trs = nodes.DescendantNodes(o => o.TagName == "tr" && (o.Class == "row1" || o.Class == "row2")).ToList();

				foreach (XmlNode tr in trs)
				{
					List<XmlNode> tds = tr.DescendantNodes(o => o.TagName == "td").ToList();

					Regex re2 = new Regex("(^[^\\w\\d]+)|([^\\w\\d]+$)");

					string decimalSeparator = ".";
					string iso = null, name = null, rate = null, reverseRate = null;
					try
					{
						decimalSeparator = System.Globalization.CultureInfo.CurrentCulture.NumberFormat.CurrencyDecimalSeparator;

						iso = re2.Replace(tds[0].InnerText, "");
						name = re2.Replace(tds[1].InnerText, "");
						rate = re2.Replace(tds[2].InnerText, "").Replace(".", decimalSeparator);
						reverseRate = re2.Replace(tds[3].InnerText, "").Replace(".", decimalSeparator);

						decimal drate, dreverseRate;

						if (decimal.TryParse(rate, out drate) && decimal.TryParse(reverseRate, out dreverseRate))
						{
							CurrencyRate currRate = new CurrencyRate(name, iso, drate, dreverseRate);
							currencyRatesInfo.Items.Add(currRate);
							currencyRatesInfo.Success = true;
						}

					}
					catch// (Exception ex)
					{
						//Logging.WriteError(logMsg + ", ERROR: " + ex.Message);
					}
				}

				if (currencyRatesInfo.Success && currencyRatesInfo.Items.Count > 0)
				{
					// save to cache
					System.Web.HttpContext.Current.Cache.Insert(defaultCurrencyISOCode, currencyRatesInfo, null, DateTime.UtcNow.AddMinutes(10), System.Web.Caching.Cache.NoSlidingExpiration);
				}

			}

			//Logging.WriteApplication(logMsg + ", Items: " + currencyRatesInfo.Items.Count);

			return currencyRatesInfo;
		}
	}

	#endregion

}
