﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;

namespace Ice.Net.Ecommerce
{
	public static class EnumerableExtensions
	{
		/// <summary>
		/// Join string elements of List collection
		/// </summary>
		/// <typeparam name="?"></typeparam>
		/// <param name="s"></param>
		/// <param name="separator"></param>
		/// <returns></returns>
		public static string ToStringJoined<TSource>(this IEnumerable<TSource> s, string separator = null)
		{
			StringBuilder sb = new StringBuilder();
			bool hasElements = false;
			foreach (TSource item in s)
			{
				if (hasElements)
				{
					sb.Append(separator);
				}
				else
				{
					hasElements = true;
				}
				sb.Append(item);
			}
			return sb.ToString();
		}

		/// <summary>
		/// Concatenate single item
		/// </summary>
		/// <typeparam name="T"></typeparam>
		/// <param name="ienum"></param>
		/// <param name="item"></param>
		/// <returns></returns>
		public static IEnumerable<TSource> Concat<TSource>(this IEnumerable<TSource> source, TSource item)
		{
			foreach (TSource m in source)
			{
				yield return m;
			}

			yield return item;
		}

		/// <summary>
		/// returns only successfuly casted values
		/// </summary>
		/// <typeparam name="TResult"></typeparam>
		/// <param name="ienum"></param>
		/// <returns></returns>
		public static IEnumerable<TResult> TryCast<TResult>(this IEnumerable source)
		{
			foreach (object obj in source)
			{
				if (obj is TResult)
				{
					yield return (TResult)obj;
				}
			}
		}

		/// <summary>
		/// Returns the first element of a sequence, or a defaultValue if the sequence contains no elements.
		/// </summary>
		public static TSource FirstOrValue<TSource>(this IEnumerable<TSource> source, TSource defaultValue = default(TSource))
		{
			source = source.Take(1);
			if (source.Count() == 1)
			{
				return source.FirstOrDefault();
			}
			else
			{
				return defaultValue;
			}
		}

		/// <summary>
		/// Returns the last element of a sequence, or a defaultValue if the sequence contains no elements.
		/// </summary>
		public static TSource LastOrValue<TSource>(this IEnumerable<TSource> source, TSource defaultValue = default(TSource))
		{
			source = source.Reverse().Take(1);

			if (source.Count() == 1)
			{
				return source.FirstOrDefault();
			}
			else
			{
				return defaultValue;
			}
		}

		/// <summary>
		/// Returns value or default value
		/// </summary>
		/// <typeparam name="TKey"></typeparam>
		/// <typeparam name="TValue"></typeparam>
		/// <param name="source"></param>
		/// <param name="key"></param>
		/// <returns></returns>
		public static TValue TryGetValueOrDefault<TKey, TValue>(this IDictionary<TKey, TValue> source, TKey key)
		{
			TValue val;
			if (source.TryGetValue(key, out val))
			{
				return val;
			}
			else
			{
				return default(TValue);
			}
		}

		/// <summary>
		/// Creates new group from all items in source
		/// </summary>
		/// <typeparam name="TSource"></typeparam>
		/// <param name="source"></param>
		/// <returns></returns>
		public static List<TSource[]> ToGroup<TSource>(this IEnumerable<TSource> source)
		{
			return new List<TSource[]>() { source.ToArray() };
		}

		public static IEnumerable<TSource> Case<TSource>(this IEnumerable<TSource> source, bool condition, Func<IEnumerable<TSource>, IEnumerable<TSource>> onTrue, Func<IEnumerable<TSource>, IEnumerable<TSource>> onFalse = null)
		{
			if (condition)
			{
				return onTrue != null ? onTrue(source) : source;
			}
			else
			{
				return onFalse != null ? onFalse(source) : source;
			}
		}

		public static IEnumerable<TSource> Distinct<TSource>(this IEnumerable<TSource> source, Func<TSource, object> keySelector)
		{
			List<object> keys = new List<object>();

			foreach (TSource item in source)
			{
				object key = keySelector(item);

				if (!keys.Contains(key))
				{
					keys.Add(key);
					yield return item;
				}
			}
		}

		public static IOrderedEnumerable<TSource> OrderBy<TSource, TKey>(this IEnumerable<TSource> source, Func<TSource, TKey> keySelector, bool descending)
		{
			if (descending)
			{
				return source.OrderByDescending(keySelector);
			}
			else
			{
				return source.OrderBy(keySelector);
			}
		}

		public static IEnumerable<TSource> Cycle<TSource>(this IEnumerable<TSource> source, int take)
		{
			int count = 0;

			while (true)
			{
				foreach (TSource item in source)
				{
					if (count++ < take)
					{
						yield return item;
					}

					if(count >= take)
					{
						break;
					}
				}

				if (count == 0 || count >= take)
				{
					// source has no items or all taken
					break;
				}
			}
		}



	}

}
