﻿/***********************************************
 * Modified by Ice to return Sequence
 * use SQLServer_create_table.sql to create db table
 * removed parameter Application from WHERE
 ***********************************************/

[assembly: Ice.Elmah.Scc("$Id: SqlErrorLog.cs 723 2010-09-26 20:38:24Z azizatif $")]

namespace Ice.Elmah
{
    #region Imports

    using System;
    using System.Data;
    using System.Data.SqlClient;
    using System.IO;
    using System.Text;
    using System.Threading;
    using System.Xml;

    using IDictionary = System.Collections.IDictionary;
    using IList = System.Collections.IList;

    #endregion

    /// <summary>
    /// An <see cref="ErrorLog"/> implementation that uses Microsoft SQL 
    /// Server 2000 as its backing store.
    /// </summary>
    
    public class SqlErrorLog : ErrorLog
    {
        private readonly string _connectionString;

        private const int _maxAppNameLength = 60;

        /// <summary>
        /// Initializes a new instance of the <see cref="SqlErrorLog"/> class
        /// using a dictionary of configured settings.
        /// </summary>
        public SqlErrorLog(IDictionary config)
        {
            if (config == null)
                throw new ArgumentNullException("config");

            string connectionString = ConnectionStringHelper.GetConnectionString(config);

            //
            // If there is no connection string to use then throw an 
            // exception to abort construction.
            //

            if (connectionString.Length == 0)
                throw new ApplicationException("Connection string is missing for the SQL error log.");

            _connectionString = connectionString;

			//// InitializeDatabase();

            //
            // Set the application name as this implementation provides
            // per-application isolation over a single store.
            //

            string appName = Mask.NullString((string) config["applicationName"]);

            if (appName.Length > _maxAppNameLength)
            {
                throw new ApplicationException(string.Format(
                    "Application name is too long. Maximum length allowed is {0} characters.",
                    _maxAppNameLength.ToString("N0")));
            }

            ApplicationName = appName;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="SqlErrorLog"/> class
        /// to use a specific connection string for connecting to the database.
        /// </summary>
        public SqlErrorLog(string connectionString)
        {
            if (connectionString == null)
                throw new ArgumentNullException("connectionString");

            if (connectionString.Length == 0)
                throw new ArgumentException(null, "connectionString");
            
            _connectionString = connectionString;
        }

        /// <summary>
        /// Gets the name of this error log implementation.
        /// </summary>
        public override string Name
        {
            get { return "Microsoft SQL Server Error Log"; }
        }

        /// <summary>
        /// Gets the connection string used by the log to connect to the database.
        /// </summary>
        public virtual string ConnectionString
        {
            get { return _connectionString; }
        }

		/// <summary>
        /// Logs an error to the database.
        /// </summary>
        /// <remarks>
        /// Use the stored procedure called by this implementation to set a
        /// policy on how long errors are kept in the log. The default
        /// implementation stores all errors for an indefinite time.
        /// </remarks>
        public override string Log(Error error)
        {
            if (error == null)
                throw new ArgumentNullException("error");

            string errorXml = ErrorXml.EncodeString(error);

            Guid id = Guid.NewGuid();

			using (SqlConnection connection = new SqlConnection(this.ConnectionString))
			{
				using (SqlCommand command = new SqlCommand())
				{
					command.Connection = connection;

					command.CommandText = @"INSERT INTO ELMAH_Error 
([ErrorId], [Application], [Host], [Type], [Source], [Message], [User], [StatusCode], [TimeUtc], [AllXml] )
	VALUES 
(@ErrorId, @Application, @Host, @Type, @Source, @Message, @User, @StatusCode, @TimeUtc, @AllXml);";

					SqlParameterCollection parameters = command.Parameters;

					parameters.Add("@ErrorId", SqlDbType.UniqueIdentifier).Value = id;
					parameters.Add("@Application", SqlDbType.NVarChar, _maxAppNameLength).Value = ApplicationName;
					parameters.Add("@Host", SqlDbType.NVarChar, 30).Value = error.HostName;
					parameters.Add("@Type", SqlDbType.NVarChar, 100).Value = error.Type;
					parameters.Add("@Source", SqlDbType.NVarChar, 60).Value = error.Source;
					parameters.Add("@Message", SqlDbType.NVarChar, 500).Value = error.Message;
					parameters.Add("@User", SqlDbType.NVarChar, 50).Value = error.User;
					parameters.Add("@StatusCode", SqlDbType.Int).Value = error.StatusCode;
					parameters.Add("@TimeUtc", SqlDbType.DateTime).Value = error.Time.ToUniversalTime();
					parameters.Add("@AllXml", SqlDbType.NText).Value = errorXml;

					connection.Open();
					command.ExecuteNonQuery();
					return id.ToString();

				}
			}
        }

        /// <summary>
        /// Returns a page of errors from the databse in descending order 
        /// of logged time.
        /// </summary>
        public override int GetErrors(int pageIndex, int pageSize, IList errorEntryList)
        {
            if (pageIndex < 0)
                throw new ArgumentOutOfRangeException("pageIndex", pageIndex, null);

            if (pageSize < 0)
                throw new ArgumentOutOfRangeException("pageSize", pageSize, null);

			using (SqlConnection connection = new SqlConnection(this.ConnectionString))
			{
				using (SqlCommand command = new SqlCommand())
				{
					command.Connection = connection;

					int first = pageIndex * pageSize + 1; // 1 is first in SQL server
					int last = (pageIndex + 1) * pageSize;

					command.CommandText =  string.Format(@"select * from (
select ROW_NUMBER() OVER (ORDER BY Sequence DESC) as rowNum1, [Sequence], [ErrorId], [Application], [Host], [Type], [Source], [Message], [User], [StatusCode], [TimeUtc] from ELMAH_Error
) as tmp1
where tmp1.rowNum1 between {0} and {1}",
				first,
				last);

					connection.Open();

					using (SqlDataReader reader = command.ExecuteReader())
                    {
                        if (errorEntryList != null)
                        {
                            while (reader.Read())
                            {
                                string id = reader["ErrorId"].ToString();

								Elmah.Error error = new Elmah.Error();
								error.Sequence = Convert.ToInt32(reader["Sequence"]); // Ice added
                                error.ApplicationName = reader["Application"].ToString();
                                error.HostName = reader["Host"].ToString();
                                error.Type = reader["Type"].ToString();
                                error.Source = reader["Source"].ToString();
                                error.Message = reader["Message"].ToString();
                                error.User = reader["User"].ToString();
                                error.StatusCode = Convert.ToInt32(reader["StatusCode"]);
                                error.Time = Convert.ToDateTime(reader["TimeUtc"]).ToLocalTime();
                                errorEntryList.Add(new ErrorLogEntry(this, id, error));
                            }
                        }
                    }
                }

                using (SqlCommand command = new SqlCommand())
                {
					command.CommandText = @"SELECT COUNT(*) FROM [ELMAH_Error]";
					command.Connection = connection;
                    return (int)command.ExecuteScalar();
                }
			}
        }

        /// <summary>
        /// Returns the specified error from the database, or null 
        /// if it does not exist.
        /// </summary>
        public override ErrorLogEntry GetError(string id)
        {
            if (id == null)
                throw new ArgumentNullException("id");

            if (id.Length == 0)
                throw new ArgumentException(null, "id");

            Guid errorGuid;

            try
            {
                errorGuid = new Guid(id);
            }
            catch (Exception e)
            {
                throw new ArgumentException(e.Message, "id", e);
            }

			using (SqlConnection connection = new SqlConnection(this.ConnectionString))
			{
				using (SqlCommand command = new SqlCommand())
				{
					command.Connection = connection;

					command.CommandText = @"SELECT [AllXml] FROM [ELMAH_Error] WHERE [ErrorId] = @ErrorId";

					command.Parameters.Add("@ErrorId", SqlDbType.UniqueIdentifier).Value = errorGuid;

					connection.Open();

					string errorXml = (string)command.ExecuteScalar();

					if (errorXml == null)
						return null;

					Error error = ErrorXml.DecodeString(errorXml);
					return new ErrorLogEntry(this, id, error);
				}
			}
        }

    }
}
