﻿using System;

namespace Ice.Orm.Core.Criteria
{
	using DataStore;

	/// <summary>
	/// Enumeration of the comparison operators supported in SQL statements.
	/// </summary>
	public enum DBWhereKeyOperator
	{
		Equals = 1,
		NotEquals = 2,
		
		/// <summary>
		/// must include % manually, % is not added by the program
		/// </summary>
		Like = 3,
		
		/// <summary>
		/// must include % manually, % is not added by the program
		/// </summary>
		NotLike = 4,

		LessThan = 5,
		LessThanOrEquals = 6,
		GreaterThan = 7,
		GreaterThanOrEquals = 8,
		In = 9,
		NotIn = 10,
		IsNULL = 11,
		IsNotNULL = 12,

		NullOrEmpty = 13,
		NotNullOrEmpty = 14,

		/// <summary>
		/// % is added by the program automatically: value%
		/// </summary>
		StartsWith = 15,

		/// <summary>
		/// % is added by the program automatically: %value
		/// </summary>
		EndsWith = 16,

		/// <summary>
		/// % is added by the program automatically: %value%
		/// </summary>
		Contains = 17,
	} ;

	public enum DBWhereKeyLogic
	{
		And,
		Or,
		NotUsed
	};

	public enum DBWhereBrackets
	{
		NotUsed,

		/// <summary>
		/// (
		/// </summary>
		Open,

		/// <summary>
		/// )
		/// </summary>
		Close,

		/// <summary>
		/// ((
		/// </summary>
		OpenDouble,

		/// <summary>
		/// ))
		/// </summary>
		CloseDouble
	}

	[Serializable]
	public class DBWhereKey : QParameter
	{
		private DBWhereKeyOperator mComparisonOperator = DBWhereKeyOperator.Equals;
		private DBWhereKeyLogic mLogicOperator = DBWhereKeyLogic.NotUsed;
		private DBWhereBrackets mBrackets = DBWhereBrackets.NotUsed;
		bool mReverseColumnNameAndValue = false; // reverse if true, egg.: value LIKE columnName

		/// <summary>
		/// Will set the default comparisonOperator 'Equals' and logicOperator 'And', example: ( ... AND [name] = value ... )
		/// </summary>
		public DBWhereKey(string columnName, 
			object value, 
			DBWhereKeyOperator comparisonOperator = DBWhereKeyOperator.Equals, 
			DBWhereKeyLogic logicOperator = DBWhereKeyLogic.And, 
			DBWhereBrackets brackets = DBWhereBrackets.NotUsed)
		{
			this.Column = columnName;
			this.Value = value;
			this.mBrackets = brackets;
			this.mComparisonOperator = comparisonOperator;
			this.mLogicOperator = logicOperator;

			if(!IsValueNull(this.Value))
			{
				this.Value = FixValueForLike(comparisonOperator, this.Value);
			}
		}

		private string GetSqlIfNull()
		{
			if (this.IsNullValue())
			{
				return "";
			}
			else
			{
				return this.Name;
			}
		}

		public static object FixValueForLike(DBWhereKeyOperator comparison, object value)
		{
			switch (comparison)
			{
				case DBWhereKeyOperator.Contains:
				case DBWhereKeyOperator.Like:
				case DBWhereKeyOperator.NotLike:
					return "%" + value + "%";

				case DBWhereKeyOperator.StartsWith:
					return value + "%";

				case DBWhereKeyOperator.EndsWith:
					return "%" + value;
			}
			return value;
		}

		public static bool IsValueNull(object value)
		{
			return value == null || value == System.DBNull.Value;
		}

		private bool IsNullValue()
		{
			return IsValueNull(this.Value);
		}

		public static string GetComparisonOperator(DBWhereKeyOperator comparison, bool isNullValue, ref bool noParameter)
		{
			switch (comparison)
			{
				case DBWhereKeyOperator.NullOrEmpty:
				case DBWhereKeyOperator.NotNullOrEmpty:
					throw new Exception("Can't create where for NullOrEmpty or NotNullOrEmpty. Use 2 steps!");
			}

			if (isNullValue)
			{
				switch (comparison)
				{
					case DBWhereKeyOperator.Equals:
						comparison = DBWhereKeyOperator.IsNULL;
						break;
					case DBWhereKeyOperator.NotEquals:
						comparison = DBWhereKeyOperator.IsNotNULL;
						break;
				}
			}

			switch (comparison)
			{
				case DBWhereKeyOperator.IsNotNULL:
					noParameter = true;
					return " IS NOT NULL ";
				case DBWhereKeyOperator.IsNULL:
					noParameter = true;
					return " IS NULL ";
				case DBWhereKeyOperator.Equals:
					return " = ";
				case DBWhereKeyOperator.NotEquals:
					return " <> "; // odbc ne podrzava !=
				case DBWhereKeyOperator.GreaterThan:
					return " > ";
				case DBWhereKeyOperator.GreaterThanOrEquals:
					return " >= ";
				case DBWhereKeyOperator.LessThan:
					return " < ";
				case DBWhereKeyOperator.LessThanOrEquals:
					return " <= ";
				case DBWhereKeyOperator.StartsWith:
				case DBWhereKeyOperator.EndsWith:
				case DBWhereKeyOperator.Contains:
				case DBWhereKeyOperator.Like:
					return " LIKE ";
				case DBWhereKeyOperator.NotLike:
					return " NOT LIKE ";
				case DBWhereKeyOperator.In:
					return " IN ";
				case DBWhereKeyOperator.NotIn:
					return " NOT IN ";

				default:
					return null;
			}
		}

		public string GetWhereSql()
		{
			string sql = null;
			bool noParameter = false;
			string comparison = GetComparisonOperator(this.mComparisonOperator, this.IsNullValue(), ref noParameter);

			if (comparison != null)
			{
				if (mReverseColumnNameAndValue)
				{
					sql = GetSqlIfNull() + comparison + SqlName.CreateFullSqlName(this.Column);
				}
				else
				{
					sql = SqlName.CreateFullSqlName(this.Column) + comparison + GetSqlIfNull();
				}
			}
			else
			{
				// build custom expression
				switch (this.mComparisonOperator)
				{
					case DBWhereKeyOperator.IsNULL:
						sql = SqlName.CreateFullSqlName(this.Column) + " IS NULL ";
						break;

					case DBWhereKeyOperator.IsNotNULL:
						sql = SqlName.CreateFullSqlName(this.Column) + " IS NOT NULL ";
						break;

					case DBWhereKeyOperator.In:
						sql = SqlName.CreateFullSqlName(this.Column) + " IN (" + this.GetInParameterNames() + ")";
						break;

					case DBWhereKeyOperator.NotIn:
						sql = SqlName.CreateFullSqlName(this.Column) + " NOT IN (" + this.GetInParameterNames() + ")";
						break;

					default:
						throw (new Exception("Not implemented"));
				}
			}

			return this.GetBrackets(true) +  sql + this.GetBrackets(false);
		}

		private string GetBrackets(bool opening)
		{
			return GetBrackets(mBrackets, opening);
		}

		public static string GetBrackets(DBWhereBrackets brackets, bool opening)
		{
			if (opening == true)
			{
				if (brackets == DBWhereBrackets.Open)
				{
					return "(";
				}
				else if (brackets == DBWhereBrackets.OpenDouble)
				{
					return "((";
				}
			}
			else if (opening == false)
			{
				if (brackets == DBWhereBrackets.Close)
				{
					return ")";
				}
				else if (brackets == DBWhereBrackets.CloseDouble)
				{
					return "))";
				}
			}

			return "";
		}

		public string GetLogicOperatorString()
		{
			return GetLogicOperatorString(this.mLogicOperator);
		}

		private static string GetLogicOperatorString(DBWhereKeyLogic logic)
		{
			switch (logic)
			{
				case DBWhereKeyLogic.And:
					return " AND ";
				case DBWhereKeyLogic.Or:
					return " OR ";
				default:
					return "";
			}
		}

		private string GetInParameterNames()
		{
			if (this.Value is Array)
			{
				int inCount = 0;
				System.Text.StringBuilder sb = new System.Text.StringBuilder();

				foreach (object o in (this.Value as Array))
				{
					if (sb.Length > 0)
					{
						sb.Append(",");
					}
					sb.Append(this.Name);
					inCount++;
				}
				return sb.ToString();
			}
			return null;
		}

	}
}
