﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.Linq.Expressions;
using Ice.Orm.Core.LINQ;
using System.Linq;
using Ice.Orm.Core.DataStore;
using System.IO;
using Ice.Orm.Core.Entity;
using Ice.Orm.Core.Schema;
using Ice.Orm.Core.Criteria.SqlBlocks;
using System.Data;

namespace Ice.Orm.Core.Criteria
{
	[System.Diagnostics.DebuggerDisplay("{_query.QueryWithValues}")]
	public class Query<TSource> : ICompilableQuery
		where TSource: EntityBase
	{
#if DEBUG
		static Query()
		{
			LINQ.SelfTest.QueryLinqSelfTests.RunSelfTest();
		}
#endif

		QueryBase _query;

		public Query()
		{
			_query = new QueryBase();
			EntitySetting es = EntitySetting.Get<TSource>();
			_query.DataStore = es.DataStore;
			_query.SchemaTableHelper = es.SchemaTableHelper; // koristi se u CompiledQuery pri on-demand kreirianju ReadOnlyCollection<IDbDataParameter> DbParameters
		}

		public Query<TSource> CreateQuery(string sql, params object[] args)
		{
			SqlCommandBlock cmd = _query.GetSqlBlock<SqlCommandBlock>();
			cmd.AddTextFormated(sql, args);
			return this;
		}

		public Query<TSource> Skip(int skip)
		{
			_query.SkipCount = skip;
			return this;
		}

		public Query<TSource> Take(int count)
		{
			_query.TakeCount = count;
			return this;
		}

		/// <summary>
		/// Add DISTINCT to sql
		/// </summary>
		/// <returns></returns>
		public Query<TSource> Distinct()
		{
			_query.IsDistinct = true;
			return this;
		}

		public Query<TSource> Where(Expression<Func<TSource, bool>> predicate)
		{
			if (predicate != null)
			{
				WhereSqlFormater ff = new WhereSqlFormater(_query); // write parameters and sql to query
				ff.Visit(predicate);
				SqlWhereBlock where = _query.GetSqlBlock<SqlWhereBlock>();
				where.AddSegments(ff.SqlSegments.AsEnumerable(), false);
			}
			return this;
		}

		public Query<TSource> Where(string whereSql, params object[] args)
		{
			if (!string.IsNullOrEmpty(whereSql))
			{
				SqlCommandBlock cmd = new SqlCommandBlock();
				cmd.Add("(");
				cmd.AddTextFormated(whereSql, args); // obavezno kreirati zagrade okolo jer je ovo zasebni Where(...). ovo je bio bug kad nije bilo zagrada
				cmd.Add(")");
				_query.GetSqlBlock<SqlWhereBlock>().AddSegments(cmd.SelectSqlSegments(), false);
			}
			return this;
		}

		/// <summary>
		/// Create SELECT query.
		/// </summary>
		/// <typeparam name="TSource"></typeparam>
		/// <typeparam name="TResult"></typeparam>
		/// <param name="source"></param>
		/// <param name="selector"></param>
		/// <returns></returns>
		public QuerySelect<TSource, TResult> Select<TResult>(Expression<Func<TSource, TResult>> selector)
		{
			this.CreateSelectSql(
				null, 
				selector, // selector can be null, k=>k or any other expression
				null);
			return new QuerySelect<TSource, TResult>(_query);
		}

		protected void CreateSelectSql(string before, Expression selector, string after)
		{
			SqlSelectBlock sel = _query.GetSqlBlock<SqlSelectBlock>();
			if (selector == null)
			{
				sel.AddSegment(new SqlSegment("*"));
			}
			else
			{
				SelectorSqlFormater frm = new SelectorSqlFormater(_query);
				frm.Visit(selector);
				sel.AddSegment(new SqlSegment(before));
				sel.AddSegments(frm.SqlSegments.AsEnumerable());
				sel.AddSegment(new SqlSegment(after));
			}
			_query.WriteToFromBlock(GetTableName<TSource>());
		}

		public TResult Max<TResult>(Expression<Func<TSource, TResult>> selector)
		{
			this.CreateSelectSql("MAX(", selector, ")");
			return _query.ExecuteScalar<TSource, TResult>();
		}

		public TResult Min<TResult>(Expression<Func<TSource, TResult>> selector)
		{
			this.CreateSelectSql("MIN(", selector, ")");
			return _query.ExecuteScalar<TSource, TResult>();
		}

		public TResult Sum<TResult>(Expression<Func<TSource, TResult>> selector)
		{
			this.CreateSelectSql("SUM(", selector, ")");
			return _query.ExecuteScalar<TSource, TResult>();
		}

		public Query<TSource> OrderBy<TKey>(Expression<Func<TSource, TKey>> selector, bool descending = false)
		{
			if (typeof(TKey).Name.Contains("AnonymousType"))
			{
				throw new ArgumentException("AnonymousType is not supported with OrderBy(): " + selector);
			}
			OrderBySqlFormater frm = new OrderBySqlFormater(_query);
			frm.Visit(selector);
			SqlOrderByBlock order = _query.GetSqlBlock<SqlOrderByBlock>();
			string col = frm.SqlSegments.AsEnumerable().Select(o=>o.Text).ToStringJoined();
			order.AddColumn(col, descending); // there should be only one column name in items
			return this;
		}

		public Query<TSource> OrderBy(string customSql)
		{
			_query.GetSqlBlock<SqlOrderByBlock>().AddText(customSql);
			return this;
		}

		public Query<TSource> OrderByDescending<TKey>(Expression<Func<TSource, TKey>> selector)
		{
			return this.OrderBy(selector, true);
		}

		private static SqlName GetTableName(Type sourceType)
		{
			EntitySetting es = AttributedTableHelper.GetEntitySetting(sourceType);
			if (es != null)
			{
				return new SqlName(es.SchemaTableHelper.TableName);
			}
			else
			{
				throw new NotSupportedException("\"" + sourceType.Name + "\" can't be translated into T-SQL, there is no EntitySetting attribute defined");
			}
		}

		private static SqlName GetTableName<TSource1>()
		{
			return GetTableName(typeof(TSource1));
		}

		internal Query<TSource> CreateCount(Expression<Func<TSource, bool>> predicate)
		{
			_query.WriteToCountBlock(GetTableName<TSource>());
			this.Where(predicate);
			return this;
		}

		/// <summary>
		/// Count objects by predicate.
		/// </summary>
		/// <typeparam name="TSource"></typeparam>
		/// <param name="source"></param>
		/// <param name="predicate"></param>
		/// <returns></returns>
		public int Count(Expression<Func<TSource, bool>> predicate)
		{
			return this.CreateCount(predicate)._query.ExecuteScalar<TSource, int>();
		}

		/// <summary>
		/// Count objects by previous where condition.
		/// </summary>
		/// <returns></returns>
		public int Count()
		{
			return this.CreateCount(null)._query.ExecuteScalar<TSource, int>();
		}

		public bool Any()
		{
			Query q = new Query();
			q.CreateQuery2("SELECT 1 WHERE EXISTS(", this.Select(o=>1), ")");
			int? res = q.ExecuteScalar<TSource, int?>();
			return res == 1;
		}

		internal Query<TSource> CreateDelete(Expression<Func<TSource, bool>> predicate)
		{
			_query.WriteToDeleteBlock(GetTableName<TSource>());
			this.Where(predicate);
			return this;
		}

		/// <summary>
		/// Delete objects by predicate condition. Returns number of rows deleted.
		/// </summary>
		/// <param name="predicate"></param>
		/// <returns></returns>
		public int DeleteDirect(Expression<Func<TSource, bool>> predicate)
		{
			return this.CreateDelete(predicate)._query.ExecuteNonQuery<TSource>();
		}

		/// <summary>
		/// Create UPDATE query.
		/// </summary>
		/// <typeparam name="TSource"></typeparam>
		/// <param name="source"></param>
		/// <param name="m"></param>
		/// <param name="predicate"></param>
		/// <returns></returns>
		internal Query<TSource> CreateUpdate(Action<TSource> m, Expression<Func<TSource, bool>> predicate)
		{
			TSource ee = Activator.CreateInstance<TSource>();
			m.Invoke(ee);

			if (ee is Entity.EntityBase)
			{
				_query.GetSqlBlock<SqlUpdateBlock>().Add(
					GetTableName<TSource>(),
					((Entity.EntityBase)(object)ee).GetModifiedValues().Select(o => new QParameter() { Column = o.Key, Value = o.Value }).ToArray());
			}
			else
			{
				throw new NotImplementedException("there is no method defined to update object of type " + ee.GetType().FullName);
			}
			this.Where(predicate);
			return this;
		}

		/// <summary>
		/// Update objects by predicate condition. Returns number of rows affected.
		/// </summary>
		/// <typeparam name="TSource"></typeparam>
		/// <param name="source"></param>
		/// <param name="m"></param>
		/// <param name="predicate"></param>
		/// <returns></returns>
		public int UpdateDirect(Action<TSource> m, Expression<Func<TSource, bool>> predicate)
		{
			return this.CreateUpdate(m, predicate)._query.ExecuteNonQuery<TSource>();
		}

		/// <summary>
		/// Create INSERT query.
		/// </summary>
		/// <typeparam name="TSource"></typeparam>
		/// <param name="source"></param>
		/// <param name="m"></param>
		/// <returns></returns>
		internal Query<TSource> CreateInsert(Action<TSource> m)
		{
			if (_query.SelectSqlSegments().Any())
			{
				throw new NotSupportedException("Insert(....) can be used on new clean Query<> instance only");
			}

			TSource ee = Activator.CreateInstance<TSource>();
			m.Invoke(ee);
			if (ee is Entity.EntityBase)
			{
				_query.GetSqlBlock<SqlInsertBlock>().Add(
					GetTableName<TSource>(), 
					((Entity.EntityBase)(object)ee).GetModifiedValues().Select(o => new QParameter() { Column = o.Key, Value = o.Value }).ToArray());
			}
			else
			{
				throw new NotImplementedException("there is no method defined to insert object of type " + ee.GetType().FullName);
			}

			return this;
		}

		/// <summary>
		/// Insert object with values. Returns inserted object filled with default data from database.
		/// </summary>
		/// <typeparam name="TSource"></typeparam>
		/// <param name="source"></param>
		/// <param name="m"></param>
		/// <returns></returns>
		public TSource InsertDirect(Action<TSource> m)
		{
			if (_query.SelectSqlSegments().Any())
			{
				throw new NotSupportedException("Insert(....) can be used on new clean Query<> instance only");
			}

			TSource ee = Activator.CreateInstance<TSource>();
			m.Invoke(ee);
			if (ee is Entity.EntityBase)
			{
				ee.SetDataStore(_query.DataStore);
				((Entity.EntityBase)(object)ee).SaveToDB();// this way we let the code to handle default values for db fields. Update identity value after insert and return object with new values.
			}
			else
			{
				throw new NotImplementedException("there is no method defined to insert object of type " + ee.GetType().FullName);
			}
			return ee;
		}

		public Query<TSource> LogQuery(StringBuilder sb, bool writeTime, bool simulation)
		{
			_query.WriteLog += new Action<string>((string obj) =>
			{
				sb.Append(obj);
			});

			_query.SimulationMode = simulation;
			_query.WriteTime = writeTime;
			return this;
		}

		public Query<TSource> LogQuery(string fileLocation, bool writeTime, bool simulation)
		{
			_query.WriteLog += new Action<string>((string obj) =>
			{
				File.AppendAllText(fileLocation, obj);
			});

			_query.SimulationMode = simulation;
			_query.WriteTime = writeTime;
			return this;
		}

		public Query<TSource> Case(bool condition, Func<Query<TSource>, Query<TSource>> funcIfTrue, Func<Query<TSource>, Query<TSource>> funcIfFalse = null)
		{
			if (condition == true && funcIfTrue != null)
			{
				return funcIfTrue.Invoke(this);
			}
			else if (condition == false && funcIfFalse != null)
			{
				return funcIfFalse.Invoke(this);
			}
			return this;
		}

		public Query<TSource> SetDataStore(DataStore.DataStoreBase dataStore)
		{
			_query.DataStore = dataStore;
			return this;
		}

		public Query<TSource> SetDataStore(string dataStoreName)
		{
			SetDataStore(DataStoreCollection.GetByDatabaseName(dataStoreName));
			return this;
		}

		public List<TSource> ToList()
		{
			return _query.ToList<TSource, TSource>();
		}

		public TSource FirstOrDefault()
		{
			_query.TakeCount = 1;
			return this.ToList().FirstOrDefault();
		}

		public List<SimpleDataTable> ToSimpleDataSet()
		{
			return _query.ToSimpleDataSet<TSource>();
		}

		public TResult ExecuteScalar<TResult>()
		{
			return _query.ExecuteScalar<TSource, TResult>();
		}

		public int ExecuteNonQuery()
		{
			return _query.ExecuteNonQuery<TSource>();
		}

		#region ICompilableQuery Members

		IEnumerable<SqlSegment> ICompilableQuery.SelectSqlSegments()
		{
			return _query.SelectSqlSegments();
		}

		CompiledQuery ICompilableQuery.CompileQuery()
		{
			return _query.CompileQuery();
		}

		#endregion
	}

}
