﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using System.Data;
using System.Linq;

namespace Ice.Orm.Core.DataStore
{
	/// <summary>
	/// case non-sensitive dictionary
	/// </summary>
	public class CaselessDictionary : IEnumerable<KeyValuePair<string, object>>
	{
		private Dictionary<string, object> mDic = new Dictionary<string, object>();
		private Dictionary<string, object> mLowerCaseDic;

		private Dictionary<string, object> LowerCaseDic { get { return mLowerCaseDic ?? (mLowerCaseDic = mDic.ToDictionary(o => o.Key.ToLower(), o => o.Value)); } }

		public bool TryGetValue(string key, out object o)
		{
			return (this.mDic.TryGetValue(key, out o) || this.LowerCaseDic.TryGetValue(key.ToLower(), out o));
		}

		/// <summary>
		/// Creates a shallow copy of GDictionary
		/// </summary>
		/// <returns></returns>
		public CaselessDictionary Clone()
		{
			CaselessDictionary dic = new CaselessDictionary();
			foreach (KeyValuePair<string, object> k in this)
			{
				dic[k.Key] = k.Value;
			}
			return dic;
		}

		/// <summary>
		/// GET returns null if key doesn't exist (compatible with Hashtable!)
		/// </summary>
		/// <param name="key"></param>
		/// <returns></returns>
		public object this[string key]
		{
			get
			{
				Object o;
				this.TryGetValue(key, out o);
				return o;
			}
			set
			{
				this.mDic[key] = value;
			}
		}

		public bool ContainsKey(string key)
		{
			return this.mDic.ContainsKey(key) || this.LowerCaseDic.ContainsKey(key.ToLower());
		}

		public CaselessDictionary()
		{
		}

		public CaselessDictionary(DataRow dr)
		{
			foreach (DataColumn col in dr.Table.Columns)
			{
				this[col.ColumnName] = dr[col.ColumnName];
			}
		}

		public CaselessDictionary(Hashtable ht)
		{
			foreach (DictionaryEntry d in ht)
			{
				this[d.Key + ""] = d.Value;
			}
		}

		public CaselessDictionary(Dictionary<string, object> dic)
		{
			this.mDic = dic;
		}

		public Dictionary<string, object> ToDictionary()
		{
			return this.mDic;
		}

		public DataRow ToDataRow()
		{
			object [] itemArray = new object[this.Count];
			int index = 0;

			DataTable dt = new DataTable();

			foreach(KeyValuePair<string, object> kv in this)
			{
				dt.Columns.Add(kv.Key);
				itemArray[index] = kv.Value; //// kv.Value == null ? System.DBNull.Value : kv.Value;
				index++;
			}
			return dt.Rows.Add(itemArray);
		}

		public int Count
		{
			get { return this.mDic.Count; }
		}

		public bool Remove(string key)
		{
			if (mLowerCaseDic != null)
			{
				mLowerCaseDic.Remove(key.ToLower());
			}
			return this.mDic.Remove(key);
		}

		#region IEnumerable<KeyValuePair<string,object>> Members

		IEnumerator<KeyValuePair<string, object>> IEnumerable<KeyValuePair<string, object>>.GetEnumerator()
		{
			return this.mDic.GetEnumerator();
		}

		#endregion

		#region IEnumerable Members

		IEnumerator IEnumerable.GetEnumerator()
		{
			return this.mDic.GetEnumerator();
		}

		#endregion
	}

}
