﻿using System;
using System.Data;
using System.Collections;
using System.Collections.Generic;
using Ice.Orm.Core.Criteria;
using Ice.Orm.Core.Schema;

namespace Ice.Orm.Core.Entity
{

	[AttributeUsage(AttributeTargets.Property | AttributeTargets.Field, AllowMultiple = false, Inherited = true)]
	public class EntityFieldSetting : Attribute
	{

		#region Constructors

		public EntityFieldSetting(string columnName)
		{
			this.ColumnName = columnName;
		}

		#endregion

		#region internal properties for debugging only

		internal string TableName { get; set; }

		/// <summary>
		/// get column name with table name: table.column
		/// </summary>
		internal string TableAndColumnName { get { return (TableName != null ? TableName + "." : null) + this.ColumnName; } }

		#endregion

		#region Public properties

		public bool AllowDBNull{get;set;}

		public bool IsRowVersion{get;set;}

		public bool IsLong{get;set;}

		public bool IsAutoIncrement { get; set; }

		public bool IsPrimaryKey { get; set; }

		public string BaseColumnName { get; set; }

		public string BaseSchemaName { get; set; }

		public string BaseTableName { get; set; }

		/// <summary>
		/// The name of the database column for storing the property decorated with this attribute.
		/// </summary>
		public string ColumnName { get; set; }

		public Type DataType { get; set; }

		public string TypeName { get; set; }

		public DataSetDateTime DateTimeMode { get; set; }

		private bool _defaultValueHasValue;
		private object _defaultValue;
		/// <summary>
		/// Database defined default value, if not defined, returns null
		/// </summary>
		public object DefaultValue
		{
			get
			{
				object value = _defaultValueHasValue ? _defaultValue : null;
				if (value == null || value == DBNull.Value)
				{
					return null; // default value is not set
				}
				else if (value is string)
				{
					if (this.DataType == typeof(string))
					{
						return value;
					}
					else if (this.DataType == typeof(DateTime) && string.Equals(value, "getdate()"))
					{
						return DateTime.Now;
					}

					// conversion is not supported
					throw new NotSupportedException();
				}
				else
				{
					// convert to this.DataType
					// byte, int, decimal, float, bool etc.
					return Convert.ChangeType(value, this.DataType);
				}
			}

			set
			{
				_defaultValue = value;
				_defaultValueHasValue = true;
			}
		}

		/// <summary>
		/// Field size in the database. If field is string, means length of the string.
		/// </summary>
		public int ColumnSize { get; set; }

		public int ColumnOrdinal { get; set; }

		public int NumericPrecision { get; set; }

		public int NumericScale { get; set; }

		public bool IsReadOnly { get; set; }

		public bool IsUnique { get; set; }

		/// <summary>
		/// <para>Idenitity or autonumber field</para>
		/// </summary>
		public bool IsIdentity { get; set; }

		/// <summary>
		/// If field's value == MagicValue, System.DBNull.Value will be saved to database
		/// </summary>
		public object MagicValue { get; set; }

		/// <summary>
		///  value is retrieved using DataReader, this value is required to be set in IDbDataParameter for update and insert on OleDb and ODBC drivers
		/// </summary>
		internal object ProviderType { get; set; }

		public string DbTypeDesc { get; set; }

		#endregion

	}
}

