﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Ice.Orm.Core.Entity;
using Ice.Orm.Core.Schema;
using System.Reflection;

namespace Ice.Orm.Core.Schema
{
	/// <summary>
	/// contains settings from attributes
	/// </summary>
	internal class AttributedTableHelper
	{
		private class AttributedTable
		{
			public bool IsValid { get; set; }
			public EntitySetting Table { get; set; }
			public Type EntityBaseType {get;set;}
			public EntityCache Cache { get; set; }
			
			/// <summary>
			/// attributed columns settings used for LINQ only, it's values are not updated from database schema
			/// </summary>
			public Dictionary<MemberInfo, EntityFieldSetting> Columns = new Dictionary<MemberInfo,EntityFieldSetting>();
			
			/// <summary>
			/// attributed associations settings used for LINQ only, it's values are not updated from database
			/// </summary>
			public Dictionary<MemberInfo, AssociationAttribute> Associations = new Dictionary<MemberInfo, AssociationAttribute>();
		}

		static Dictionary<Type, AttributedTable> mAttributedTables = new Dictionary<Type, AttributedTable>();

		private static AttributedTable GetTable(Type type)
		{
			if (type == null)
			{
				return null;
			}

			AttributedTable tt;

			lock (mAttributedTables)
			{
				if (!mAttributedTables.TryGetValue(type, out tt))
				{
					tt = new AttributedTable();
					tt.EntityBaseType = type;

					EntitySetting es = type.GetCustomAttributes(typeof(EntitySetting), true).Cast<EntitySetting>().FirstOrDefault();
					if (es != null)
					{
						tt.IsValid = true;
						tt.Table = es;
						tt.Cache = type.GetCustomAttributes(typeof(EntityCache), true).Cast<EntityCache>().FirstOrDefault();
						
						var items = type.GetProperties()
							.Select(o => new
							{
								mi = (MemberInfo)o,
								pi = o,
								efs = o.GetCustomAttributes(typeof(EntityFieldSetting), true).Cast<EntityFieldSetting>().FirstOrDefault(),
								assoc = o.GetCustomAttributes(typeof(AssociationAttribute), true).Cast<AssociationAttribute>().FirstOrDefault()
							}).ToList();

						tt.Columns = items.Where(o => o.efs != null).Select(o=>
						{
							// o.efs.TableName = es.SchemaTableHelper.TableName; // slow, has to read schema from database
							o.efs.TableName = es.TableName; // faster way
							return o;
						})
						.ToDictionary(o => o.mi, o => o.efs);
						
						tt.Associations = items.Where(o => o.assoc != null).Select(o=>
						{
							o.assoc.EntitySetting = es;
							return o;
						})
						.ToDictionary(o => o.mi, o => o.assoc);

						mAttributedTables.Add(type, tt);
					}
					else
					{
						mAttributedTables.Add(type, new AttributedTable() { IsValid = false });
					}
				}
			}
			return tt;
		}

		public static EntitySetting GetEntitySetting(Type type)
		{
			AttributedTable tt = GetTable(type);
			if (tt != null && tt.IsValid)
			{
				return tt.Table;
			}
			else
			{
				return null;
			}
		}

		public static EntityCache GetEntityCachePolicy(Type type)
		{
			AttributedTable tt = GetTable(type);
			if (tt != null && tt.IsValid)
			{
				return tt.Cache;
			}
			else
			{
				return null;
			}
		}

		public static EntityFieldSetting GetEntityFieldSetting(MemberInfo mi)
		{
			AttributedTable tt = GetTable(mi.DeclaringType);
			if (tt != null && tt.IsValid)
			{
				EntityFieldSetting efs;
				tt.Columns.TryGetValue(mi, out efs);
				return efs;
			}
			else
			{
				return null;
			}
		}

		public static AssociationAttribute GetAssociation(MemberInfo mi)
		{
			AttributedTable tt = GetTable(mi.DeclaringType);
			if(tt != null && tt.IsValid)
			{
				AssociationAttribute assoc;
				tt.Associations.TryGetValue(mi, out assoc);
				return assoc;
			}
			else
			{
				return null;
			}
		}


	}
}
