﻿using System;
using System.Data;
using System.Collections;
using System.Reflection;
using System.Linq;
using System.Data.Linq;

namespace Ice.Orm.Core.Schema
{
	using Entity;
	using DataStore;
	using System.Collections.Generic;
	using System.Text;
	using Utility;
	using Ice.Orm.Core.Criteria;
	using LINQ;

	/// <summary>
	/// contains settings from schema read from database
	/// </summary>
	public class SchemaTableHelper
	{
		#region Private members

		SchemaColumnCollection mSchemaColumns = null;

		SimpleDataTable mGSchemaTable;
		DataTable mSchemaTable;
		DataTable mReaderSchema;

		EntityFieldSetting mAutoIncrementColumn = null;
		EntityFieldSetting mIdentityColumn = null; // use it as reference for update and return value afer insert
		SchemaColumnCollection mPrimaryKey = null;
		SchemaColumnCollection mWriteReference = null;

		#endregion

		#region Public properties

		public DataTable SchemaTable
		{
			get
			{
				EnsureSchemaLoaded();
				return mSchemaTable;
			}
		}

		public string TableName { get; private set; }

		public string DatabaseName { get; private set; }

		/// <summary>
		/// schema columns settings created from database schema
		/// </summary>
		public SchemaColumnCollection SchemaColumns
		{
			get
			{
				EnsureSchemaLoaded();
				return mSchemaColumns;
			}
		}

		/// <summary>
		/// <para>Identity or autonumber column</para>
		/// </summary>
		public EntityFieldSetting IdentityColumn
		{
			get 
			{
				EnsureSchemaLoaded();
				return mIdentityColumn;
			}
		}

		/// <summary>
		/// <para>Null if no PK, else collection of SchemaColumn objects</para>
		/// <para>Multiple SchemaColumn columns may exsit, used on update or delete</para>
		/// </summary>
		public SchemaColumnCollection PrimaryKey
		{
			get 
			{
				EnsureSchemaLoaded(); 
				return mPrimaryKey;
			}
		}

		/// <summary>
		/// <para>update or delete reference</para>
		/// </summary>
		public SchemaColumnCollection WriteReference
		{
			get
			{
				EnsureSchemaLoaded();
				return mWriteReference;
			}
		}

		#endregion

		#region Constructor

		private SchemaTableHelper(string dbName, string tableName)
		{
			// will use EntityFieldSettings attributes only
			this.TableName = tableName;
			this.DatabaseName = dbName;
		}

		#endregion

		#region Private methods

		private static object mLock = new object();

		private void EnsureSchemaLoaded()
		{
			if (mSchemaTable == null)
			{
				lock (mLock)
				{
					if (mSchemaTable == null)
					{
						// load from db on demand
						DataStoreBase store = DataStoreCollection.GetByDatabaseName(this.DatabaseName);
						DataStoreBase.Schemas schemas = store.GetSchemaTable(this.TableName);
						mSchemaTable = schemas.SchemaTable;
						mReaderSchema = schemas.ReaderSchema;
						mGSchemaTable = SimpleDataTable.Create(mSchemaTable);

						if (mSchemaColumns == null) // only if schema columns are not already created from attributed settings
						{
							ApplyColumnsSettingsFromSchema(mSchemaTable, mReaderSchema);
						}
					}
				}
			}
		}

		private void ApplyColumnsSettingsFromSchema(DataTable schemaTable, DataTable readerSchema)
		{
			////readerSchema.WriteXml("d:\\sch_reader_" + DateTime.Now.Ticks + ".xml");
			////schemaTable.WriteXmlSchema("d:\\schema_" + DateTime.Now.Ticks + ".xml");

			if (mSchemaColumns == null)
			{
				mSchemaColumns = new SchemaColumnCollection();
			}

			if (readerSchema == null)
			{
				throw new NotSupportedException("ReaderSchema is not found for current provider");
			}

			// create settings from schemaTable
			foreach (DataColumn column in schemaTable.Columns)
			{
				// this values are overwriten by EntitySetting Attributes in method ApplyEntityFieldSettings(...)

				// create new or update existing
				// mSchemaColumns should be empty because efs are created from database schema only, not from attributes
				EntityFieldSetting efs = new EntityFieldSetting(column.ColumnName);
				mSchemaColumns[column.ColumnName] = efs;
				efs.TableName = this.TableName;

				efs.AllowDBNull = column.AllowDBNull;
				efs.IsAutoIncrement = column.AutoIncrement;
				efs.IsUnique = column.Unique;
				efs.DataType = column.DataType;
				efs.DateTimeMode = column.DateTimeMode;
				efs.DefaultValue = column.DefaultValue; // will be set using EntitySetting attributes
				efs.IsReadOnly = column.ReadOnly;

				// VistaCart compatibility
				if (column.AllowDBNull && column.ColumnName.EndsWith("ID"))
				{
					////efs.MagicValue = 0; // 0 will be saved as NULL for ...ID columns
				}

				if (readerSchema != null)
				{
					// from readerschema
					//schemaCol.ProviderSpecificDataType = readerSchemaValues["ProviderSpecificDataType"] as Type; // MsSQL provider has this parameter only!!!! ODBC and OleDB don't have parameter!
					////string ssr = readerSchema.Rows.Cast<DataRow>().Select(o => o["ColumnName"]).ToStringJoined(", ");// == column.ColumnName).FirstOrDefault();
					DataRow rr = readerSchema.Rows.Cast<DataRow>().Where(o => o["ColumnName"] + "" == column.ColumnName).FirstOrDefault();
					if (rr == null)
					{
						throw new NotSupportedException("ReaderSchema doesn't have column " + column.ColumnName);
					}
					// provider type is always integger except for SqlCE where it is System.Data.SqlServerCe.SqlCeType
					efs.ProviderType = rr["ProviderType"]; // exists with SQL, MsSQL, ODBC, OleDB providers, tested!
				}
				else
				{
					throw new NotSupportedException("ReaderSchema not found");
				}

				if (column.AutoIncrement)
				{
					efs.IsReadOnly = true;
					efs.IsAutoIncrement = true;
					mAutoIncrementColumn = efs;
					efs.IsIdentity = true;
					mIdentityColumn = efs; // identity column
				}

				if (IsPK(column, schemaTable))
				{
					efs.IsPrimaryKey = true;

					// add to primary key collection
					if (mPrimaryKey == null)
					{
						mPrimaryKey = new SchemaColumnCollection();
					}
					mPrimaryKey[column.ColumnName] = efs;
				}

				// set column size for String and Byte[]
				if (column.MaxLength > 0 && (column.DataType == typeof(System.String) || column.DataType == typeof(System.Byte[])))
				{
					efs.ColumnSize = column.MaxLength;
				}
			}

			if (mWriteReference == null && mIdentityColumn != null)
			{
				mWriteReference = new SchemaColumnCollection();
				mWriteReference[mIdentityColumn.ColumnName] = mIdentityColumn;
			}

			if (mWriteReference == null && mPrimaryKey != null && mPrimaryKey.Count > 0)
			{
				mWriteReference = mPrimaryKey;
			}

			if (mWriteReference == null)
			{
				var list = mSchemaColumns.Values.Where(o => o.IsUnique).ToList();
				if (list.Count == 1)
				{
					var unique = list.First();
					mWriteReference = new SchemaColumnCollection();
					mWriteReference[unique.ColumnName] = unique;
				}
			}

			// to be done: if there is no mWriteReference, use Entity Attributed settings

		}

		private bool IsPK(DataColumn column, DataTable schemaTable)
		{
			foreach (DataColumn col in schemaTable.PrimaryKey)
			{
				if (col.ColumnName == column.ColumnName)
				{
					return true;
				}
			}
			return false;
		}

		private DataRow GetValuesFromReaderSchema(DataTable readerSchema, string columnName)
		{
			foreach (DataRow dr in readerSchema.Rows)
			{
				if ((dr["ColumnName"] + "") == columnName)
				{
					return dr;
				}
			}
			return null;
		}

		#endregion

		#region Public methods

		public SimpleDataRow NewSchemaTableDataRow()
		{
			EnsureSchemaLoaded();
			return mGSchemaTable.NewSimpleDataRow();
		}

		#endregion

		#region cache manager

		private static Dictionary<string, SchemaTableHelper> mItems = new Dictionary<string, SchemaTableHelper>();

		/// <summary>
		/// get cached or create new if not already created
		/// </summary>
		/// <param name="databaseName"></param>
		/// <param name="tableName"></param>
		/// <returns></returns>
		public static SchemaTableHelper GetSchemaTableHelper(string databaseName, string tableName)
		{
			// load SchemaTableHelper from database
			// it has nothing to do with SchemaTableHelper created from Entity attributes
			lock (mItems)
			{
				SchemaTableHelper sch;

				string key = databaseName + "." + tableName;
	
				if (!mItems.TryGetValue(key, out sch))
				{
					// not found in the collection create new instance
					// schema from database is read on-demand with SchemaTableHelper.EnsureSchemaLoaded()
					sch = new SchemaTableHelper(databaseName, tableName);

					mItems.Add(key, sch);
				}
				return sch;
			}
		}


		#endregion
	}
}
