﻿using System;
using System.Data;
using System.Data.Odbc;
using System.Data.OleDb;

namespace Ice.Orm.Core.Utility
{
	internal class TypeConverters
	{

		#region GetDefaultNetValue, return default(type) or "", or ' ' or byte[]{0} value

		/// <summary>
		/// Obtain the default NullValue to use with the TableColumn attribute for columns mapping
		/// to a system type that does not allow null assignment. This method is intended for use
		/// by code generators.
		/// </summary>
		public static object GetDefaultNotNullNetValue(Type type)
		{
			return GetDefaultNotNullNetValue(Type.GetTypeCode(type));
		}

		/// <summary>
		/// Obtain the default NullValue to use with the TableColumn attribute for columns mapping
		/// to a system type that does not allow null assignment. This method is intended for use
		/// by code generators.
		/// </summary>
		public static object GetDefaultNotNullNetValue(TypeCode tc)
		{
			switch (tc)
			{
				case TypeCode.DBNull:
					return System.DBNull.Value;
				case TypeCode.SByte:
					return 0;
				case TypeCode.Empty:
					return null; // nullable
				case TypeCode.Object: // byte[], Nullable
					return new byte[]{(byte)0}; // nullable
				case TypeCode.Byte:
				case TypeCode.Int16:
				case TypeCode.Int32:
				case TypeCode.Int64:
				case TypeCode.UInt16:
				case TypeCode.UInt32:
				case TypeCode.UInt64:
				case TypeCode.Single:
				case TypeCode.Double:
				case TypeCode.Decimal:
					return 0;
				case TypeCode.DateTime:
					return DateTime.MinValue;
				case TypeCode.Boolean:
					return false;
				case TypeCode.Char: // char is value type!!!
					return ' ';
				case TypeCode.String:
					return ""; // nullable
				default:
					throw new Exception("No DefaultNetValue for TypeCode " + tc);
			}
		}

		#endregion

		#region OdbcType -> TypeCode conversion

		public static TypeCode GetTypeCode(OdbcType odbcType)
		{
			switch (odbcType)
			{
				case OdbcType.BigInt:
					return TypeCode.Int64;
				case OdbcType.Binary:
					return TypeCode.Object; // byte[]
				case OdbcType.Bit:
					return TypeCode.Boolean;
				case OdbcType.Char:
					return TypeCode.Char;
				case OdbcType.Date:
					return TypeCode.DateTime;
				case OdbcType.DateTime:
					return TypeCode.DateTime;
				case OdbcType.Decimal:
					return TypeCode.Decimal;
				case OdbcType.Double:
					return TypeCode.Double;
				case OdbcType.Image:
					return TypeCode.Object;
				case OdbcType.Int:
					return TypeCode.Int32;
				case OdbcType.NChar:
					return TypeCode.Char;
				case OdbcType.NText:
					return TypeCode.String;
				case OdbcType.Numeric:
					return TypeCode.Decimal;
				case OdbcType.NVarChar:
					return TypeCode.String;
				case OdbcType.Real:
					return TypeCode.Single;
				case OdbcType.SmallDateTime:
					return TypeCode.DateTime;
				case OdbcType.SmallInt:
					return TypeCode.Int16;
				case OdbcType.Text:
					return TypeCode.String;
				case OdbcType.Time:
					return TypeCode.DateTime;
				case OdbcType.Timestamp:
					return TypeCode.Object;
				case OdbcType.TinyInt:
					return TypeCode.Byte;
				case OdbcType.UniqueIdentifier:
					return TypeCode.Object;
				case OdbcType.VarBinary:
					return TypeCode.Object;
				case OdbcType.VarChar:
					return TypeCode.String;
				default:
					throw new Exception("No TypeCode for OdbcType " + odbcType);
			}
		}

		#endregion

		#region SqlDbType -> TypeCode conversion

		public static TypeCode GetTypeCode(SqlDbType sqlDbType)
		{
			switch (sqlDbType)
			{
				case SqlDbType.BigInt:
					return TypeCode.Int64;
				case SqlDbType.Binary:
					return TypeCode.Object;
				case SqlDbType.Bit:
					return TypeCode.Boolean;
				case SqlDbType.Char:
					return TypeCode.Char;
				case SqlDbType.Date:
					return TypeCode.DateTime;
				case SqlDbType.DateTime:
					return TypeCode.DateTime;
				case SqlDbType.DateTime2:
					return TypeCode.DateTime;
				case SqlDbType.DateTimeOffset:
					return TypeCode.DateTime;
				case SqlDbType.Decimal:
					return TypeCode.Decimal;
				case SqlDbType.Float:
					return TypeCode.Double;
				case SqlDbType.Image:
					return TypeCode.Object;
				case SqlDbType.Int:
					return TypeCode.Int32;
				case SqlDbType.Money:
					return TypeCode.Decimal;
				case SqlDbType.NChar:
					return TypeCode.Char;
				case SqlDbType.NText:
					return TypeCode.String;
				case SqlDbType.NVarChar:
					return TypeCode.String;
				case SqlDbType.Real:
					return TypeCode.Single;
				case SqlDbType.SmallDateTime:
					return TypeCode.DateTime;
				case SqlDbType.SmallInt:
					return TypeCode.Int16;
				case SqlDbType.SmallMoney:
					return TypeCode.Decimal;
				case SqlDbType.Structured: // uknown
					return TypeCode.String;
				case SqlDbType.Text:
					return TypeCode.String;
				case SqlDbType.Time:
					return TypeCode.DateTime;
				case SqlDbType.Timestamp:
					return TypeCode.Object;
				case SqlDbType.TinyInt:
					return TypeCode.Byte;
				case SqlDbType.Udt: // unknown
					return TypeCode.String;
				case SqlDbType.UniqueIdentifier:
					return TypeCode.Object;
				case SqlDbType.VarBinary:
					return TypeCode.Object;
				case SqlDbType.VarChar:
					return TypeCode.String;
				case SqlDbType.Variant: // unknown
					return TypeCode.String;
				case SqlDbType.Xml: // unknown
					return TypeCode.String;
				default:
					throw new Exception("No TypeCode for SqlDbType " + sqlDbType);
			}
		}

		#endregion

		#region SqlDbType -> OdbcType conversion

		public static OdbcType GetOdbcType(SqlDbType sqlDbType)
		{
			switch (sqlDbType)
			{
				case SqlDbType.BigInt:
					return OdbcType.BigInt;
				case SqlDbType.Binary:
					return OdbcType.Binary;
				case SqlDbType.Bit:
					return OdbcType.Bit;
				case SqlDbType.Char:
					return OdbcType.Char;
				case SqlDbType.Date:
					return OdbcType.Date;
				case SqlDbType.DateTime:
					return OdbcType.DateTime;
				case SqlDbType.DateTime2:
					return OdbcType.DateTime;
				case SqlDbType.DateTimeOffset:
					return OdbcType.DateTime;
				case SqlDbType.Decimal:
					//return DbType.Decimal; // decimal is not allowed to be used with ODBC driver
					return OdbcType.Double;
				case SqlDbType.Float:
					//return DbType.Decimal; // decimal is not allowed to be used with ODBC driver
					return OdbcType.Double;
				case SqlDbType.Image:
					return OdbcType.Image;
				case SqlDbType.Int:
					return OdbcType.Int;
				case SqlDbType.Money:
					//return DbType.Decimal; // decimal is not allowed to be used with ODBC driver
					return OdbcType.Double;
				case SqlDbType.NChar:
					return OdbcType.NChar;
				case SqlDbType.NText:
					return OdbcType.NText;
				case SqlDbType.NVarChar:
					return OdbcType.NVarChar;
				case SqlDbType.Real:
					return OdbcType.Real;
				case SqlDbType.SmallDateTime:
					return OdbcType.SmallDateTime;
				case SqlDbType.SmallInt:
					return OdbcType.SmallInt;
				case SqlDbType.SmallMoney:
					//return DbType.Decimal; // decimal is not allowed to be used with ODBC driver
					return OdbcType.Double;
				case SqlDbType.Structured: // uknown
					return OdbcType.Text;
				case SqlDbType.Text:
					return OdbcType.Text;
				case SqlDbType.Time:
					return OdbcType.Time;
				case SqlDbType.Timestamp:
					return OdbcType.Timestamp;
				case SqlDbType.TinyInt:
					return OdbcType.TinyInt;
				case SqlDbType.Udt: // unknown
					return OdbcType.NText;
				case SqlDbType.UniqueIdentifier:
					return OdbcType.UniqueIdentifier;
				case SqlDbType.VarBinary:
					return OdbcType.VarBinary;
				case SqlDbType.VarChar:
					return OdbcType.VarChar;
				case SqlDbType.Variant: // unknown
					return OdbcType.NText;
				case SqlDbType.Xml: // unknown
					return OdbcType.NText;
				default:
					throw new Exception("No OdbcType for SqlDbType " + sqlDbType);
			}
		}

		#endregion

		#region SqlDbType -> OleDbType conversion

		public static OleDbType GetOleDbType(SqlDbType sqlDbType)
		{
			switch (sqlDbType)
			{
				case SqlDbType.BigInt:
					return OleDbType.BigInt;
				case SqlDbType.Binary:
					return OleDbType.Binary;
				case SqlDbType.Bit:
					return OleDbType.Boolean;
				case SqlDbType.Char:
					return OleDbType.Char;
				case SqlDbType.Date:
					return OleDbType.Date;
				case SqlDbType.DateTime:
					return OleDbType.Date;
				case SqlDbType.DateTime2:
					return OleDbType.Date;
				case SqlDbType.DateTimeOffset:
					return OleDbType.Date;
				case SqlDbType.Decimal:
					return OleDbType.Decimal;
				case SqlDbType.Float:
					return OleDbType.Double;
				case SqlDbType.Image:
					return OleDbType.LongVarBinary;
				case SqlDbType.Int:
					return OleDbType.Integer;
				case SqlDbType.Money:
					return OleDbType.Currency;
				case SqlDbType.NChar:
					return OleDbType.WChar;
				case SqlDbType.NText:
					return OleDbType.LongVarWChar;
				case SqlDbType.NVarChar:
					return OleDbType.VarWChar;
				case SqlDbType.Real:
					return OleDbType.Single;
				case SqlDbType.SmallDateTime:
					return OleDbType.Date;
				case SqlDbType.SmallInt:
					return OleDbType.SmallInt;
				case SqlDbType.SmallMoney:
					return OleDbType.Currency;
				case SqlDbType.Structured: // uknown
					return OleDbType.IUnknown;
				case SqlDbType.Text:
					return OleDbType.VarWChar;
				case SqlDbType.Time:
					return OleDbType.DBTime;
				case SqlDbType.Timestamp:
					return OleDbType.DBTimeStamp;
				case SqlDbType.TinyInt:
					return OleDbType.TinyInt;
				case SqlDbType.Udt: // unknown
					return OleDbType.IUnknown;
				case SqlDbType.UniqueIdentifier:
					return OleDbType.VarChar;
				case SqlDbType.VarBinary:
					return OleDbType.VarBinary;
				case SqlDbType.VarChar:
					return OleDbType.VarChar;
				case SqlDbType.Variant: // unknown
					return OleDbType.IUnknown;
				case SqlDbType.Xml: // unknown
					return OleDbType.LongVarWChar;
				default:
					throw new Exception("No OleDbType for SqlDbType " + sqlDbType);
			}
		}

		#endregion

		#region TypeCode -> DbType conversions

		public static DbType GetDbType(TypeCode typeCode)
		{
			switch (typeCode)
			{
				case TypeCode.Boolean:
					return DbType.Boolean;
				case TypeCode.Byte:
					return DbType.Byte;
				case TypeCode.Char:
					return DbType.String;
				case TypeCode.DateTime:
					return DbType.DateTime;
				case TypeCode.DBNull:
					return DbType.String;
				case TypeCode.Decimal:
					//return DbType.Decimal; // decimal is not allowed to be used with ODBC driver
					return DbType.Double;
				case TypeCode.Double:
					return DbType.Double;
					////return DbType.Decimal;
				case TypeCode.Empty:
					return DbType.String;
				case TypeCode.Int16:
					return DbType.Int16;
				case TypeCode.Int32:
					return DbType.Int32;
				case TypeCode.Int64:
					return DbType.Int64;
				case TypeCode.Object: //// binary (byte[])
					return DbType.Binary;
				case TypeCode.SByte:
					return DbType.SByte;
				case TypeCode.Single:
					return DbType.Single;
				case TypeCode.String:
					return DbType.String;
				case TypeCode.UInt16:
					return DbType.UInt16;
				case TypeCode.UInt32:
					return DbType.UInt32;
				case TypeCode.UInt64:
					return DbType.UInt64;
				default:
					throw new Exception("No DbType for TypeCode " + typeCode);
			}
		}

		#endregion

		#region DbType -> TypeCode conversions

		public static TypeCode GetTypeCode(DbType dbType)
		{
			switch (dbType)
			{
				case DbType.Boolean:
					return TypeCode.Boolean;
				case DbType.Byte:
					return TypeCode.Byte;
				case DbType.DateTime:
					return TypeCode.DateTime;
				case DbType.Decimal:
					return TypeCode.Decimal;
				case DbType.Double:
					return TypeCode.Double;
				case DbType.Int16:
					return TypeCode.Int16;
				case DbType.Int32:
					return TypeCode.Int32;
				case DbType.Int64:
					return TypeCode.Int64;
				case DbType.Object: // binary (byte[])
					return TypeCode.Object;
				case DbType.SByte:
					return TypeCode.SByte;
				case DbType.Single:
					return TypeCode.Single;
				case DbType.String:
					return TypeCode.String;
				case DbType.UInt16:
					return TypeCode.UInt16;
				case DbType.UInt32:
					return TypeCode.UInt32;
				case DbType.UInt64:
					return TypeCode.UInt64;
				default:
					throw new Exception("No TypeCode for DbType " + dbType);
			}
		}

		#endregion

		public static DbType GetDbType(SqlDbType sqlDbType)
		{
			return GetDbType(GetTypeCode(sqlDbType));
		}

		public static DbType GetDbType(OdbcType odbcType)
		{
			return GetDbType(GetTypeCode(odbcType));
		}

		public static DbType GetDbType(Type type)
		{
			return GetDbType(Type.GetTypeCode(type));
		}

	}
}
