﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Linq;

namespace Ice.Converter
{
	public class IceRadixConvert
	{
		private string _charsTable = IceBaseConvert.JavaScriptRFCTable64;
		public string CharsTable { get { return _charsTable; } set { _charsTable = value; } }

		#region static helper methods

		/// <summary>
		/// convert from decimal number to any radix (up to radix 64)
		/// </summary>
		private static string FromDecimal(long decimalValue, int toRadix, string table)
		{
			bool isNegative = decimalValue < 0;
			decimalValue = Math.Abs(decimalValue);

			StringBuilder sb = new StringBuilder();
			long A = decimalValue;
			while (A >= toRadix)
			{
				long modulo = A % toRadix;
				A = A / toRadix;
				sb.Insert(0, table[(int)modulo]+"");
			}
			sb.Insert(0, table[(int)A] + "");
			return (isNegative ? "-" : null) + sb.ToString();
		}

		/// <summary>
		/// Convert to decimal number from any radix (up to radix 64)
		/// </summary>
		private static long ToDecimal(string value, int fromRadix, string table)
		{
			if (string.IsNullOrEmpty(value))
			{
				throw new ArgumentNullException("value");
			}
			bool isNegative = value.StartsWith("-");
			if (isNegative)
			{
				value = value.Substring(1);
			}

			value = Reverse(value);
			long decimalValue = 0;
			for (int n = 0; n < value.Length; n++)
			{
				int index = table.IndexOf(value[n]);
				if (index < 0)
				{
					// invalid char, not found in table
					return -1;
				}
				long factor = (long)Math.Pow(fromRadix, n);
				decimalValue += index * factor;
			}
			return (isNegative ? -1 : 1) * decimalValue;
		}

		private static string Reverse(string text)
		{
			StringBuilder sb = new StringBuilder();
			foreach (char ch in text)
			{
				sb.Insert(0, ch + "");
			}
			return sb.ToString();
		}

		#endregion

		public T Convert<T>(string value, int fromRadix, int toRadix)
		{
			if (string.IsNullOrEmpty(value))
			{
				throw new ArgumentNullException("value");
			}
			if (fromRadix < 2 || fromRadix > 64)
			{
				throw new ArgumentOutOfRangeException("fromRadix");
			}
			if (toRadix < 2 || toRadix > 64)
			{
				throw new ArgumentOutOfRangeException("toRadix");
			}

			string[] parts = value.Split('.', ',').Select(o =>
			{
				long dec = ToDecimal(o, fromRadix, this.CharsTable);
				string val = FromDecimal(dec, toRadix, this.CharsTable);
				return val;
			}).ToArray();

			StringBuilder sb = new StringBuilder();
			for (int n = 0; n < parts.Length; n++)
			{
				if (n > 0)
				{
					sb.Append('.');
				}
				sb.Append(parts[n]);
			}
			string str = sb.ToString();

			return (T)System.Convert.ChangeType(str, Nullable.GetUnderlyingType(typeof(T)) ?? typeof(T), null); // can't just cast values (e.g. double to decimal, string to int, etc., must use Convert)
		}

		public decimal Convert(string value, int fromRadix, int toRadix)
		{
			return this.Convert<decimal>(value, fromRadix, toRadix);
		}



	}
}