﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection;

namespace Ice.Extensions
{
	/// <summary>
	/// This attribute is used to represent a string value
	/// for a value in an enum.
	/// </summary>
	[AttributeUsage(AttributeTargets.Field)]
	public class StringValueAttribute : Attribute
	{
		#region Properties

		/// <summary>
		/// Holds the stringvalue for a value in an enum.
		/// </summary>
		public string StringValue { get; private set; }

		#endregion

		#region Constructor

		/// <summary>
		/// Constructor used to init a StringValue Attribute
		/// </summary>
		/// <param name="value"></param>
		public StringValueAttribute(string value)
		{
			this.StringValue = value;
		}

		#endregion
	}

	public static class EnumExtensions
	{
		/// <summary>
		/// Will get the string value for a given enums value, this will
		/// only work if you assign the StringValue attribute to
		/// the items in your enum. If StringValue attribute is not defined, return value.ToString() if useEnumName==true, else throw Exception
		/// </summary>
		/// <param name="value"></param>
		/// <param name="useEnumName">true: if StringValue attribute is not found, return enumvalue.ToString(), false: always return StringValue or null if StringValue is not defined</param>
		/// <returns></returns>
		public static string GetStringValue(this Enum value, bool useEnumName = false)
		{
			string strval = value + "";

			// Get fieldinfo for this type
			FieldInfo fieldInfo = value.GetType().GetField(strval);

			// Get the stringvalue attributes
			StringValueAttribute[] attribs = fieldInfo.GetCustomAttributes(typeof(StringValueAttribute), false) as StringValueAttribute[];

			// Return the first if there was a match.
			if (attribs != null && attribs.Length > 0)
			{
				return attribs[0].StringValue;
			}
			else if (useEnumName)
			{
				return strval;
			}
			else
			{
				throw new Exception("StringValue for " + value + " not found in enum type " + value.GetType());
			}
		}
	}
}
