﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;

namespace Ice.Utility
{
	public class IceTypeConverter
	{
		private static bool TryChangeTypeWorker(object obj, Type toType, out object result, bool exceptionIfFailed)
		{
			if (obj == System.DBNull.Value)
			{
				obj = null;
			}

			if (obj == null)
			{
				if (toType == typeof(string)) // string has no parameterless constructor
				{
					result = null;
				}
				else if (toType == typeof(byte[])) // byte[] has no parameterless constructor
				{
					result = null;
				}
				else
				{
					result = Activator.CreateInstance(toType, true);
				}
				bool success = result == null; // if nullable type, and obj is null, conversion is successful, return true
				if (!success && exceptionIfFailed)
				{
					throw new Exception("conversion failed, " + obj + " can not be converted to type " + toType.FullName);
				}
				return success;
			}
			else if (obj.GetType() == toType)
			{
				result = obj;
				return true;
			}
			else // convert, eg. "432" to int 432
			{
				// can't convert to Nullable type, must convert to it's underlying type (int, decimal, ...)
				try
				{
					// to int is used a lot and this is faster way
					if(toType == typeof(bool) || toType == (typeof(bool?)))
					{
						bool val;
						string str = obj + "";
						if(str == "1")
						{
							result = true;
							return true;
						}
						else if(str == "0")
						{
							result = false;
							return true;
						}
						else if(bool.TryParse(str, out val))
						{
							result = val;
							return true;
						}
					}
					else if (toType == typeof(int) || toType == (typeof(int?)))
					{
						int val;
						if(int.TryParse(obj+"", out val))
						{
							result = val;
							return true;
						}
					}
					if (toType == typeof(long) || toType == (typeof(long?)))
					{
						long val;
						if (long.TryParse(obj + "", out val))
						{
							result = val;
							return true;
						}
					}
					else if(toType == typeof(decimal) || toType == (typeof(decimal?)))
					{
						decimal val;
						if(decimal.TryParse(obj + "", out val))
						{
							result = val;
							return true;
						}
					}
					else if (toType == typeof(float) || toType == (typeof(float?)))
					{
						float val;
						if (float.TryParse(obj + "", out val))
						{
							result = val;
							return true;
						}
					}

					toType = Nullable.GetUnderlyingType(toType) ?? toType;

					result = Convert.ChangeType(obj, toType, null); // can't just cast values (e.g. double to decimal, string to int, etc., must use Convert)
					return true;
				}
				catch(Exception ex)
				{
					try
					{
						// on exception, try dynamic cast (this is required to deserialize Enum in Silverlight)
						result = DynamicCast(obj, toType);
						return true;
					}
					catch
					{ }

					if (exceptionIfFailed)
					{
						throw ex;
					}

					result = Activator.CreateInstance(toType);
					return false;
				}
			}
		}

		/// <summary>
		/// returns converted object, else throws exception if conversion failed
		/// </summary>
		/// <param name="obj"></param>
		/// <param name="toType"></param>
		/// <returns></returns>
		public static object ChangeType(object obj, Type toType)
		{
			object res = null;
			TryChangeTypeWorker(obj, toType, out res, true);
			return res;
		}

		/// <summary>
		/// returns converted object, else throws exception if conversion failed
		/// </summary>
		/// <typeparam name="TResult"></typeparam>
		/// <param name="obj"></param>
		/// <returns></returns>
		public static TResult ChangeType<TResult>(object obj)
		{
			object res = null;
			TryChangeTypeWorker(obj, typeof(TResult), out res, true);
			return (TResult)res;
		}

		/// <summary>
		/// Returns true if conversion is successful
		/// </summary>
		/// <param name="obj"></param>
		/// <param name="toType"></param>
		/// <param name="res"></param>
		/// <returns></returns>
		public static bool TryChangeType(object obj, Type toType, out object res)
		{
			return TryChangeTypeWorker(obj, toType, out res, false);
		}

		/// <summary>
		/// Returns true if conversion is successful
		/// </summary>
		/// <typeparam name="TResult"></typeparam>
		/// <param name="obj"></param>
		/// <param name="o"></param>
		/// <returns></returns>
		public static bool TryChangeType<TResult>(object obj, out TResult res)
		{
			object tmp = null;
			bool success = TryChangeTypeWorker(obj, typeof(TResult), out tmp, false);
			res = (TResult)tmp;
			return success;
		}

		private static TResult DynamicCastWorker<TResult>(object obj)
		{
			if (obj == null)
			{
				return default(TResult);
			}
			return (TResult)obj;
		}

		public static object DynamicCast(object obj, Type toType)
		{
			MethodInfo mi = typeof(IceTypeConverter).GetMethod("DynamicCastWorker", BindingFlags.Static | BindingFlags.NonPublic);
			MethodInfo castMethod = mi.MakeGenericMethod(toType);
			object castedObject = castMethod.Invoke(null, new object[] { obj });
			return castedObject;
		}

		public static object DefaultNetValue(Type toType)
		{
			if (toType == typeof(string)) // string has no parameterless constructor
			{
				return null;
			}
			else if (toType == typeof(byte[])) // byte[] has no parameterless constructor
			{
				return null;
			}
			else
			{
				return Activator.CreateInstance(toType);
			}
		}

	}
}
