﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Data;
using System.Collections.ObjectModel;
using System.Linq;
using System.Reflection;

namespace System.Data
{
	public class SimpleDataRow
	{
		private SimpleDataTable mGDataTable;
		private object[] mItemArray;

		public SimpleDataRow()
		{
		}

		public SimpleDataTable Table
		{
			get { return mGDataTable; }
			set { mGDataTable = value; }
		}

		public object[] ItemArray
		{
			get { return mItemArray; }
			set { mItemArray = value; }
		}

		/// <summary>
		/// Case non-sensitive columnName
		/// </summary>
		/// <param name="columnName"></param>
		/// <returns></returns>
		public object this[string columnName]
		{
			get { return mItemArray[mGDataTable.GetColumnIndex(columnName, true)]; }
			set { mItemArray[mGDataTable.GetColumnIndex(columnName, true)] = value; }
		}

		public object this[int columnIndex]
		{
			get { return mItemArray[columnIndex]; }
			set {mItemArray[columnIndex] = value; }
		}
	}

	public class SimpleDataTable
	{
		#region Private members

		/// <summary>
		/// ColumnName, ColumnIndex
		/// </summary>
		private Dictionary<String, int> mColumns;
		private Dictionary<String, int> mLowercaseColumns;
		private Dictionary<string, int> LowercaseColumns{get{return mLowercaseColumns ?? (mLowercaseColumns = mColumns.ToDictionary(o => o.Key.ToLower(), o => o.Value));}}
		private List<Object[]> mRows;
		private string mTableName;
		private int? mTotalCount;

#if !SILVERLIGHT
		private DataTable mSchemaTable;
#endif

		private Dictionary<string, int> CreateColums(IEnumerable<string> columns)
		{
			if (columns == null)
			{
				return new Dictionary<string, int>();
			}
			else
			{
				int index = 0;
				return columns.ToDictionary(o => o, o => index++);
			}
		}

		#endregion

		#region Public properties

		/// <summary>
		/// fast way to retreive columns count without converting it to array or list
		/// </summary>
		public int GetColumnsCount()
		{
			return this.mColumns.Count;
		}

		public ReadOnlyCollection<string> Columns
		{
			get { return (mColumns ?? (mColumns = new Dictionary<string, int>())).Keys.ToList().AsReadOnly(); }
			set { mColumns = CreateColums(value); mLowercaseColumns = null; } // set is required for json deserialization to work
		}

		public List<Object[]> Rows
		{
			get { return mRows ?? (mRows = new List<object[]>()); }
			set { mRows = value; } // set is required for json deserialization to work
		}

		public string TableName
		{
			get { return mTableName; }
			set { mTableName = value; }
		}

		public int? TotalCount
		{
			get { return mTotalCount; }
			set { mTotalCount = value; }
		}

		#endregion

		#region Constructors

		public SimpleDataTable()
		{
		}

		public static SimpleDataTable Create(IEnumerable<IDictionary> dictionaryList)
		{
			SimpleDataTable dd = new SimpleDataTable();
			if (dictionaryList.Any())
			{
				dd.mColumns = dictionaryList.First().Keys.Cast<string>().Select((o, n) => new { Key = o, Index = n }).ToDictionary(o => o.Key, o => o.Index);
				dd.mRows = dictionaryList.Select(o => o.Values.Cast<object>().ToArray()).ToList();
			}
			return dd;
		}

		#if !SILVERLIGHT
		public static SimpleDataTable Create(DataTable dataTable)
		{
			SimpleDataTable dd = new SimpleDataTable();
			dd.mTableName = dataTable.TableName;
			dd.mSchemaTable = dataTable;

			if (dataTable.ExtendedProperties.Contains("TotalCount"))
			{
				dd.mTotalCount = Convert.ToInt32(dataTable.ExtendedProperties["TotalCount"]);
			}

			dd.mColumns = new Dictionary<string,int>();
			dd.mRows = new List<object[]>();
			int index = 0;
			foreach (DataColumn column in dataTable.Columns)
			{
				dd.mColumns.Add(column.ColumnName, index++);
			}

			for (int n = 0; n < dataTable.Rows.Count; n++)
			{
				dd.mRows.Add(dataTable.Rows[n].ItemArray);
			}
			return dd;
		}
		#endif

		#endregion

		#region Helper public & private methods

		////private Object[] GetItemArray(IDictionary obj)
		////{
		////	Object[] values = new Object[obj.Count];
		////	int index = 0;
		////	foreach (DictionaryEntry dic in obj)
		////	{
		////		values[index++] = dic.Value;
		////	}
		////	return values;
		////}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="includeColumnsNames">if true, first row will be columns names</param>
		public string ToCSV(bool includeColumnsNames, string delimiter = ",")
		{
			// http://tools.ietf.org/html/rfc4180

			return new[] { this.Columns.Cast<object>().ToArray() }
				.Concat(this.Rows)
				.Skip(includeColumnsNames ? 0 : 1) // skip columns if required
				.Select(row => row.Select(m => "\"" + (m + "").Replace("\"", "\"\"") + "\"").ToStringJoined(delimiter))
				.ToStringJoined("\r\n");
		}

		#if !SILVERLIGHT
		public DataTable ToDataTable(bool useSchemaTable = false)
		{
			DataTable dataTable;

			if (mSchemaTable != null && useSchemaTable)
			{
				dataTable = mSchemaTable.Clone();
			}
			else
			{
				dataTable = new DataTable();

				foreach (string colName in mColumns.Keys)
				{
					dataTable.Columns.Add(colName);
				}
			}
	
			foreach (Object[] rowData in mRows)
			{
				dataTable.Rows.Add(rowData);
			}

			dataTable.TableName = this.TableName;
			dataTable.ExtendedProperties["TotalCount"] = this.TotalCount;

			return dataTable;
		}
		#endif

		/// <summary>
		/// case non-sensitive search
		/// </summary>
		/// <param name="columnName"></param>
		/// <returns></returns>
		public int GetColumnIndex(string columnName, bool exeptionIfNotFound)
		{
			int index;
			if (this.mColumns.TryGetValue(columnName, out index))
			{
				return index;
			}

			// try caseless search
			if(this.LowercaseColumns.TryGetValue(columnName.ToLower(), out index))
			{
				return index;
			}
			
			if (exeptionIfNotFound)
			{
				throw new Exception("column " + columnName + " does not exist");
			}
			else
			{
				return -1;
			}
		}

		/// <summary>
		/// Clone GDataTable definition without data
		/// </summary>
		/// <returns></returns>
		public SimpleDataTable Clone()
		{
			SimpleDataTable gdt = new SimpleDataTable();
			#if !SILVERLIGHT
			gdt.mSchemaTable = this.mSchemaTable;
			#endif
			gdt.mColumns = this.mColumns;
			return gdt;
		}

		public SimpleDataRow GetDataRow(int index)
		{
			SimpleDataRow dr = new SimpleDataRow();
			dr.Table = this;
			dr.ItemArray = this.Rows[index];
			return dr;
		}

		public SimpleDataRow NewSimpleDataRow()
		{
			SimpleDataRow dr = new SimpleDataRow();
			dr.Table = this;
			dr.ItemArray = new object[this.mColumns.Count];
			return dr;
		}

		public object[] NewRow()
		{
			return new object[this.mColumns.Count];
		}

		public IEnumerable<SimpleDataRow> AsSimpleDataRows()
		{
			foreach(object[] row in this.Rows)
			{
				yield return new SimpleDataRow() { ItemArray = row, Table = this };
			}
		}

		#endregion

	}

}
