﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

namespace System.IO
{
	public static class StreamExtensions
	{
		static int _bufferSize = 8 * 4096;

		public static IEnumerable<byte> AsEnumerable(this Stream stream)
		{
			int readCount;
			byte[] buffer = new byte[_bufferSize];
			do
			{
				readCount = stream.Read(buffer, 0, buffer.Length);
				foreach (byte b in buffer.Take(readCount))
				{
					yield return b;
				}
			}
			while (readCount > 0);
		}

		public static MemoryStream ToMemoryStream(this Stream stream)
		{
			if (stream.Position != 0)
			{
				stream.Position = 0;
			}
			byte[] buffer = new byte[_bufferSize];
			int read;
			MemoryStream ms = new MemoryStream();
			while ((read = stream.Read(buffer, 0, buffer.Length)) > 0)
			{
				ms.Write(buffer, 0, read);
			}
			return ms;
		}

		private class State
		{
			public byte[] Buffer = new byte[_bufferSize];
			public MemoryStream MS = new MemoryStream();
			public Stream Stream;
			public int ExpectedLength;
			public int TimeoutMs;
			public IAsyncResult BeginReadResult;
			public double TimeSpanMs;
			public bool Completed;
		}

		private static void EndRead(IAsyncResult res)
		{
			State state = (State)res.AsyncState;
			lock (state)
			{
				if (state.Completed)
				{
					return;
				}

				int read = state.Stream.EndRead(res);
				if (read > 0)
				{
					state.MS.Write(state.Buffer, 0, read);
					// continue read from stream
					BeginRead(state);
				}
				else if (state.ExpectedLength > 0 && state.MS.Position < state.ExpectedLength)
				{
					// wait for more data from stream
					BeginRead(state);
				}
				else
				{
					// finished
					Monitor.Pulse(state);
				}
			}
		}

		private static void BeginRead(State state)
		{
			state.BeginReadResult = state.Stream.BeginRead(state.Buffer, 0, state.Buffer.Length, new AsyncCallback(EndRead), state);
		}

		public static MemoryStream ToMemoryStream(this Stream stream, int expectedLength, int timeoutMs)
		{
			DateTime dtStart = DateTime.Now;

			State state = new State() { Stream = stream, ExpectedLength = expectedLength, TimeoutMs = timeoutMs };

			lock (state)
			{
				BeginRead(state);

				Monitor.Wait(state, state.TimeoutMs);
				state.BeginReadResult.AsyncWaitHandle.Close();
				state.TimeSpanMs = (DateTime.Now - dtStart).TotalMilliseconds;
				state.Completed = true;

				if (state.MS.Length < expectedLength)
				{
					throw new TimeoutException("Timeout waiting for data from stream.");
				}
			}
			return state.MS;
		}

	}
}
