﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;
using Ice.Threading;
using System.Data;

namespace System.Linq
{
	public static class EnumerableExtensions
	{
		/// <summary>
		/// Join string elements of List collection
		/// </summary>
		/// <typeparam name="?"></typeparam>
		/// <param name="s"></param>
		/// <param name="separator"></param>
		/// <returns></returns>
		public static string ToStringJoined<TSource>(this IEnumerable<TSource> s, string separator = null)
		{
			StringBuilder sb = new StringBuilder();
			bool hasElements = false;
			bool hasSeparator = !string.IsNullOrEmpty(separator);

			foreach (TSource item in s)
			{
				if(!hasElements)
				{
					hasElements = true;
				}
				else if (hasSeparator)
				{
					sb.Append(separator);
				}

				sb.Append(item);
			}
			return sb.ToString();
		}

		/// <summary>
		/// Concatenate single item
		/// </summary>
		/// <typeparam name="T"></typeparam>
		/// <param name="ienum"></param>
		/// <param name="item"></param>
		/// <returns></returns>
		public static IEnumerable<TSource> Concat<TSource>(this IEnumerable<TSource> source, TSource item)
		{
			foreach (TSource m in source)
			{
				yield return m;
			}

			yield return item;
		}

		/// <summary>
		/// Returns items from first collection where predicate result is true
		/// </summary>
		/// <typeparam name="TSource"></typeparam>
		/// <param name="first"></param>
		/// <param name="second"></param>
		/// <param name="predicate"></param>
		/// <returns></returns>
		public static IEnumerable<TSource> Intersect<TSource, TSource2>(this IEnumerable<TSource> first, IEnumerable<TSource2> second, Func<TSource, TSource2, bool> predicate)
		{
			return first.Where(f => second.Any(s => predicate(f, s)));
		}

		/// <summary>
		/// Returns items from first collection where firstKey == secondKey. Very fast, uses dictionary.
		/// </summary>
		/// <typeparam name="TSource"></typeparam>
		/// <typeparam name="TSource2"></typeparam>
		/// <typeparam name="TResult"></typeparam>
		/// <param name="first"></param>
		/// <param name="second"></param>
		/// <param name="firstKey"></param>
		/// <param name="secondKey"></param>
		/// <param name="excludeIntersection"></param>
		/// <returns></returns>
		public static IEnumerable<TSource> Intersect<TSource, TSource2, TResult>(this IEnumerable<TSource> first, IEnumerable<TSource2> second, Func<TSource, TResult> firstKey, Func<TSource2, TResult> secondKey, bool excludeIntersection = false )
		{
			Dictionary<TResult, bool> secondKeys = null;

			foreach (TSource item in first)
			{
				if (secondKeys == null)
				{
					secondKeys = second.Select(secondKey).Distinct(k=>k).ToDictionary(k => k, k => true);
				}

				if (secondKeys.ContainsKey(firstKey(item)) != excludeIntersection)
				{
					yield return item;
				}
			}
		}

		public static IEnumerable<TSource> Except<TSource, TSource2, TResult>(this IEnumerable<TSource> first, IEnumerable<TSource2> second, Func<TSource, TResult> firstKey, Func<TSource2, TResult> secondKey)
		{
			return first.Intersect(second, firstKey, secondKey, true);
		}


		/// <summary>
		/// Returns both element from first, and from second where firstKey == secondKey. Very fast, uses dictionary.
		/// </summary>
		/// <typeparam name="TElement"></typeparam>
		/// <typeparam name="TResult"></typeparam>
		/// <param name="first"></param>
		/// <param name="second"></param>
		/// <param name="firstKey"></param>
		/// <param name="secondKey"></param>
		/// <returns></returns>
		public static IEnumerable<TElement> Union<TElement, TResult>(this IEnumerable<TElement> first, IEnumerable<TElement> second, Func<TElement, TResult> firstKey, Func<TElement, TResult> secondKey)
		{
			Dictionary<TResult, TElement> dic = null;

			foreach (TElement e1 in first)
			{
				if (dic == null) { dic = second.ToDictionary(secondKey, o => o); }
				TResult key = firstKey(e1);

				if (dic.ContainsKey(key))
				{
					yield return e1;
					yield return dic[key];
				}
			}
		}

		/// <summary>
		/// Returns the first element of a sequence, or a defaultValue if the sequence contains no elements.
		/// </summary>
		public static TSource FirstOrValue<TSource>(this IEnumerable<TSource> source, TSource defaultValue = default(TSource))
		{
			var items = source.Take(1).ToList();
			if (items.Count == 1)
			{
				return items.First();
			}
			else
			{
				return defaultValue;
			}
		}

		/// <summary>
		/// Returns the last element of a sequence, or a defaultValue if the sequence contains no elements.
		/// </summary>
		public static TSource LastOrValue<TSource>(this IEnumerable<TSource> source, TSource defaultValue = default(TSource))
		{
			var items = source.Reverse().Take(1).ToList();

			if (items.Count == 1)
			{
				return items.First();
			}
			else
			{
				return defaultValue;
			}
		}

		/// <summary>
		/// Get value from dictionary. Returns value or default value.
		/// </summary>
		/// <typeparam name="TKey"></typeparam>
		/// <typeparam name="TValue"></typeparam>
		/// <param name="source"></param>
		/// <param name="key"></param>
		/// <returns></returns>
		public static TValue TryGetValueOrDefault<TKey, TValue>(this IDictionary<TKey, TValue> source, TKey key, TValue defaultValue = default(TValue))
		{
			TValue val;
			if (source.TryGetValue(key, out val))
			{
				return val;
			}
			else
			{
				return defaultValue;
			}
		}

		public static IEnumerable<TSource> Case<TSource>(this IEnumerable<TSource> source, bool condition, Func<IEnumerable<TSource>, IEnumerable<TSource>> onTrue, Func<IEnumerable<TSource>, IEnumerable<TSource>> onFalse = null)
		{
			if (condition)
			{
				return onTrue != null ? onTrue(source) : source;
			}
			else
			{
				return onFalse != null ? onFalse(source) : source;
			}
		}

		/// <summary>
		/// ice dictionary based distinct which is much faster
		/// </summary>
		/// <typeparam name="TSource"></typeparam>
		/// <param name="source"></param>
		/// <param name="keySelector"></param>
		/// <returns></returns>
		public static IEnumerable<TSource> Distinct<TSource, TResult>(this IEnumerable<TSource> source, Func<TSource, TResult> keySelector)
		{
			Dictionary<TResult, bool> keys = new Dictionary<TResult, bool>();

			foreach (TSource item in source)
			{
				TResult key = keySelector(item);

				if (!keys.ContainsKey(key))
				{
					keys[key] = true;
					yield return item;
				}
			}
		}

		public static IOrderedEnumerable<TSource> OrderBy<TSource, TKey>(this IEnumerable<TSource> source, Func<TSource, TKey> keySelector, bool descending)
		{
			if (descending)
			{
				return source.OrderByDescending(keySelector);
			}
			else
			{
				return source.OrderBy(keySelector);
			}
		}

		/// <summary>
		/// Cycle current items endlessly. Use Take() to limit number of items.
		/// </summary>
		/// <typeparam name="TSource"></typeparam>
		/// <param name="source"></param>
		/// <returns></returns>
		public static IEnumerable<TSource> Cycle<TSource>(this IEnumerable<TSource> source, int? take = null)
		{
			bool hasItems = false;
			int count = 0;

			do
			{
				foreach (TSource item in source)
				{
					hasItems = true;
					if (!(take >= 0) || (++count <= take))
					{
						yield return item;
					}
					else
					{
						// exit foreach and do-while
						hasItems = false;
						break;
					}
				}
			}
			while (hasItems);
		}

		private static int GetColIndex(bool balancedFill, int currentIndex, int fromItemIndex, int itemsPerColumn)
		{
			if(balancedFill && currentIndex >= fromItemIndex)
			{
				int delta = currentIndex - fromItemIndex;
				int reducedColStartIndex = fromItemIndex / itemsPerColumn;

				int col2 = delta / (itemsPerColumn - 1) + reducedColStartIndex;
				return col2;
			}
			return currentIndex / itemsPerColumn;
		}

		/// <summary>
		/// From total count calculate itemsPerColumn. Add itemsPerColumn to group1, than add itemsPerColumn to group2, etc. If balancedFill is true, add itemsPerColumn-1 to second half of groups to fill empty space in last column.
		/// </summary>
		/// <typeparam name="TElement"></typeparam>
		/// <param name="source"></param>
		/// <param name="columnsCount"></param>
		/// <param name="balancedFill"></param>
		/// <returns></returns>
		public static IEnumerable<IGrouping<int, TElement>> GroupToColumns<TElement>(this IEnumerable<TElement> source, int columnsCount, bool balancedFill = false)
		{
			if(columnsCount < 1)
			{
				throw new ArgumentException("columnsCount must be greater than zero");
			}

			List<TElement> srcList = source.ToList();
			int srcCount = srcList.Count;

			int itemsPerColumn = srcCount / columnsCount;
			if(srcCount % columnsCount > 0)
			{
				itemsPerColumn++;
			}

			int fromItemIndex = srcCount;

			if (balancedFill && srcCount % columnsCount > 0)
			{
				int missingItems = columnsCount - srcCount % columnsCount;
				int fromColumnIndex = columnsCount - missingItems;
				fromItemIndex = fromColumnIndex * itemsPerColumn;
			}

			var items = srcList.Select((o, n) => new
			{
				columnIndex = GetColIndex(balancedFill, n, fromItemIndex, itemsPerColumn),
				item = o
			}).ToList();
			
			return items.GroupBy(o => o.columnIndex, o => o.item);
		}

		/// <summary>
		/// Add itemsPerGroup items to group1, than add itemsPerGroup to group2, etc. Last group may have less than itemsPerGroup items
		/// </summary>
		/// <typeparam name="TElement"></typeparam>
		/// <param name="source"></param>
		/// <param name="itemsPerGroup"></param>
		/// <returns></returns>
		public static IEnumerable<IGrouping<int, TElement>> GroupToItemsGroups<TElement>(this IEnumerable<TElement> source, int itemsPerGroup)
		{
			List<TElement> srcList = source.ToList();
			int srcCount = srcList.Count;

			return srcList.Select((o, n) => new
			{
				columnIndex = n / itemsPerGroup,
				item = o
			}).GroupBy(o => o.columnIndex, o => o.item);
		}

		/// <summary>
		/// The same as SelectMany, select all items from column0, than from column1, etc.
		/// </summary>
		/// <typeparam name="TElement"></typeparam>
		/// <param name="source"></param>
		/// <returns></returns>
		public static IEnumerable<TElement> SelectVertically<TElement>(this IEnumerable<IGrouping<int, TElement>> source)
		{
			return source.SelectMany(columnGroup=>columnGroup.Select(rowItem=>rowItem));
		}

		/// <summary>
		/// Select 1st element from group0, then select 1st element from group1, .... then select 2nd element from group0, select 2nd element from group1, etc.
		/// </summary>
		/// <typeparam name="TElement"></typeparam>
		/// <param name="source"></param>
		/// <returns></returns>
		public static IEnumerable<TElement> SelectHorizontally<TElement>(this IEnumerable<IGrouping<int, TElement>> source)
		{
			return source.ToSimpleDataTable().SelectHorizontally<TElement>();
		}

		public static SimpleDataTable ToSimpleDataTable<TElement>(this IEnumerable<IGrouping<int, TElement>> source)
		{
			// each group in source has to be converted into datacolumn with data
			SimpleDataTable dtsimple = new SimpleDataTable();
			dtsimple.Columns = source.Select((group, columnIndex) => "Column_" + columnIndex).ToList().AsReadOnly();

			source.ForEach((group, columnIndex) =>
			{
				group.ForEach((value, rowIndex) =>
				{
					if(dtsimple.Rows.Count < rowIndex + 1)
					{
						dtsimple.Rows.Add(dtsimple.NewRow());
					}
					dtsimple.Rows[rowIndex][columnIndex] = value;
				});
			});

			return dtsimple;
		}

		public static IEnumerable<TElement> SelectHorizontally<TElement>(this SimpleDataTable table)
		{
			for (int rowIndex = 0; rowIndex < table.Rows.Count; rowIndex++)
			{
				for (int colIndex = 0; colIndex < table.Columns.Count; colIndex++)
				{
					object obj = table.Rows[rowIndex][colIndex];
					if(obj != null && typeof(TElement).IsAssignableFrom(obj.GetType()))
					{
						yield return (TElement) obj;
					}
				}
			}
		}

		public static IEnumerable<TElement> SelectVertically<TElement>(this SimpleDataTable table)
		{
			for (int colIndex = 0; colIndex < table.Columns.Count; colIndex++)
			{
				for (int rowIndex = 0; rowIndex < table.Rows.Count; rowIndex++)
				{
					yield return (TElement)table.Rows[rowIndex][colIndex];
				}
			}
		}

		public static IEnumerable<TSource> ForEach<TSource>(this IEnumerable<TSource> source, Action<TSource> action)
		{
			var items = source.ToList();
			items.ForEach(action);
			return items;
		}

		/// <summary>
		/// 
		/// </summary>
		/// <typeparam name="TSource"></typeparam>
		/// <param name="source"></param>
		/// <param name="action">(obj, index)...</param>
		/// <returns></returns>
		public static IEnumerable<TSource> ForEach<TSource>(this IEnumerable<TSource> source, Action<TSource, int> action)
		{
			int index = 0;
			var items = source.ToList();
			items.ForEach(o=> action(o, index++));
			return items;
		}

		/// <summary>
		/// Thread safe yielder.
		/// </summary>
		/// <typeparam name="T"></typeparam>
		/// <param name="source"></param>
		/// <returns></returns>
		public static IceYielder<T> GetYielder<T>(this IEnumerable<T> source)
		{
			return new IceYielder<T>(source);
		}

		/// <summary>
		/// ForEach async, using thread safe yielder. Exception is not handled and is throw.
		/// </summary>
		/// <typeparam name="T"></typeparam>
		/// <param name="source"></param>
		/// <param name="threadsCount"></param>
		/// <param name="action"></param>
		/// <returns></returns>
		public static void ForEachAsync<T>(this IEnumerable<T> source, int threadsCount, Action<T> action, Action<IceThreadPool> onComplete = null, Action<IceThreadPool.PoolItem> onException = null)
		{
			source.ForEachAsync(threadsCount, (o, n) => action(o), onComplete, onException);
		}

		/// <summary>
		/// 
		/// </summary>
		/// <typeparam name="T"></typeparam>
		/// <param name="source"></param>
		/// <param name="threadsCount"></param>
		/// <param name="action">(obj, index)...</param>
		/// <param name="onComplete"></param>
		/// <param name="onException"></param>
		public static void ForEachAsync<T>(this IEnumerable<T> source, int threadsCount, Action<T, int> action, Action<IceThreadPool> onComplete = null, Action<IceThreadPool.PoolItem> onException = null)
		{
			var yielder = source.GetYielder();
			var pool = new IceThreadPool(threadsCount, () =>
			{
				T obj;
				int index = 0;
				while (yielder.YieldNext(out obj))
				{
					action(obj, index++);
				}
			});

			if (onComplete != null)
			{
				pool.PoolCompleteEvent += onComplete;
			}

			if (onException != null)
			{
				pool.ThreadExceptionEvent += onException;
			}
			pool.Start(); // can't use Wait(), it would block action method
		}

	}

}
