﻿#if !SILVERLIGHT

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;
using System.Text.RegularExpressions;
using Ice.Extensions;
using Ice.Utility;
using System.Web.UI;
using System.Web.UI.HtmlControls;
using System.Web.UI.WebControls;
using Ice.Web.Script.Services;
using Ice.Web.UI;
using Ice.Web.UI.WebControls;

namespace Ice.Web.Extensions
{
	public static class ControlExtensions
	{
		#region Utility extensions

		public static T Set<T>(this T control, Action<T> action) where T : Control
		{
			action.Invoke(control);
			return control;
		}

		public static T AppendJsEvent<T>(this T control, JavascriptEventEnum eventName, string js) where T : IAttributeAccessor
		{
			string name = "on" + eventName.GetStringValue();
			string curr = control.GetAttribute(name) ?? "";
			if (!string.IsNullOrEmpty(curr) && !curr.EndsWith(";"))
			{
				curr += ";";
			}
			curr += js;
			if (!curr.EndsWith(";"))
			{
				curr += ";";
			}
			control.SetAttribute(name, curr);
			return control;
		}

		public static string ToHtmlRendered(this Control c)
		{
			System.IO.StringWriter sr = new System.IO.StringWriter();
			System.Web.UI.HtmlTextWriter h = new HtmlTextWriter(sr);
			c.RenderControl(h);
			return sr.ToString();
		}

		public static TextBox TrimText(this TextBox control)
		{
			control.Text = control.Text.Trim();
			return control;
		}

		public static IEnumerable<TextBox> TrimText(this IEnumerable<TextBox> source)
		{
			return source.ForEach(o => o.Text = o.Text.Trim());
		}

		#endregion

		#region Display alert extensions

		public static string JSEncodeString(string str, string encapsulateWith = null)
		{
			return encapsulateWith +
				str.Replace("\r", "\\r")
				.Replace("\n", "\\n")
				.Replace("'", "\\'")
				.Replace("\"", "\\\"")
				+ encapsulateWith;
		}

		public static void DisplayAlert(this Page page, string message)
		{
			page.ExecuteJavaScript("alert(\"" + JSEncodeString(message) + "\");");
		}

		/// <summary>
		/// register with MS ScriptManager; register with IceResponse;
		/// </summary>
		/// <param name="page"></param>
		/// <param name="javaScriptCode"></param>
		public static void ExecuteJavaScript(this Page page, string javaScriptCode)
		{
			if (!javaScriptCode.EndsWith(";"))
			{
				javaScriptCode += ";";
			}

			if (IceScriptManager.IsIceAjaxRequest())
			{
				IceResponse.Current.ExecuteJavaScript(javaScriptCode);
			}
			else
			{
				IceScriptPlaceholder plc;

				if (ScriptManager.GetCurrent(page) != null)
				{
					// register with MS ScriptManager
					ScriptManager.RegisterStartupScript(page, page.GetType(), Guid.NewGuid().ToString(), javaScriptCode + ";", true);
				}
				else if (page.Form != null)
				{
					page.ClientScript.RegisterStartupScript(page.Page.GetType(), Guid.NewGuid().ToString(), javaScriptCode + ";", true);
				}
				else if ((plc = page.DescendantControls().OfType<IceScriptPlaceholder>().FirstOrDefault()) != null)
				{
					// no form, create own script tag
					plc.AddChildren(new LiteralControl(@"<script type=""text/javascript"">
//<![CDATA[
" + javaScriptCode + ";" + @"
//]]>
</script>"));
				}
				else
				{
					throw new Exception("Page must have IceScriptPlaceholder control at the end of the body where script tag will be placed.");
				}
			}
		}

		public static string CreateWindowLocation(string url, int waitMs = 0)
		{
			if(waitMs == 0)
			{
				return "window.location='" + url + "';";
			}
			else
			{
				return "setTimeout(function(){window.location='" + url + "';}, " + waitMs + ");";
			}
		}

		public static void SetWindowLocation(this Page page, string url, int waitMs = 0)
		{
			page.ExecuteJavaScript(CreateWindowLocation(url, waitMs));
		}

		public static void ReloadWindowLocation(this Page page, int waitMs = 0)
		{
			string js;
			if (waitMs == 0)
			{
				js = "window.location.reload()";
			}
			else
			{
				js = "setTimeout(function(){window.location.reload();}, " + waitMs + ");";
			}
			page.ExecuteJavaScript(js);
		}

		#endregion

		#region DOM manipulation extensions

		/// <summary>
		/// returns descendants of TResult type
		/// </summary>
		/// <typeparam name="TResult"></typeparam>
		/// <param name="control"></param>
		/// <returns></returns>
		public static IEnumerable<Control> DescendantControls<TSource>(this TSource control) where TSource:Control
		{
			// select children than concat children's descendants
			// return control.Controls.Cast<Control>().Concat(control.Controls.Cast<Control>().SelectMany(n => n.DescendantControls()));
			// better, dig children's children immediately:
			foreach(Control c in control.Controls)
			{
				yield return c;
				
				foreach(Control c1 in c.DescendantControls())
				{
					yield return c1;
				}
			}
		}

		public static IEnumerable<Control> AncestorControls<TSource>(this TSource control) where TSource:Control
		{
			Control cc = control;
			while ((cc = cc.Parent) != null)
			{
				yield return cc;
			}
		}

		public static IEnumerable<Control> DescendantControls<TSource>(this IEnumerable<TSource> controls) where TSource:Control
		{
			return controls.SelectMany(o => o.DescendantControls());
		}

		public static IEnumerable<Control> AncestorControls<TSource>(this IEnumerable<TSource> controls) where TSource : Control
		{
			return controls.SelectMany(o => o.AncestorControls());
		}

		public static IEnumerable<Control> ParentControls<TSource>(this IEnumerable<TSource> controls) where TSource : Control
		{
			return controls.Select(o => o.Parent);
		}

		public static IEnumerable<Control> ChildControls<TSource>(this IEnumerable<TSource> controls) where TSource : Control
		{
			return controls.SelectMany(o => o.Controls.Cast<Control>());
		}

		public static IEnumerable<Control> ChildControls<TSource>(this TSource control) where TSource : Control
		{
			return control.Controls.Cast<Control>();
		}

		public static TSource RemoveChildren<TSource>(this TSource control) where TSource : Control
		{
			control.Controls.Clear();
			return control;
		}

		public static TSource AddChildren<TSource>(this TSource control, params Control[] children) where TSource : Control
		{
			// must create collection copy because controles are removed from children collection when added to different collection
			children.Where(o=>o != null).ToList().ForEach(o => control.Controls.Add(o));
			return control;
		}

		public static TSource AddChildren<TSource>(this TSource control, params ITemplate[] templates) where TSource : Control
		{
			templates.Where(o=>o != null).ForEach(o=>o.InstantiateIn(control));
			return control;
		}

		public static TSource InsertChildren<TSource>(this TSource control, params Control[] children) where TSource : Control
		{
			// must create collection copy because controles are removed from children collection when added to different collection
			children.Where(o => o != null).Reverse().ToList().ForEach(o => control.Controls.AddAt(0,o));
			return control;
		}

		public static TSource InsertBefore<TSource>(this TSource control, params Control[] items) where TSource : Control
		{
			int index = control.Parent.Controls.IndexOf(control);
			items.Reverse().ForEach(o => control.Parent.Controls.AddAt(index, o));
			return control;
		}

		public static TSource InsertAfter<TSource>(this TSource control, params Control[] items) where TSource : Control
		{
			int index = control.Parent.Controls.IndexOf(control);
			items.Reverse().ForEach(o => control.Parent.Controls.AddAt(index + 1, o));
			return control;
		}

		#endregion

		#region next siblings

		public static IEnumerable<Control> NextSiblings<TSource>(this TSource node) where TSource:Control
		{
			if (node.Parent != null)
			{
				foreach (Control s in node.Parent.Controls.Cast<Control>().SkipWhile(o => o != node).Skip(1))
				{
					yield return s;
				}
			}
		}

		public static IEnumerable<Control> NextSiblings<TSource>(this TSource node, Func<Control, bool> predicate) where TSource:Control
		{
			return node.NextSiblings().Where(predicate);
		}

		public static IEnumerable<Control> NextSiblings<TSource>(this IEnumerable<TSource> list) where TSource:Control
		{
			return list.SelectMany(o => o.NextSiblings());
		}

		public static IEnumerable<Control> NextSiblings<TSource>(this IEnumerable<TSource> list, Func<Control, bool> predicate) where TSource:Control
		{
			return list.NextSiblings().Where(predicate);
		}

		#endregion

		#region previous siblings

		public static IEnumerable<Control> PreviousSiblings<TSource>(this TSource node) where TSource:Control
		{
			if (node.Parent != null)
			{
				foreach (Control s in node.Parent.Controls.Cast<Control>().TakeWhile(o => o != node).Reverse())
				{
					// backward:
					// n0, n1, n2, n3, n4=current, n5, n5, will return nodes: n3, n2, n1, n0
					yield return s;
				}
			}
		}

		public static IEnumerable<Control> PreviousSiblings<TSource>(this TSource node, Func<Control, bool> predicate) where TSource:Control
		{
			return node.PreviousSiblings().Where(predicate);
		}

		public static IEnumerable<Control> PreviousSiblings<TSource>(this IEnumerable<TSource> list) where TSource:Control
		{
			return list.SelectMany(o => o.PreviousSiblings());
		}

		public static IEnumerable<Control> PreviousSiblings<TSource>(this IEnumerable<TSource> list, Func<Control, bool> predicate) where TSource:Control
		{
			return list.PreviousSiblings().Where(predicate);
		}

		#endregion

		#region Css Style extensions

		public static List<KeyValuePair<string, string>> ParseStyleValues(string style)
		{
			return new Regex("([^:;]+):([^:;]*)").Matches(style ?? "").Cast<Match>().Select(m=>new KeyValuePair<string, string>(m.Groups[1].Value.Trim(), m.Groups[2].Value.Trim()))
				.Where(o=>!string.IsNullOrEmpty(o.Key)).ToList();
		}

		/// <summary>
		/// Remove previous style values with the same names and append new vlues.
		/// </summary>
		/// <param name="sc"></param>
		/// <param name="style"></param>
		/// <returns></returns>
		public static CssStyleCollection ApplyValues(this CssStyleCollection sc, List<KeyValuePair<string, string>> style)
		{
			foreach (KeyValuePair<string, string> kv in style)
			{
				// we must remove preivous values, not just overwrite. By removing first, new values will be appended to the end. Css always use last stye value.
				// eg. in this scenario, if "background" is not removed first, background-color will always be blue because "background" is changed to green, but "background-color" overrides it
				//Panel ppp = new Panel();
				//ppp.Style.SetValues("background:red; background-color:blue;");
				//ppp.Style.SetValues("background:green;");

				sc.Remove(kv.Key); // remove it first so the new value will be appended to the end
				
				sc.Add(kv.Key, kv.Value);
			}
			return sc;
		}

		/// <summary>
		/// Remove previous style values with the same names and append new vlues.
		/// </summary>
		/// <param name="sc"></param>
		/// <param name="style"></param>
		/// <returns></returns>
		public static CssStyleCollection ApplyValues(this CssStyleCollection sc, string style)
		{
			return sc.ApplyValues(ParseStyleValues(style));
		}

		/// <summary>
		/// Returns true if it is MS Ajax posback or IceAjax request
		/// </summary>
		/// <param name="page"></param>
		/// <returns></returns>
		public static bool IsInAsyncPostBack(this Page page)
		{
			ScriptManager sm = ScriptManager.GetCurrent(page);
			if(sm != null && sm.IsInAsyncPostBack)
			{
				return true;
			}

			return IceScriptManager.IsIceAjaxRequest();

		}

		public static T ApplyStyle<T>(this T control, string style) where T : IAttributeAccessor
		{
			if (!string.IsNullOrEmpty(style))
			{
				CssStyleCollection col = new HtmlGenericControl().Style.ApplyValues(control.GetAttribute("style")).ApplyValues(style);
				control.SetAttribute("style", col.Value);
			}
			return control;
		}

		public static IEnumerable<T> ApplyStyle<T>(this IEnumerable<T> source, string style) where T : IAttributeAccessor
		{
			return source.ForEach(o => o.ApplyStyle(style));
		}

		#endregion

		#region Css class extensions

		public static T AddClass<T>(this T control, string cssClass) where T : IAttributeAccessor
		{
			if (!string.IsNullOrEmpty(cssClass))
			{
				control.RemoveClass(cssClass); // to avoid duplicates and to place cssClass at the end
				
				if (control is WebControl)
				{
					WebControl cc = (WebControl)(object)control;
					cc.CssClass = (cc.CssClass + " " + cssClass).Trim();
				}
				else
				{
					string cls = control.GetAttribute("class");
					control.SetAttribute("class", (cls + " " + cssClass).Trim());
				}
			}
			return control;
		}

		public static IEnumerable<T> AddClass<T>(this IEnumerable<T> source, string cssClass) where T : IAttributeAccessor
		{
			return source.ForEach(o => o.AddClass(cssClass));
		}

		public static T RemoveClass<T>(this T control, string cssClass) where T : IAttributeAccessor
		{
			if (!string.IsNullOrEmpty(cssClass))
			{
				if (control is WebControl)
				{
					WebControl cc = (WebControl)(object)control;
					cc.CssClass = (" " + cc.CssClass + " ").Replace(" " + cssClass + " ", " ").Trim();
				}
				else
				{
					string cls = " " + control.GetAttribute("class") + " ";
					control.SetAttribute("class", (cls.Replace(" " + cssClass + " ", " ")).Trim());
				}
			}
			return control;
		}

		public static IEnumerable<T> RemoveClass<T>(this IEnumerable<T> source, string cssClass) where T : IAttributeAccessor
		{
			return source.ForEach(o => o.RemoveClass(cssClass));
		}

		public static bool HasClass<T>(this T control, string cssClass) where T : IAttributeAccessor
		{
			if (!string.IsNullOrEmpty(cssClass))
			{
				if (control is WebControl)
				{
					WebControl cc = (WebControl)(object)control;
					return (" " + cc.CssClass + " ").Contains(" " + cssClass + " ");
				}
				else
				{
					string cls = " " + control.GetAttribute("class") + " ";
					return cls.Contains(" " + cssClass + " ");
				}
			}
			else
			{
				return false;
			}
		}

		public static T ToggleClass<T>(this T control, string cssClass) where T : IAttributeAccessor
		{
			if (control.HasClass(cssClass))
			{
				return control.RemoveClass(cssClass);
			}
			else
			{
				return control.AddClass(cssClass);
			}
		}

		public static IEnumerable<T> ToggleClass<T>(this IEnumerable<T> source, string cssClass) where T : IAttributeAccessor
		{
			return source.ForEach(o => o.ToggleClass(cssClass));
		}

		#endregion

		#region ListControl extensions

		public static bool SetSelectedValue(this ListControl control, object value)
		{
			control.SelectedIndex = -1; // select none or first, if value will not be found, this way we have first (default) item selected

			ListItem item = control.Items.FindByValue(value + "");
			if (item != null)
			{
				control.SelectedIndex = control.Items.IndexOf(item);
				return true;
			}
			else
			{
				return false;
			}
		}

		public static bool SetSelectedText(this ListControl control, string text)
		{
			control.SelectedIndex = -1; // select none or first, if value will not be found, this way we have first (default) item selected

			ListItem item = control.Items.FindByText(text);
			if (item != null)
			{
				control.SelectedIndex = control.Items.IndexOf(item);
				return true;
			}
			else
			{
				return false;
			}
		}

		public static TResult GetSelectedValue<TResult>(this ListControl control)
		{
			TResult o;
			IceTypeConverter.TryChangeType<TResult>(control.SelectedValue, out o);
			return o;
		}

		#endregion

		#region TreeView extensions

		public static IEnumerable<TreeNode> DescendantNodes(this TreeNode node)
		{
			if (node != null && node.ChildNodes != null && node.ChildNodes.Count > 0)
			{
				foreach (TreeNode n in node.ChildNodes)
				{
					yield return n;

					foreach (TreeNode n1 in n.DescendantNodes())
					{
						yield return n1;
					}
				}
			}
		}

		public static IEnumerable<TreeNode> AncestorNodes(this TreeNode node)
		{
			while ((node = node.Parent) != null)
			{
				yield return node;
			}
		}

		public static IEnumerable<TreeNode> DescendantNodes(this TreeView treeView)
		{
			return treeView.Nodes.Cast<TreeNode>().Concat(treeView.Nodes.Cast<TreeNode>().SelectMany(o => o.DescendantNodes()));
		}

		#endregion

	}
}

#endif