﻿/// <reference path="Ice.js" />
/// <reference path="Ice.Enumerable.js" />

/*
Modifier	Description
i		Perform case-insensitive matching
g		Perform a global match (find all matches rather than stopping after the first match)
m		Perform multiline matching

Expression	Description
[abc]	Find any character between the brackets
[^abc]	Find any character NOT between the brackets
[0-9]	Find any digit between the brackets
[^0-9]	Find any digit NOT between the brackets
(x|y)	Find any of the alternatives specified

Metacharacter	Description
.		Find a single character, except newline or line terminator
\w	Find a word character
\W	Find a non-word character
\d		Find a digit
\D		Find a non-digit character
\s		Find a whitespace character
\S		Find a non-whitespace character
\b		Find a match at the beginning/end of a word
\B		Find a match not at the beginning/end of a word
\0		Find a NUL character
\n		Find a new line character
\f		Find a form feed character
\r		Find a carriage return character
\t		Find a tab character
\v		Find a vertical tab character
\xxx		Find the character specified by an octal number xxx
\xdd		Find the character specified by a hexadecimal number dd
\uxxxx		Find the Unicode character specified by a hexadecimal number xxxx

Quantifier	Description
n+	Matches any string that contains at least one n
n*	Matches any string that contains zero or more occurrences of n
n?		Matches any string that contains zero or one occurrences of n
n{X}		Matches any string that contains a sequence of X n's
n{X,Y}	Matches any string that contains a sequence of X to Y n's
n{X,}	Matches any string that contains a sequence of at least X n's
n$	Matches any string with n at the end of it
^n	Matches any string with n at the beginning of it
?=n	Matches any string that is followed by a specific string n
?!n	Matches any string that is not followed by a specific string n
*/

Ice.Regex = function (regexPattern, ignoreCase)
{
	/// <summary>
	/// .NET like Regex class.
	/// </summary>
	/// <param name="regexPattern" type="String"></param>
	/// <param name="ignoreCase" type="Boolean" optional="true"></param>
	this._regexPattern = regexPattern || "";
	this._ignoreCase = !!ignoreCase;
};

Ice.Regex.prototype = new (function ()
{
	this._regexPattern = "";
	this._ignoreCase = false;

	function CreateRegex(isGlobal)
	{
		return new RegExp(this._regexPattern, (this._ignoreCase ? "i" : "") + (isGlobal ? "g" : ""));
	}

	function CreateGroupItem(str)
	{
		str = str || "";
		return {
			// length of match text
			Length: str.length,
			// true if match is successful, else false
			Success: !!str,
			// group text value
			Value: str
		};
	}

	function CreateGroups(match)
	{
		var arr = [];
		arr.Any = function () { this.length > 0; };

		Ice.ForEach(match, function (item, index)
		{
			arr.push(CreateGroupItem(item));
		});

		return arr;
	}

	function CreateMatchItem(match)
	{
		var str = match ? (match[0] || "") : "";

		return {
			Groups: CreateGroups(match),
			// start position in original string
			Index: match ? match.index : -1,
			// length of match text
			Length: str.length,
			// true if match is successful, else false
			Success: !!str,
			// complete match text value
			Value: str
		};
	}

	this.IsMatch = function (inputStr)
	{
		/// <summary>
		/// Returns true if regex match is found.
		/// </summary>
		/// <param name="inputStr"></param>
		/// <returns type="Boolean"></returns>
		return CreateRegex.call(this, false).test(inputStr);
	};

	this.Matches = function (inputStr, startIndex, takeMatches)
	{
		/// <summary>
		/// Find all matches. Returns IEnumerable of matched items.
		/// </summary>
		/// <param name="inputStr" type="String"></param>
		/// <param name="startIndex" type="Int" optional="true"></param>
		/// <param name="takeMatches" type="Int" optional="true"></param>

		var arr = [];
		var m;
		var cnt = 0;

		arr.Any = function () { this.length > 0; };

		//#if VSINTELLISNSE
		if (Ice.IsVSIntellisene)
		{
			return new Ice.Enumerable([["all", "group1", "group2"]]).Select(function (o, n) { return CreateMatchItem(o); });
			return arr;
		}
		//#endif

		if (startIndex > 0) { inputStr = inputStr.substr(startIndex); }

		// g flag returns all matches, without g flag, exec returns first match only
		// in while, always use g flag, without g flag, endless loop will occure
		var re = CreateRegex.call(this, true);

		while ((!takeMatches || (takeMatches > 0 && cnt++ < takeMatches)) && (m = re.exec(inputStr)) && m[0])
		{
			arr.push(m);
		}

		return new Ice.Enumerable(arr).Select(function (o, n) { return CreateMatchItem(o); });
	};

	this.Match = function (inputStr, startIndex)
	{
		/// <summary>
		/// Find and returns first match.
		/// </summary>
		/// <param name="inputStr" type="String"></param>
		/// <param name="startIndex" type="Int" optional="true"></param>

		/*
			new RegExp(...).exec(inputStr) with "g" flag returns all matches, without "g" flag, returns first match only
			match item is array with properties:
				0: "ante" // full match
				1: "te"		// first group
				2: "mb"		// second group
				index: 0	// position in inputStr
				input: "anteantefadsantefdsante" // inputStr
				length: 2	// items in this array
		*/

		//#if VSINTELLISNSE
		if (Ice.IsVSIntellisene)
		{
			return CreateMatchItem(["all", "group1", "group2"]);
		}
		//#endif

		return CreateMatchItem(CreateRegex.call(this, false).exec(inputStr));
	};


});

