﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Globalization;
using System.IO;
using System.Reflection;
using System.Text;
using System.Web;
using System.Collections;
using Ice.Web.Script;
using System.Linq;

namespace Ice.Web.Script.Services
{
	using Resources;
	using Serialization;
	using Ice.JavaScript.Packer;
	using Ice.Serialization.Json.Core.Serializer;
	using System.Threading;
	using System.Web.Services;
	using Ice.Web.UI;

    public class IceAjaxHandler
    {
		/// <summary>
		/// -IceRex.axd
		/// </summary>
		public const string IceHandlerPath = "-IceRes.axd";
		/// <summary>
		/// ICE-SC
		/// </summary>
		public const string IceScriptResourceRequest = "ICE-SC";
		/// <summary>
		/// ICE-EX
		/// </summary>
		public const string IceElmahRequest = "ICE-EX";
		/// <summary>
		/// ICE-TM
		/// </summary>
		public const string IceTypeServiceRequest = "ICE-TM";

		static Dictionary<string, Dictionary<string, WebServiceMethodData>> _webServicesCachedDic = new Dictionary<string, Dictionary<string, WebServiceMethodData>>();

		private const string _ajaxPostPrefix = "jsonarg="; // prefix is for Elmah to write request as form parameter (on exception)
		private const string _ajaxGetParam = "jsonarg";

		private IceAjaxVersionEnum _ajaxVersion;
		private HttpContext _httpContext;
		private string _methodName;
		private string _crossdomainOrigin;

		private static FieldInfo _rawUrlFieldInfo = typeof(HttpRequest).GetField("_rawUrl", BindingFlags.Instance | BindingFlags.NonPublic);

		/// <summary>
		/// <para>Retruns true if request was processed, else returns false.</para>
		/// <para>to get JS web service methods with ajax engine, enter url.aspx/wsjs</para>
		/// </summary>
		/// <param name="context"></param>
		/// <param name="webServiceType"></param>
		/// <returns></returns>
		internal void ProcessRequest(HttpContext context, string typeName, long timestampTicks, string methodName, string currentVirtualPath)
		{
			if (methodName == null)
			{
				methodName = GetMethodName(context);
			}

			if (methodName == null)
			{
				throw new InvalidOperationException(AtlasWeb.WebService_InvalidWebServiceCall);
			}

			IceAjaxHandler handler = this;
			
			handler._crossdomainOrigin = context.Request.Headers.Get("Origin");

			handler._methodName = methodName;
			handler._ajaxVersion = IceWebSettings.AjaxVersion;
			handler._httpContext = context;

			object target = null;
			WebServiceMethodData methodData = null;

			try
			{
				// this is critical: if app was precompiled, aspx or ascx type name has changed, so this will throw exception
				// if method doesn't exist, it will throw exception and send js to reload client window
				CompiledTypeInfo ctinfo = null;

				try { ctinfo = WebServiceHelper.GetCompiledType(typeName); } // if type doesn't exists, exception is thrown
				catch { }

				if(!(timestampTicks > 1000) || ctinfo == null || ctinfo.TimestampTicks != timestampTicks)
				{
					////throw new Exception("Assembly timestamp is different, reload window, type: " + typeName);
					// nek ne baca exception
					// nek klijent reloadira window da se povuku nove verzije js engine skripti
					handler.FlushResponseObject(new IceResponse().ReloadWindowLocation(2000));// prevent looping
					return;
				}

				Type serviceType = ctinfo.CompiledType;

				methodData = GetMethodData(context, serviceType, methodName);

				if (!methodData.IsStatic)
				{
					if(typeof(System.Web.UI.Page).IsAssignableFrom(serviceType))
					{
						target = CompiledPageHelper.LoadPage<System.Web.UI.Page>(serviceType);
					}
					else if(typeof(System.Web.UI.UserControl).IsAssignableFrom(serviceType))
					{
						// works with MasterPage too
						target = CompiledPageHelper.LoadControl<System.Web.UI.UserControl>(serviceType);
					}
					else
					{
						target = Activator.CreateInstance(serviceType);
					}
				}
			}
			catch (Exception ex)
			{
				// if exception is thrown, means app was precompilled, it is required to reload the whole page to get new version of js script
				Ice.Elmah.ErrorLog.WriteLog(ex); // write to Elmah
				
				////// sleep on server to prevent looping
				////Thread.Sleep(5 * 1000); // sleep 5 seconds

				// nek klijent reloadira window da se povuku nove verzije js engine skripti
				handler.FlushResponseObject(new IceResponse().ReloadWindowLocation(5000)); // prevent looping
				
				// stop execution here
				return;
			}


			try
			{
				// if webServicemethods return void, it actually returns null and null has to be sent to client to finish the request and to be able to use onSuccess(data) funcion.
				IDictionary<string, object> rawParams = GetRawParams(methodData, context);

				if (!string.IsNullOrEmpty(currentVirtualPath))
				{
					try
					{
						_rawUrlFieldInfo.SetValue(context.Request, currentVirtualPath); // set RawUrl
						// if currentLocation doesn't have query string, RewritePath will keep previous query string
						// if currentLocation has query string, RewritePath will use new query string currentLocation vpath replacing old query string
						// so, we have to remove old query string for sure
						context.RewritePath("/", "", ""); // to remove query string from context.Request.Url, use empty string "", beause null will not remove query string
						context.RewritePath(currentVirtualPath); // so that Request.QueryString may be used in aspx pages

					}
					catch { }
				}

				object resp = methodData.CallMethodFromRawParams(target, rawParams, this._ajaxVersion);

				if (context.Response.RedirectLocation != null)
				{
					handler.FlushResponseObject(new IceResponse().SetWindowLocation(context.Response.RedirectLocation));
				}
				else
				{
					if (methodData.ReturnType.FullName != "System.Void")
					{
						IceResponse.Current.SetResponseObject(resp);
					}

					handler.FlushResponseObject(IceResponse.Current);
				}
			}
			catch (Exception ex)
			{
				if (ex is ThreadAbortException && IceResponse.Current.ResponseEnded)
				{
					handler.FlushResponseObject(IceResponse.Current);
				}
				else
				{
					// unhandled exception or IceWebServiceResponse exception which is not exception, but custom message or something
					handler.FlushResponseObject(ex);
				}
			}
		}

		static object _methodsLock = new object();

		/// <summary>
		/// Get method data from cache or create method data using reflection and cache it
		/// </summary>
		/// <param name="context"></param>
		/// <param name="serviceType"></param>
		/// <param name="methodName"></param>
		/// <returns></returns>
		private WebServiceMethodData GetMethodData(HttpContext context, Type serviceType, string methodName)
		{
			// web service hashtable contains methods per for web service ( WebServiceMethodData  objects)
			Dictionary<string, WebServiceMethodData> methodsInService = _webServicesCachedDic.TryGetValueOrDefault(serviceType.FullName); 
			if (methodsInService == null)
			{
				lock(_methodsLock)
				{
					methodsInService = _webServicesCachedDic.TryGetValueOrDefault(serviceType.FullName); 

					if(methodsInService == null)
					{
						WebServiceData ws = new WebServiceData(serviceType);

						methodsInService = WebServiceHelper.GetWebMethods(serviceType)
						.Select(o => new WebServiceMethodData(ws, o))
						.ToDictionary(o => o.MethodName, o => o);

						_webServicesCachedDic[serviceType.FullName] = methodsInService;
					}
				}
			}

			return methodsInService.TryGetValueOrDefault(methodName);
		}

		private string GetMethodName(HttpContext context)
		{
			/*
			Path		"/pth11/WebForm5.aspx/GetMyNames"
			PathInfo	"/GetMyNames"
			FilePath	"/pth11/WebForm5.aspx"
			*/

			// method name may be queryString parameter or PathInfo:
			// "/pth11/WebForm5.aspx?GetMyNames"
			// "/pth11/WebForm5.aspx/GetMyNames"

			if (context.Request.QueryString.Count > 0)
			{
				return context.Request.QueryString[0];
			}
			else if ((context.Request.PathInfo.Length < 2) || (context.Request.PathInfo[0] != '/'))
			{
				return null;
			}
			else
			{
				return context.Request.PathInfo.Substring(1);
			}
		}

        private IDictionary<string, object> GetRawParams(WebServiceMethodData methodData, HttpContext context)
        {
            return GetRawParamsFromPostRequest(context, methodData.Owner.Serializer);
        }

		private IDictionary<string, object> GetRawParamsFromGetRequest(HttpContext context, IceJavaScriptSerializer serializer, WebServiceMethodData methodData)
        {
            NameValueCollection queryString = context.Request.QueryString;
            Dictionary<string, object> dictionary = new Dictionary<string, object>();
            foreach (WebServiceParameterData data in methodData.ParameterDatas)
            {
                string name = data.ParameterInfo.Name;
                string input = queryString[name];
                if (input != null)
                {
                    dictionary.Add(name, serializer.DeserializeObject(input));
                }
            }
            return dictionary;
        }

		private IDictionary<string, object> GetRawParamsFromPostRequest(HttpContext context, IceJavaScriptSerializer serializer)
        {
			string str = null;

			if (context.Request.HttpMethod == "POST")
			{
				str = new StreamReader(context.Request.InputStream).ReadToEnd();

				if (str.StartsWith(_ajaxPostPrefix))
				{
					// for form-request, eg. if there is input name="jsonarg" value="json string"/>
					str = str.Substring(_ajaxPostPrefix.Length);
				}
				else
				{
					if (str.Length > 500)
					{
						str = str.Substring(500) + "...";
					}
					throw new NotSupportedException("Unknown ajax request: " + str);
				}
			}
			else
			{
				str = context.Request.QueryString[_ajaxGetParam];
			}

            if (string.IsNullOrEmpty(str))
            {
                return new Dictionary<string, object>();
            }

			// we're using double escapeing in Ice.Net.js to prevent from IIS validation (both get and post)
			// first unescape is already done by the sistem
			////for (int n = 1; n < IceWebSettings.EscapeLevel; n++)
			////{
			////	str = HttpUtility.UrlDecode(str);
			////}

			try
			{
				str = Ice.Converter.IceBase64.DecodeBase64StringToString(str);
			}
			catch { }

			if(this._ajaxVersion == IceAjaxVersionEnum.Array)
			{
				IList<Object> args = null;
				try
				{
					args = serializer.Deserialize<IList<Object>>(str);
				}
				catch
				{
					throw new ArgumentException("Arguments not an array!");
				}
				if (args == null)
				{
					throw new ArgumentException("Invalid or missing arguments!");
				}
				Dictionary<string, object> parameters = new Dictionary<string, object>(args.Count);
				for (int nn = 0; nn < args.Count; nn++)
				{
					parameters.Add(nn + "", args[nn]);
				}
				return parameters;
			}
			else // strong names
			{
				return serializer.Deserialize<IDictionary<string, object>>(str);
			}
        }

		string SerializeToJSON(object o)
		{
			if (this._ajaxVersion == IceAjaxVersionEnum.StrongNames_NET35)
			{
				return "{\"d\":" + Ice.Serialization.Json.IceJson.Serialize(o) + "}";
			}
			else
			{
				return Ice.Serialization.Json.IceJson.Serialize(o);
			}
		}

		private bool FlushResponseObject(Exception ex)
		{
			////while(ex is TargetInvocationException && ex.InnerException != null)
			while (ex.InnerException != null)
			{
				ex = ex.InnerException;
			}

			if (ex is IceWebServiceResponse)
			{
				return FlushResponseObject(IceResponse.Current);
			}

			// unhandled exception
			Ice.Elmah.ErrorLog.WriteLog(ex); // write to Elmah

			this._httpContext.Response.TrySkipIisCustomErrors = true;

			if (IceWebSettings.IsDebugDefined) //// || this._httpContext.IsDebuggingEnabled)
			{
				// always full stack
				return FlushResponseObject(new IceResponse().SetExceptionMessage(ex.Message + " in " + _methodName + "\r\n" + ex.StackTrace + " in " + ex.GetType().FullName));
			}
			else
			{
				return FlushResponseObject(new IceResponse().SetExceptionMessage(ex.Message));
			}
		}

		private bool FlushResponseObject(IceResponse resp)
		{
			if (this._httpContext.Response.IsClientConnected)
			{
				this._httpContext.Response.Clear();
				this._httpContext.Response.ClearContent();
				this._httpContext.Response.Cache.SetNoStore();
				this._httpContext.Response.Cache.SetCacheability(HttpCacheability.NoCache);
				this._httpContext.Response.Cache.SetExpires(DateTime.Now.AddDays(-100));

				string json;

				if (!string.IsNullOrEmpty(this._crossdomainOrigin))
				{
					this._httpContext.Response.Headers.Set("Access-Control-Allow-Origin", this._crossdomainOrigin);
				}

				json = SerializeToJSON(resp.GetObjectForSerialization());
				this._httpContext.Response.ContentType = "application/json; charset=utf-8"; // if BinaryWrite is used, add charset=utf-8, if .Write(string) is used, utf-8 is added by default

				if (json.Length > 500)
				{
					// enable compression only if json.Length > 500
					// if length < 500, compressed size is greater than uncompressed!!!!
					Ice.Web.IceHttpCompressionFilter.AddCompressionFilter(this._httpContext);
				}
				else
				{
					Ice.Web.IceHttpCompressionFilter.RemoveCompressionFilter(this._httpContext); // do not use compression because compressed data is larger than original
				}

				this._httpContext.Response.BufferOutput = this._httpContext.Response.Buffer = true;
				// send json
				byte[] data = Encoding.UTF8.GetBytes(json);
				////byte[] delimit = Encoding.ASCII.GetBytes("\r\n\r\n");
				this._httpContext.Response.BinaryWrite(data);
				////this._httpContext.Response.BinaryWrite(delimit);
				//this._httpContext.Response.TrySkipIisCustomErrors = true;
				//this._httpContext.Response.SuppressContent = false;
				//this._httpContext.Response.Flush();
			}

			return true;
		}


    }
}

