﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Text;
using System.Web;
using System.Web.Services;
using System.Text.RegularExpressions;
using Ice.Web.UI;
using System.Linq;
using System.Web.UI;

namespace Ice.Web.Script.Services
{
	public class CompiledTypeInfo
	{
		public Type CompiledType;
		public long TimestampTicks;

		public CompiledTypeInfo(Type serviceType)
		{
			this.CompiledType = serviceType;
			FileInfo fi = new FileInfo(serviceType.Assembly.Location);
			this.TimestampTicks = Math.Max(fi.LastWriteTime.Ticks, fi.CreationTime.Ticks);
		}
	}

	class WebServiceHelper
	{
		#region CompiledTypesHashtable

		private static Dictionary<string, CompiledTypeInfo> mCompiledTypesHashtable = new Dictionary<string, CompiledTypeInfo>();

		/// <summary>
		/// Returns cached Type
		/// </summary>
		/// <param name="context"></param>
		/// <param name="httpPath"></param>
		/// <returns></returns>
		public static CompiledTypeInfo GetCompiledType(HttpContext context, string httpPath)
		{
			CompiledTypeInfo cc;
			if(!mCompiledTypesHashtable.TryGetValue(httpPath, out cc))
			{
				Type service = System.Web.Compilation.BuildManager.GetCompiledType(httpPath); // accepts virtual or app relative path ~/file.asmx
				mCompiledTypesHashtable[httpPath] = cc = new CompiledTypeInfo(service);
			}
			return cc;
		}

		public static CompiledTypeInfo GetCompiledType(string typeName)
		{
			CompiledTypeInfo cc;
			if (!mCompiledTypesHashtable.TryGetValue(typeName, out cc))
			{
				Type service = Type.GetType(typeName, true, true);
				mCompiledTypesHashtable[typeName] = cc = new CompiledTypeInfo(service);
			}
			return cc;
		}

		#endregion

		#region Web Service JS methods definition builder

		public static List<MethodInfo> GetWebMethods(Type type)
		{
			// since we can not retrieve static methods from inherited aspx, or ascx type, use it's base type, and to make sure, use BaseType.BaseType too:

			List<Type> list = new List<Type>() { type };

			if (type.BaseType != null)
			{
				list.Add(type.BaseType);

				if (type.BaseType.BaseType != null)
				{
					list.Add(type.BaseType.BaseType);
				}
			}

			var items = list.SelectMany(o => o.GetMethods(BindingFlags.Public | BindingFlags.Instance | BindingFlags.Static)
					.Where(mi => mi.GetCustomAttributes(typeof(IceWebMethodAttribute), true).Length > 0))
				.Distinct(o=>o.Name) // same function name with different set of arguments is not supported in Javascript - it is not checked here
				.ToList();

			return items;
		}

		/// <summary>
		/// Returns JS string representing web service methods
		/// </summary>
		/// <param name="webServiceType"></param>
		/// <param name="appRelativeUrl"></param>
		/// <returns></returns>
		public static string GetWebServiceJS(FileLocation fileLoc)
		{
			CompiledTypeInfo ctinfo;
			Type webServiceType;

			if(fileLoc.CompiledType != null)
			{
				webServiceType = fileLoc.CompiledType;
			}
			////else if (fileLoc.AppRelativePath != null)
			////{
			////	// web service with many methods
			////	ctinfo = GetCompiledType(HttpContext.Current, fileLoc.AppRelativePath);
			////	webServiceType = ctinfo.CompiledType;
			////}
			else
			{
				throw new NotSupportedException();
			}

			string typeName = webServiceType.AssemblyQualifiedName.Split(',').Take(2).Select(o=>o.Trim()).ToStringJoined(","); // take type name and namespace only, without version

			ctinfo = new CompiledTypeInfo(webServiceType);
			mCompiledTypesHashtable[typeName] = ctinfo; // cache it now
			
			string ns = webServiceType.FullName;
			string baseUrl = VirtualPathUtility.ToAbsolute("~/") + IceAjaxHandler.IceTypeServiceRequest + IceAjaxHandler.IceHandlerPath;


			// get all web methods in CompiledType at once
			List<MethodInfo> methods;
			if (fileLoc.CompiledType != null)
			{
				methods = GetWebMethods(webServiceType);
			}
			else
			{
				throw new NotImplementedException("FileLocation has no AppRelativePath and no CompiledType");
			}

			var items = methods.Select(mi =>
			{
				WebServiceMethodData data = new WebServiceMethodData(null, mi);

				// don't register types, may cause error if registering type which may not be converted to JS object
				if (!mi.ReturnType.FullName.StartsWith("System.") && !mi.ReturnType.FullName.Contains('['))
				{
					// do not add System.Collections....List whic is generic type, it is impossible to create JS valid namespace from List<>
					fileLoc.FileCompiler.RegisterType(mi.ReturnType);
				}

				return new
				{
					MethodInfo = mi,
					Synchronous = data.IsSynchronous ? "true" : "false",
					SessionState = data.SessionState ? "true" : "false",
					RequestTimeout = data.Timeout == null ? "null" : data.Timeout.Value.ToString(),
					JsFunctionArguments = data.ParameterDataDictionary.Select(o => o.Key).Concat(new []{"onComplete", "onError", "onLoading", "timeoutMs"}).ToStringJoined(", "),
					ParametersNames = new Func<string>(() =>
					{
						if (IceWebSettings.AjaxVersion != IceAjaxVersionEnum.Array)
						{
							// IceAjaxVersion.Array send values to server serialized into array: [parentId]
							// this.GetChildren = function (parentId, onComplete, onError, userContext) { return service.F("GetChildren", arguments); };

							// IceAjaxVersion.StrongNames send values to server serialized as: {"parentId": parentId}
							// this.GetChildren = function (parentId, onComplete, onError, userContext) { return service.F("GetChildren", arguments, ["parentId"]); };

							return ", [" + data.ParameterDataDictionary.Select(o => "\"" + o.Key + "\"").ToStringJoined(", ") + "]";
						}
						return null;
					}).Invoke(),
					ReturnType = mi.ReturnType,
					MethodNamespace = mi.DeclaringType.FullName,
					MethodName = mi.Name
				};
			}).ToList();

			StringBuilder sb = new StringBuilder();
			
			if(!items.Any())
			{
				throw new Exception("IceScriptManager no methods found in " + fileLoc.DebuggingFilename);
			}

			// warning: CompiledType may be an aspx page, while method namespace is class where method is declared

			var first = items.First();

			sb.AppendLine("Ice.RegisterNs(\"" + first.MethodNamespace + "\");");

			var items2 = items.Select(o =>
			{
				foreach (var par in o.MethodInfo.GetParameters())
				{
					if (!par.ParameterType.FullName.StartsWith("System.") && !par.ParameterType.FullName.Contains('['))
					{
						// do not add System.Collections....List whic is generic type, it is impossible to create JS valid namespace from List<>
						fileLoc.FileCompiler.RegisterType(par.ParameterType);
					}
				}

				string ttn1 = Ice.Utility.IceTools.GetJavaScriptTypeName(o.ReturnType);
				string callbackArg = ttn1 == "System.Void" ? "" : (ttn1 + " data");

				string parameters = o.MethodInfo.GetParameters().Select((pp, nn) => (nn == 0 ? "\r\n" : null) + "\t\t/// <param name=\"" + pp.Name
					+ "\" type=\"" + Ice.Utility.IceTools.GetJavaScriptTypeName(pp.ParameterType, true)
					+ "\" optional=\"" + (pp.IsOptional ? "true" : "false") + "\"></param>").ToStringJoined("\r\n");

				return "\t" + "this." + o.MethodName + " = function(" + o.JsFunctionArguments + @"){" + parameters + @"
		/// <param name=""onComplete"" type=""Function"" optional=""true"">function(" + callbackArg + @"){...}</param>
		/// <param name=""onError"" type=""Function"" optional=""true"">function(string errorMsg){...}</param>
		/// <param name=""onLoading"" type=""Function"" optional=""true"">function(bool isLoading){....}</param>
		/// <param name=""timeoutMs"" type=""Int"" optional=""true"">request timeout in miliseconds</param>
		//#if DEBUG
		if (Ice.IsVSIntellisene)
		{
			var ttn1 = " + ttn1 + @";
			onComplete.call(this, typeof(ttn1)=='function' ? new ttn1() : ttn1);
			onError.call(this, 'errMsg');
			onLoading.call(this, true);
		}
		//#endif
		return service.FF({method:'" + o.MethodName + "', timeout:" + o.RequestTimeout + ", sync:" + o.Synchronous + ", sstate:" + o.SessionState + "} , arguments" + o.ParametersNames + @");
	};
";
			}).ToList();

			string jsFunctions = items2.ToStringJoined("\r\n");

			// web service url must include assembly timestamp, on webservice request, if timestamp is not the same, send client js to reload window
			
			// do not create absolute url: crossdomain requests has limitations: can't send custom x-ice-ajax header, can't set custom header at all, avoid using crossdomain requests
			////// always create absolute url, cross domain requests are supported now
			////if(string.IsNullOrEmpty(fileLoc.BaseUrl))
			////{
			////	fileLoc.BaseUrl = IceScriptManager.GetServerHttpRoot();
			////}

			string serverUrl = (fileLoc.BaseUrl + WsjsTypeHandler.CreateQstr(baseUrl, typeName, ctinfo.TimestampTicks))
				.Replace("/", "/\" + \"").Replace(".", ".\" + \"").Replace("&", "&\" + \""); // zakamuflirati url da google ne indeksira

			sb.AppendLine(first.MethodNamespace + @".constructor = function(defaultOnComplete, defaultOnError, defaultOnLoading, defaultTimeoutMs)
{
	/// <param name=""defaultOnComplete"" type=""Function"" optional=""true"">function(object data){...}</param>
	/// <param name=""defaultOnError"" type=""Function"" optional=""true"">function(string errorMsg){...}</param>
	/// <param name=""defaultOnLoading"" type=""Function"" optional=""true"">function(bool isLoading){....}</param>
	/// <param name=""defaultTimeoutMs"" type=""Int"" optional=""true"">function(bool isLoading){....}</param>

	var service = this.service = new Ice.Net.WebService(""" + serverUrl + @""", defaultOnComplete, defaultOnError, defaultOnLoading, defaultTimeoutMs);
	
" + jsFunctions + @"
};
// use Ice.Extend to create static methods in " + first.MethodNamespace + @", this way we may use static methods with default timeout and oncomplete and onerror, or we may create new service instance using
// var custom = new " + first.MethodNamespace + @"(defaultOnComplete, defaultOnError, defaultOnLoading, defaultTimeoutMs); with different settings, this is the same as MS asmx service works
Ice.Extend(" + first.MethodNamespace + ", new " + first.MethodNamespace + @".constructor());
");

			if (!string.IsNullOrEmpty(fileLoc.CompiledTypeCustomNamespace))
			{
				sb.AppendLine("Ice.RegisterNs(\"" + fileLoc.CompiledTypeCustomNamespace + "\");");
				sb.AppendLine(fileLoc.CompiledTypeCustomNamespace + " = " + "new " + first.MethodNamespace + ".constructor();");
			}

			////File.WriteAllText("d:\\script" + DateTime.Now.Ticks + ".js", sb.ToString());
			////System.Threading.Thread.Sleep(1);

			return sb.ToString();
		
		}

		/* GENERATED CODE EXAMPLE:
Ice.RegisterNs("Imagine.Web.Customer.Admin.categories");
Imagine.Web.Customer.Admin.categories = function(defaultOnComplete, defaultOnError, defaultTimeoutMs)
{
	var service = this.service = new Ice.Net.WebService("/ICE-TM-IceRes.axd?c=Imagine.Web.Customer.Admin.categories%2cImagine.Web&m=", defaultOnComplete, defaultOnError, defaultTimeoutMs);
	
	this.IceTreeView1_GetChildren = function(parentID, onComplete, onError, userContext){ return service.F("IceTreeView1_GetChildren", arguments, ["parentID"]);};
	this.IceTreeView1_NewNode = function(node, onComplete, onError, userContext){ return service.F("IceTreeView1_NewNode", arguments, ["node"]);};
	this.IceTreeView1_DeleteNodes = function(nodes, onComplete, onError, userContext){ return service.F("IceTreeView1_DeleteNodes", arguments, ["nodes"]);};
	this.IceTreeView1_SaveNodes = function(nodes, onComplete, onError, userContext){ return service.F("IceTreeView1_SaveNodes", arguments, ["nodes"]);};
};
// use Ice.Extend to create static methods in Imagine.Web.Customer.Admin.categories, this way we may use static methods with default timeout and oncomplete and onerror, or we may create new service instance using
// var custom = new Imagine.Web.Customer.Admin.categories(oncomplete, onerror, timeout); with different settings, this is the same as MS asmx service works
Ice.Extend(Imagine.Web.Customer.Admin.categories, new Imagine.Web.Customer.Admin.categories());
		*/

		#endregion


	}
}
