﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Web.UI;
using System.Web;
using System.Text.RegularExpressions;
using System.Collections;
using System.IO;
using Ice.Web.Script.Services;
using System.Linq;
using Ice.Web.UI.WebControls;
using Ice.Web.Extensions;
using Ice.Web.Resources;
using System.Web.UI.WebControls;
using Ice.Web.Resources.Themes;
using Ice.Extensions;
using System.Reflection;
using Ice.JavaScript.Packer;
using Ice.Web.UI.Css;
using System.Web.Services;
using System.Diagnostics;

namespace Ice.Web.UI
{
	public enum IceThemeEnum
	{
		/// <summary>
		/// Default, always included
		/// </summary>
		Basic,
		Aero,
		Silver,
		DeepBlue,
		LightBlue,
		Black
	}

	public enum jQueryVersion
	{
		None = 0,
		jQuery_1_7_1,
		jQuery_1_10_2,
		jQuery_2_1_0
	}

	public class IceScriptManager : Control
	{
		#region asseblies registration

		static Dictionary<Assembly, Assembly> _assemblies = new Dictionary<Assembly, Assembly>() { { Assembly.GetExecutingAssembly(), Assembly.GetExecutingAssembly() } };

		public static void RegisterAssembly(Assembly assembly)
		{
			// has to be static method, otherwise GetCallingAssembly() won't work
			if(assembly != null && !_assemblies.ContainsKey(assembly))
			{
				_assemblies.Add(assembly, assembly);
			}
		}

		#endregion

		#region Dynamic scripts

		List<string> _debugDumpLocations = new List<string>();

		/// <summary>
		/// Location where compiled script will be saved.
		/// </summary>
		/// <param name="location"></param>
		/// <returns></returns>
		public IceScriptManager DumpLocation(string location)
		{
			_debugDumpLocations.Add(location);
			return this;
		}

		/// <summary>
		/// Location where compiled script will be saved.
		/// </summary>
		/// <param name="control"></param>
		/// <returns></returns>
		public IceScriptManager DumpLocation(Control control)
		{
			if (control.TemplateControl != null && control.TemplateControl.AppRelativeVirtualPath != null)
			{
				_debugDumpLocations.Add(control.TemplateControl.AppRelativeVirtualPath);
			}
			return this;
		}

		Dictionary<string, FileLocation> _fileLocations = new Dictionary<string, FileLocation>();

		public IceScriptManager RegisterType(Type objType)
		{
			string key = "type_" + objType.FullName;

			if (!_fileLocations.ContainsKey(key))
			{
				_fileLocations.Add(key, new FileLocation()
				{
					RegisterType = objType
				});
			}
			return this;		
		}

		public IceScriptManager RegisterPageScriptMethods(Type pageType)
		{
			string key = "ws_" + pageType.FullName;

			if (!_fileLocations.ContainsKey(key))
			{
				_fileLocations.Add(key, new FileLocation()
				{
					CompiledType = pageType
				});
			}
			return this;		
		}

		public IceScriptManager RegisterJavaScriptCode(string jscode, string blockName)
		{
			if (!_fileLocations.ContainsKey(jscode))
			{
				_fileLocations.Add(jscode, new FileLocation()
				{
					BlockName = blockName,
					FileContent = jscode
				});
			}
			return this;		
		}

		public IceScriptManager RegisterCssCode(string cssCode, string blockName)
		{
			if (!_fileLocations.ContainsKey(cssCode))
			{
				_fileLocations.Add(cssCode, new FileLocation()
				{
					BlockName = blockName,
					FileContent = cssCode,
					IsCss = true
				});
			}
			return this;
		}

		/// <summary>
		/// 
		/// </summary>
		/// <param name="filename">file to include, embedded or physical path</param>
		/// <param name="appRelativeVirtualPath">main aspx or master file, required for js and css with relative links inside and for intellisense dump to be saved</param>
		/// <returns></returns>
		private IceScriptManager RegisterScriptFileImpl(string filename, string appRelativeVirtualPath)
		{
			if (!_fileLocations.ContainsKey(filename))
			{
				_fileLocations.Add(filename, new FileLocation()
				{
					FileName = filename,
					IsCss = filename.EndsWith(".css"),
					AppRelativeVirtualPath = appRelativeVirtualPath
				});
			}
			return this;
		}

		/// <summary>
		/// Register script files using control's virtual path as root to follow references inside of files content
		/// </summary>
		/// <param name="control"></param>
		/// <param name="files"></param>
		public IceScriptManager RegisterScriptFile(Control control, params string[] files)
		{
			List<string> filesList = new List<string>();

			// always use ascx path first if script is in asxc, this allows .ascx with WebMethod-s to be included in JS with relative path: /// <reference path="ponudaDana.ascx" />
			if (control.TemplateControl != null && control.TemplateControl.AppRelativeVirtualPath != null)
			{
				// vpath is site.master path if exists, else it is .aspx page
				filesList.Add((control.TemplateControl.AppRelativeVirtualPath));
			}

			// than use aspx path, and will include only if there is not control.TemplateControl
			if (control.Page != null && control.Page.AppRelativeVirtualPath != null)
			{
				// vpath is .aspx page now
				filesList.Add((control.Page.AppRelativeVirtualPath));
			}

			foreach (string vpath in filesList)
			{
				files.ForEach(o => this.RegisterScriptFileImpl(o, vpath));
			}

			return this;
		}

		public static string GetServerHttpRoot()
		{
			// value not cached, http <-> https switching wouldn't work if cached
			return (HttpContext.Current.Request.IsSecureConnection ? "https://" : "http://")
				+ HttpContext.Current.Request.ServerVariables["HTTP_HOST"] // includes port number
				;
		}

		public static void EnsureIceWebMethodAttribute(MethodInfo mi)
		{
			if (mi.GetCustomAttributes(typeof(IceWebMethodAttribute), true).Length == 0)
			{
				// custom methods must have WebMethod attribute
				throw new Exception("There is no IceWebMethodAttribute for method: " + mi.Name);
			}
		}


		#endregion

		#region Create scripts code

		static readonly Regex mInlineRegex = new Regex("([\\s]*<script[^>]*?>(?<ttjs>[^\\255]+?)<\\/script>[\\s]*)|([\\s]*<style[^>]*?>(?<ttcss>[^\\255]+?)<\\/style>[\\s]*)", RegexOptions.IgnoreCase);

		/// <summary>
		/// Returns engine, client scripts, and dynamic scripts
		/// </summary>
		/// <returns></returns>
		private List<FileLocation> GetFilesLocations()
		{
			// process child controls or inner scripts
			if (this.Controls.Count > 0) // include inline code
			{
				foreach (LiteralControl c in this.Controls)
				{
					MatchCollection matches = mInlineRegex.Matches(c.Text);

					foreach (Match m in matches)
					{
						if (m.Groups["ttcss"].Success)
						{
							string txt = "\r\n/* INLINE " + this.ID + " */\r\n" + mInlineRegex.Match(c.Text).Groups["ttcss"].Value + "\r\n";
							//files.Add(new FileLocation(this) { IsCss = true, FileContent = txt, FileGroup = FileGroupEnum.ClientScript });
						}
						else if (m.Groups["ttjs"].Success)
						{
							string txt = "\r\n/* INLINE " + this.ID + " */\r\n" + mInlineRegex.Match(c.Text).Groups["ttjs"].Value + "\r\n";
							//files.Add(new FileLocation(this) { IsCss = false, FileContent = txt, FileGroup = FileGroupEnum.ClientScript });
						}
						else
						{
							// unknown match
						}
					}
				}
			}

			// clear controls
			this.Controls.Clear();

			return new List<FileLocation>();
		}

		#endregion

		#region properties

		public IceThemeEnum Theme { get; set; }

		/// <summary>
		/// scripts or web services to include separated by comma ,
		/// </summary>
		public string Scripts { get; set; }

		private CompressionLevel _compression = CompressionLevel.Auto;

		/// <summary>
		/// Script compression level
		/// </summary>
		public CompressionLevel Compression { get { return _compression; } set { _compression = value; } }

		/// <summary>
		/// If true, create inline script
		/// </summary>
		public bool Inline { get; set; }

		// engine is included on first non-inline, non-css script manager
		/// <summary>
		/// If true, always include engine scripts. Engine must be included if there is asmx script added
		/// </summary>
		public bool Engine { get; set; }

		/// <summary>
		/// Include specific jquery version
		/// </summary>
		public jQueryVersion JQuery { get; set; }

		/// <summary>
		/// if true, will not follow references inside js files
		/// </summary>
		public bool SkipReferences { get; set; }

		#endregion

		#region url create & decode

		internal static string CreateUrl(IconEnum icon, string embeddedFile)
		{
			if (icon != IconEnum.None)
			{
				return CreateUrl(icon);
			}
			////else if (btn != WindowButtonsEnum.None)
			////{
			////	return CreateUrl(btn);
			////}
			else if (!string.IsNullOrEmpty(embeddedFile))
			{
				return CreateUrl(embeddedFile);
			}
			else
			{
				throw new InvalidOperationException("icon or btn has to be defined");
			}
		}

		private static string GetAppVirtualPath()
		{
			return VirtualPathUtility.ToAbsolute("~/"); // creates virtual path with / at the end: /pat1/
			//return HttpRuntime.AppDomainAppVirtualPath.TrimEnd('/') + "/"; // create virtual path with / at the end always: path like / or /path/
		}

		public static string CreateUrl(IconEnum icon)
		{
			// insert R into path just to tell VS debugger this is different kind of file
			return GetAppVirtualPath() + IceAjaxHandler.IceScriptResourceRequest + "-R" + IceAjaxHandler.IceHandlerPath + "?i=" + (int)icon + "&t=" + AssemblyTimestamp;
		}

		////internal static string CreateUrl(WindowButtonsEnum btn)
		////{
		////	return GetAppVirtualPath() + IceAjaxHandler.IceScriptResourceRequest + "-R" + IceAjaxHandler.IceHandlerPath + "?w=" + (int)btn + "&t=" + AssemblyTimestamp;
		////}

		internal static string CreateUrl(string embeddedPath)
		{
			return GetAppVirtualPath() + IceAjaxHandler.IceScriptResourceRequest + "-R" + IceAjaxHandler.IceHandlerPath + "?e=" + embeddedPath + "&t=" + AssemblyTimestamp;
		}

		private static string CreateUrl(CachedScript script)
		{
			StringBuilder sb = new StringBuilder();
			sb.Append(script.SourceDirectory);
			sb.Append(IceAjaxHandler.IceScriptResourceRequest);
			sb.Append("-I");
			sb.Append(script.Index); // for VS debugging, VS doesnt read query string, filenames must be unique or VS breakpoints are set into first file available with the same path
			sb.Append(IceAjaxHandler.IceHandlerPath); // /admin/css/ICE-SC-IceRes.axd?p=scriptPathKey&t=timestamp
			sb.Append("?");
			sb.Append("p=");
			sb.Append(script.ScriptPathKey);
			sb.Append("&t=");
			sb.Append(script.Timestamp);
			return sb.ToString();
		}

		internal static string CreateElmahUrl()
		{
			return GetAppVirtualPath() + IceAjaxHandler.IceElmahRequest + IceAjaxHandler.IceHandlerPath;
		}

		internal class DecodedKey
		{
			public string ScriptPathKey { get; set; }

			/// <summary>
			/// version control
			/// </summary>
			public string Timestamp { get; set; }

			public string ResourceFile { get; set; }
		}

		internal static DecodedKey DecodeScriptPathKey(HttpContext context)
		{
			// i: IconEnum
			// w: WindowButtonsEnum
			// t: timestamp ticks
			// p: ScriptPathKey 
			// e: embedded file path

			DecodedKey dd = new DecodedKey();
			dd.Timestamp = context.Request.QueryString["t"];
			dd.ScriptPathKey = context.Request.QueryString["p"];
			dd.ResourceFile = context.Request.QueryString["e"];

			int tmp;
			if (int.TryParse(context.Request.QueryString["i"] ?? "", out tmp))
			{
				dd.ResourceFile = ((IconEnum)tmp).GetStringValue();
			}
			////else if(int.TryParse(context.Request.QueryString["w"] ?? "", out tmp))
			////{
			////	dd.ResourceFile = ((WindowButtonsEnum)tmp).GetStringValue();
			////}


			return dd;
		}

		#endregion

		#region parse Scripts files

		private List<string> SplitFileNames(string fileNames)
		{
			////Regex re = new Regex("\\<\\%[^\\%]+\\%\\>"); // remove asp.net comments
			////string values = re.Replace(value, "");

			// split to lines first, than remove commented lines, than split by , ; \r \n
			// this way we may comment more js files at once if they are at same line separated by , ; ' '

			var items = (fileNames ?? "") // if it is null, must return empty list
				.Split('\r', '\n') // split to lines first
				.Select(o => o.Trim())
				.Where(o => !o.StartsWith("//")) // remove commented lines
				.SelectMany(o => o.Split(',', ';', ' ', '\t')) // split lines content with , ; ' '
				.Select(o => o.Trim())
				.Select(o => o.Contains("?") ? o.Substring(0, o.IndexOf("?")) : o) // remove qstr from files,  eg. versioned files
				.Where(o => !string.IsNullOrEmpty(o))
				.Distinct()
				.ToList();

			return items;
		}

		private FileLocation CreateFileLocation(string file)
		{
			if (file.StartsWith(".") && !file.StartsWith("..")) // embedded file
			{
				return new FileLocation() { FileName = file, IsCss = file.EndsWith(".css", StringComparison.OrdinalIgnoreCase) };
			}
			else
			{
				// virtualPath is needed for css files which must be included with virtual path for images inside css to download correctly
				string appRelativePath = null;

				if (file.StartsWith("~/"))
				{
					appRelativePath = file;
				}
				else if (file.StartsWith("/"))
				{
					// just consider it to be app root
					// svi fileovi na mojem kvartu su ovako inkludani
					// novo, mora biti ovak jer ako se inkludaju css-ovi s / na pocetku pa se promijeni app path iz / u npr. /app1/, ovak ce i dalje sve raditi dobro jer se poistovjecuje / == ~/
					appRelativePath = "~" + file;
				}
				else if (file.StartsWith("../")) // upper directory from aspx or ascx file where is included
				{
					// this.AppRelativeTemplateSourceDirectory returns app-relative directory where current file is (aspx or ascs): "~/NewFolder1/"
					appRelativePath = this.AppRelativeTemplateSourceDirectory + file;
				}
				else if (char.IsLetterOrDigit(file[0])) // "myscript.css" , relative from aspx or ascx file where is included
				{
					appRelativePath = this.AppRelativeTemplateSourceDirectory + file;
				}
				else
				{
					throw new Exception("IceScriptManager error! Can't include file: " + file);
				}

				string extension = Path.GetExtension(appRelativePath);

				bool isCss = false;
				string sourceDirectory = null; // always include / for js files to create single url rooted at app root, not to create different url for every web page
				if (extension == ".css")
				{
					isCss = true;
					// css has to be included using sourceDirectory for images inside or css to work
					string virtualFilePath = VirtualPathUtility.ToAbsolute(appRelativePath); // convert apprelative ~/ into virtual path /...
					sourceDirectory = VirtualPathUtility.GetDirectory(virtualFilePath);
				}
				else
				{
					sourceDirectory = VirtualPathUtility.ToAbsolute("~/");
				}

				return new FileLocation() { FileName = appRelativePath, IsCss = isCss };
			}
		}

		private IEnumerable<FileLocation> ParseInlineScripts()
		{
			// process child controls or inner scripts
			if (this.Controls.Count > 0) // include inline code
			{
				foreach (LiteralControl c in this.Controls)
				{
					MatchCollection matches = mInlineRegex.Matches(c.Text);

					foreach (Match m in matches)
					{
						if (m.Groups["ttcss"].Success)
						{
							string txt = "\r\n/* INLINE " + this.ID + " */\r\n" + mInlineRegex.Match(c.Text).Groups["ttcss"].Value + "\r\n";
							yield return(new FileLocation() { IsCss = true, FileContent = txt });
						}
						else if (m.Groups["ttjs"].Success)
						{
							string txt = "\r\n/* INLINE " + this.ID + " */\r\n" + mInlineRegex.Match(c.Text).Groups["ttjs"].Value + "\r\n";
							yield return(new FileLocation() { IsCss = false, FileContent = txt });
						}
						else
						{
							// unknown match
						}
					}
				}
			}
		}

		#endregion

		#region render tags

		public string GetControlPathKey()
		{
			StringBuilder sb = new StringBuilder("E");

			if (this.TemplateControl != null)
			{
				// this.TemplateControl.AppRelativeVirtualPath is app relative path to aspx, ascx or site.master control where this control is placed (exact file path where IceScriptManager is)
				sb.Append(Math.Abs(this.TemplateControl.AppRelativeVirtualPath.ToLower().GetHashCode()));

				// ako imam master stranicu i ovu kontrolu na njoj, ne zelim novu skriptu za svaku aspx stranicu, zato necu ukljuciti Page.TemplateControl
				////sb.Append(Math.Abs(this.Page.TemplateControl.AppRelativeVirtualPath.GetHashCode()));
			}

			sb.Append(Math.Abs((this.ID ?? "").GetHashCode()));
			sb.Append(Math.Abs((this.ClientID ?? "").GetHashCode()));
			sb.Append(this.Engine ? 1 : 0);
			sb.Append((int)this.Theme);
			sb.Append((int)this.Compression);
			sb.Append(this.Inline ? 1 : 0);
			sb.Append((int)this.JQuery);
			sb.Append(Math.Abs((this.Scripts ?? "").GetHashCode())); // ne gleda se sadrzaj skripti, zato treba mijenjati neki parametear u nazivu, npr. dodati qstr v=x, ili najbolje je koristiti ApplicationTimestamp
			sb.Append(Math.Abs(_fileLocations.Values.Sum(o => (long)o.GetHashCode()))); // obavezno varirati prema dinamicki ukljucenim skriptama
			sb.Append(AssemblyTimestamp); // najsigurnije je ukljuciti assebmly timestamp
			
			return sb.ToString();
		}

		private static string AssemblyTimestamp = "T" + new CompiledTypeInfo(typeof(IceScriptManager)).TimestampTicks;
		//private static string ApplicationTimestamp = "T" + DateTime.Now.Ticks; // must use app start time as var parameter in qstr for debugging to reload on app restart and for js and css which are added via .Scripts property

		static object readWriteLock = new object();

		private List<CachedScript> GetCachedScript(string controlPathKey)
		{
			// urlKey ovisi samo o lokaciji (ClientID)	

			// moram ukljuciti appstarttime jer se url mora promijeniti ako se promijeni sadrzaj skritpte, to mora tako biti definitivno
			// url mora biti drugi kad se app restarta da bi se iz browserra slao request na novi url buduci da se skripta keshira na clijentu, browser ne bi slao request na server, a na serveru je mozda nova verzija skripte
			// pazi: buduci da se skripta keshira na clijentu, svaki novi start aplikacije mora generirati novi url da bi klijentov browser poslao reqest za novom verzijom skripte
			// eTag koristi server, kad user lupi F5, vraca se response 304 not modified ako je etag poklapa

			// create cached script if not already created
			List<CachedScript> scriptsList = CachedScript.GetByControlPathKey(controlPathKey);

			if (scriptsList == null)
			{
				lock (readWriteLock)
				{
					scriptsList = CachedScript.GetByControlPathKey(controlPathKey);

					if (scriptsList == null)
					{
						// compile scripts, always include current assembly
						var asss = _assemblies.Keys.ToArray();
						FileCompiler compiler = new FileCompiler(this, asss);

						////Ice.Elmah.ErrorLog.WriteLog(_assemblies.Keys.Select(o=>o.FullName).ToStringJoined(";;;"), "");

						if (this.JQuery > 0)
						{
							compiler.RegisterJQuery();
						}

						// add engine (there may by engine only without any other files)
						if (this.Engine)
						{
							compiler.RegisterEngine();
						}

						// add other files
						var vv = _fileLocations.Values.ToList();
						
						foreach(FileLocation o in vv)
						{
							if (o.CompiledType != null)
							{
								compiler.RegisterWebService(o.CompiledType);
							}
							else if(o.RegisterType != null)
							{
								compiler.RegisterType(o.RegisterType);
							}
							else if (o.FileContent != null)
							{
								compiler.RegisterContent(o.FileContent, o.IsCss, o.BlockName);
							}
							else if (o.FileName != null)
							{
								compiler.RegisterFile(o.FileName, o.AppRelativeVirtualPath);
							}
							else
							{
								throw new NotSupportedException("missing file content or file name");
							}
						}

						////Ice.Elmah.ErrorLog.WriteLog(vv.Select(o => o.AppRelativeVirtualPath).ToStringJoined(";; "), "");

						// ad Scripts="..."
						// scripts added using property this.Sripts:
						SplitFileNames(this.Scripts).ForEach(o =>
						{
							FileLocation loc = CreateFileLocation(o);
							compiler.RegisterFile(loc.FileName, this.AppRelativeTemplateSourceDirectory);
						});


						// add inline scripts
						ParseInlineScripts().ForEach(o =>
						{
							compiler.RegisterContent(o.FileContent, o.IsCss, "INLINE SCRIPT BLOCK");
						});
						this.Controls.Clear();


						// generate final scripts:
						var cssBlocks = compiler.GetCssBlocks().Select(o => new FileLocation()
						{
							BaseUrl = o.Key,
							FileContent = o.Value,
							IsCss = true,
						}).ToList();

						FileLocation js = new FileLocation() { 
							FileContent = compiler.GetJavaScriptContent(),
							IsCss = false,
						};

						////// for intellisense dump:
						////// there may be .aspx files if they were included
						////List<string> pathsList = _fileLocations.Select(o => o.Value.AppRelativeVirtualPath).Where(o => !string.IsNullOrEmpty(o)).Distinct().ToList();

						////if(IceWebSettings.IsDebugDefined && this.TemplateControl != null && !pathsList.Any())
						////{
						////	pathsList.Add(this.TemplateControl.AppRelativeVirtualPath);
						////}


						// eTag koristi server, kad user lupi F5, vraca se response 304 not modified ako je etag poklapa
						// etag generiram, koristi server za provjeru kad klijent lupi F5, vidi u WsjsClientProxyHandler.cs

						DateTime now = DateTime.Now;

						scriptsList = new List<CachedScript>();
						int index = 0;
						int cssIndex = -1;
						int jsIndex = -1;

						foreach (FileLocation loc in cssBlocks.Concat(js).Where(o => !string.IsNullOrEmpty(o.FileContent)))
						{
							//*************************************************************************
							// preprocess for #define, #if, #else, #endif
							List<string> definedWords = new List<string>();
							if (IceWebSettings.IsDebugDefined)
							{
								definedWords.Add("DEBUG");
							}
							List<string> linesList = new List<string>();

							Regex re = new Regex("((\\/\\/)|(\\/\\*))[\\s]*(?<kw>#[\\w]+)([\\s]*(?<negation>[\\!]*)[\\s]*(?<word>[\\d\\w]+))*");

							Stack<bool> includeState = new Stack<bool>();
							includeState.Push(true); // include lines if not defined otherwise
							
							foreach (string line in loc.FileContent.Split('\n').Select(o=>o.Trim('\r')))
							{
								Match m = re.Match(line);
								if (m.Success)
								{
									linesList.Add(line); // add command line just to have it for debugging

									switch (m.Groups["kw"].Value)
									{
										case "#define":
											definedWords.Add(m.Groups["word"].Value);
											break;
										case "#if":
											bool contains = definedWords.Contains(m.Groups["word"].Value);
											bool isNegation = m.Groups["negation"].Value == "!";
											includeState.Push(contains != isNegation);
											break;
										case "#else":
											includeState.Push(!includeState.Pop()); // invert value
											break;
										case "#endif":
											includeState.Pop();
											break;
									}
								}
								else if (includeState.Peek())
								{
									linesList.Add(line);
								}
								else if (!loc.IsCss)
								{
									linesList.Add("// " + line); // add commented line in JS only
								}
							}

							string scriptCode = linesList.ToStringJoined("\r\n");

							//*************************************************************************

							// pazi: radi dinamicki generiranih css-ova sa urlovima do embedded slika, css nakon svakog recompilea ima drugiciji sadrzaj i hash value
							// pazi: radi dinamicke verzije servisa u url-u koji se ubacuje u js, js nakon recompilea ima drugaciji sadrzaj i hash value
							CompressionLevel compressLevel = this.Compression;
							if (compressLevel == CompressionLevel.Auto)
							{
								compressLevel = IceWebSettings.IsDebugDefined ? CompressionLevel.None : CompressionLevel.Pack;
							}

							if (compressLevel > CompressionLevel.None)
							{
								// css or js, always do basic minify to remove comments
								scriptCode = IcePacker.Pack(scriptCode, CompressionLevel.Minify, loc.IsCss);

								if (!loc.IsCss)
								{
									if (compressLevel == CompressionLevel.Shrink || compressLevel == CompressionLevel.Pack)
									{
										// shrink first
										Microsoft.Ajax.Utilities.Minifier min = new Microsoft.Ajax.Utilities.Minifier();
										Microsoft.Ajax.Utilities.CodeSettings set = new Microsoft.Ajax.Utilities.CodeSettings();
										scriptCode = min.MinifyJavaScript(scriptCode, set);

										// than pack if required
										if (compressLevel == CompressionLevel.Pack)
										{
											scriptCode = IcePacker.Pack(scriptCode + ";", CompressionLevel.Pack, loc.IsCss);
										}
									}
								}
							}

							if(string.IsNullOrEmpty(scriptCode))
							{
								// ignore empty blocks
								continue;
							}

							// string.GetHashCode() ovisi o sadrzaju stringa, znaci, isti sadraj, uvijek daje isti hash
							string etag = "T" + Math.Abs((scriptCode).GetHashCode()).ToString() + scriptCode.Length; // vary by content, not by app start time any more

							CachedScript ss = new CachedScript()
							{
								Manager = this,
								ETag = etag,
								IsCSS = loc.IsCss,
								ScriptContent = scriptCode,
								ScriptLength = scriptCode.Length,
								// just to make 100% sure there is source directory set
								// if source directory is not set, it will genereate url with virtual path /, even if eg. app path is /app/
								// embedded files don't have source directory set, so it must be set here:
								SourceDirectory = VirtualPathUtility.ToAbsolute(loc.BaseUrl ?? "~/"),
								ControlPathKey = controlPathKey,
								Timestamp = etag, // vary by content, not by app start time any more
								Index = index++,
								LastModified = new DateTime(now.Year, now.Month, now.Day, now.Hour, now.Minute, now.Second) // DateTime bez milisekundi jer browser ne salje milisekunde
							};

							ss.ScriptPathKey = ss.ControlPathKey + "I" + ss.Index;
							ss.Url = CreateUrl(ss); // without / at start so files will be relative to current page, this feature is required for css and images inside to work
							scriptsList.Add(ss);

							if (IceWebSettings.IsDebugDefined && _debugDumpLocations.Any())
							{
								string sufix = null;

								if(loc.IsCss)
								{
									cssIndex++;
									sufix = (cssIndex > 0 ? ("." + cssIndex) : "") + ".compiled.css";
								}
								else
								{
									jsIndex++;
									sufix = (jsIndex > 0 ? ("." + jsIndex) : "") + ".compiled.js";
								}

								_debugDumpLocations.Distinct().ForEach(p =>
								{
									// there may be many css files
									// there may be only 1 js file
									string path = System.Web.Hosting.HostingEnvironment.MapPath(p + sufix); // make sure sufix is included, not to owervrite original .aspx file
									File.WriteAllText(path, loc.FileContent);
								});
							}
						};

						CachedScript.AddToCache(scriptsList);
					}
				}
			}
			return scriptsList;
		}

		private void WriteScript(HtmlTextWriter writer, CachedScript script)
		{

			if (this.Inline)
			{
				if (script.IsCSS)
				{
					writer.Write(string.Format(@"
<style type=""text/css"">
{0}
</style>", script.ScriptContent));
				}
				else
				{
					writer.Write(string.Format(@"
<script type=""text/javascript"">
{0}
</script>", script.ScriptContent));
				}
			}
			else
			{
				if (script.IsCSS)
				{
					writer.Write(string.Format(@"
<link rel=""stylesheet"" type=""text/css"" href=""{0}"" />",
						script.Url));
				}
				else
				{
					writer.Write(String.Format(@"
<script type=""text/javascript"" src=""{0}""></script>", script.Url));
				}
			}

			base.Render(writer);
		}

		protected override void OnPreRender(EventArgs e)
		{
			this.EnableViewState = false;

			base.OnPreRender(e);

			if (this.Engine && this.Page != null && this.Page.ClientScript != null && !(this.Page.Items["_iceOnSubmit"] as bool?).GetValueOrDefault())
			{
				this.Page.Items["_iceOnSubmit"] = true;
				this.Page.ClientScript.RegisterOnSubmitStatement(this.GetType(), Guid.NewGuid().ToString(), "if(typeof(Ice) != 'undefined'){return Ice.Page.OnBeforeSubmit.Invoke();}");
			}

		}

		protected override void RenderChildren(HtmlTextWriter writer)
		{
			// this override is required to prevent children from beeng rendered (eg. script or style inline content)

			////StringWriter s = new StringWriter();
			////HtmlTextWriter w = new HtmlTextWriter(s);
			////base.RenderChildren(w);

			// do not render children, add it to the main script file		 
			////base.RenderChildren(writer);
		}

		protected override void Render(HtmlTextWriter writer)
		{
			if (!this.Visible)
			{
				return;
			}

			if (this.DesignMode)
			{
				return;
			}

			foreach (CachedScript script in GetCachedScripts())
			{
				WriteScript(writer, script);
			}
		}

		public List<CachedScript> GetCachedScripts()
		{
			string controlPathKey = GetControlPathKey();
			List<CachedScript> scriptsList = GetCachedScript(controlPathKey);
			return scriptsList;
		}

		public string GetJavascriptCode()
		{
			return GetCachedScripts().Where(o => !o.IsCSS).Select(o => o.ScriptContent).ToStringJoined("\r\n");
		}

		public string GetCssCode()
		{
			return GetCachedScripts().Where(o => !o.IsCSS).Select(o => o.ScriptContent).ToStringJoined("\r\n");
		}

		#endregion

		#region helper methods

		public static IceScriptManager GetCurrent(Control control)
		{
			Page page = control.Page;

			IceScriptManager manager = page.Items["_icescrmanager"] as IceScriptManager;

			if (manager == null)
			{
				page.Items["_icescrmanager"] = manager = page.DescendantControls().OfType<IceScriptManager>().Where(o => o.Engine).FirstOrDefault();
			}

			if (manager == null)
			{
				throw new Exception("Page must include IceScriptManager control with Engine=true");
			}

			return manager;
		}

		internal ThemeBase CurrentTheme
		{
			get { return ThemeHelper.GetTheme(this.Theme); }
		}

		public static bool IsIceAjaxRequest()
		{
			return HttpContext.Current != null && HttpContext.Current.Request.Headers["x-ice-ajax"] == "ice-ajax";
		}

		#endregion

		#region automated page js inclusion
		
		/// <summary>
		/// Ensure file exists. Throw exception if file not found.
		/// </summary>
		/// <param name="filenames"></param>
		private static void EnsureFileExists(params string[] filenames)
		{
			foreach (string filename in filenames)
			{
				if (!_assemblies.Keys.Where(o => Ice.Utility.EmbeddedHelper.ContainsFile(Ice.Utility.EmbeddedHelper.GetEmbeddedFilePath(filename), o)).Any())
				{
					throw new FileNotFoundException(filename);
				}
			}
		}


		class AutoFile
		{
			public Control Control { get; set; }
			public string VirtualFile { get; set; }
			public List<string> EmbeddedFiles = new List<string>();
		}

		// Key: virtual path filename
		private static Dictionary<string, AutoFile> _embeddedFile = new Dictionary<string, AutoFile>();

		private void AutomatedFileInclusion()
		{
			//************************************************************
			//.ascx files can't be included this way, there is no reference to ascx file 
			//************************************************************

			return; // not used, since 

			if (this.Page != null && !this.Page.IsInAsyncPostBack())
			{
				// autmatically include .js files for site.master and for .aspx files
				// this.TemplateControl.AppRelativeVirtualPath is app relative path to aspx, ascx or site.master control where this control is placed (exact file path where IceScriptManager is)
				List<AutoFile> filesList = new List<AutoFile>();

				// .ascx files can't be included this way, there is no reference to ascx file

				if (this.Page != null && this.Page.AppRelativeVirtualPath != null)
				{
					// vpath is .aspx page now
					filesList.Add(new AutoFile() { Control = this.Page, VirtualFile = this.Page.AppRelativeVirtualPath });
				}

				if (this.TemplateControl != null && this.TemplateControl.AppRelativeVirtualPath != null)
				{
					// vpath is site.master path if exists, else it is .aspx page
					// if there is no site.master, TemplateControl == Page, so use Distinct(o=>o.VirtualFile) later
					filesList.Add(new AutoFile() { Control = this.TemplateControl, VirtualFile = this.TemplateControl.AppRelativeVirtualPath });
				}

				foreach (AutoFile tempAutoFile in filesList.Distinct(o => o.VirtualFile))
				{
					AutoFile cachedAutoFile;

					if (!_embeddedFile.TryGetValue(tempAutoFile.VirtualFile, out cachedAutoFile))
					{
						// test all files combinations and save files which exists
						foreach (string file in new string[] { ".ice.css", ".css", ".js" }.Select(o => tempAutoFile.VirtualFile + o))
						{
							string tmp = Ice.Utility.EmbeddedHelper.GetEmbeddedFilePath(file);

							if (_assemblies.Keys.Where(a => Ice.Utility.EmbeddedHelper.ContainsFile(tmp, a)).Any())
							{
								tempAutoFile.EmbeddedFiles.Add(tmp);
							}
						}
						cachedAutoFile = tempAutoFile;
						_embeddedFile.Add(tempAutoFile.VirtualFile, tempAutoFile);
					}

					// register files which exists
					cachedAutoFile.EmbeddedFiles.ForEach(o => this.RegisterScriptFile(cachedAutoFile.Control, o));

				}
			}
		}

		protected override void OnInit(EventArgs e)
		{
			base.OnInit(e);

			// don't use it, it searches for files every time on every aspx page
			AutomatedFileInclusion();
		}
		

		#endregion
	}
}